/*
 * SpringProx or K531 SDK :: direct/ascii.c
 * ----------------------------------------
 *
 * (Minimal) implementation of K531 or SpringProx ASCII protocl
 *
 * IMPORTANT NOTICE:
 *     This implementation works only for payload size < 128 bytes
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>

#include "ref_direct.h"

int AsciiSendB(unsigned char b)
{
  unsigned char c;
  int e;

  c = (b >> 4) & 0x0F;
  if (c >= 10)
    c += ('A' - 10);
  else
    c += '0';

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf("%c", c);
#endif

  e = SerialSend(&c, 1);
  if (e != 1) return e; /* Communication error */

  c = b & 0x0F;
  if (c >= 10)
    c += ('A' - 10);
  else
    c += '0';

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf("%c", c);
#endif

  e = SerialSend(&c, 1);
  if (e != 1) return e; /* Communication error */

  return 1; /* 1 byte sent */
}

/*
 * AsciiSend
 * ---------
 * Format and send a frame to reader according to ASCII protocol
 * Returns size of payload on success, error code otherwise
 *
 * Payload is [CMD CODE] [ .... OPTIONAL DATA .... ]
 */

int AsciiSend(unsigned char *payload, int payload_sz)
{
  int e, i;
  unsigned char b;

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf("PCD < $");
#endif
  
  if (payload_sz >= 128) return E_INTERNAL;

  /* Start stream with '$' */
  b = '$';
  e = SerialSend(&b, 1);
  if (e != 1) return e; /* Communication error */
  
  /* We expect some echo... */
  do
  {
    e = SerialRecv(&b, 1);
    if (e < 0)  return e;
    if (e == 0) return E_READER_MUTE;
  } while (b != '$');

  /* Send command (first byte of Buffer) */
  e = AsciiSendB(payload[0]);
  if (e <= 0) return e;

  /* Send length of data (this is size of payload minus 1 for the command code) */
  e = AsciiSendB((unsigned char) (payload_sz - 1));
  if (e <= 0) return e;

  /* Send data in a loop */
  for (i=1; i<payload_sz; i++)
  {
    e = AsciiSendB(payload[i]);
    if (e <= 0) return e;
  }

  /* End of stream (CR/LF) */
  b = 0x0D;
  e = SerialSend(&b, 1);
  if (e <= 0) return e;

  b = 0x0A;
  e = SerialRecv(&b, 1);
  if (e <= 0) return e;

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf(" ...");
#endif

  /* Now we expect the whole frame in echo */
  do
  {
    e = SerialRecv(&b, 1);
    if (e < 0) return e;
    if (e == 0) break; /* No more echo ? */

  } while (b != 0x0A); /* LF */

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf("\r\n");
#endif

  /* OK */
  return payload_sz;
}

unsigned char AsciiDecode2(unsigned char c[2])
{
  unsigned char r, b;

  b = c[0];
  if ((b >= '0') && (b <= '9'))
    b = b - '0';
  else if ((b >= 'A') && (b <= 'F'))
    b = b + 10 - 'A';
  r = b << 4;

  b = c[1];
  if ((b >= '0') && (b <= '9'))
    b = b - '0';
  else if ((b >= 'A') && (b <= 'F'))
    b = b + 10 - 'A';
  r |= b;

  return r;
}

int AsciiRecv2(unsigned char c[2])
{
  int i;
  unsigned char b;

  for (i = 0; i < 2; i++)
  {
    for (;;)
    {
      if (SerialRecv(&b, 1) != 1)
        return E_READER_STOP; /* No more byte ? */

#ifdef WITH_DEBUG_CODE
      if (TraceReader)
        printf("%c", b);
#endif

      if (((b >= '0') && (b <= '9'))
       || ((b >= 'A') && (b <= 'F')))
      {
        /* Valid Ascii byte */
        c[i] = b;
        break;
      }
      /* Discard invalid bytes */
    }
  }

  return 0;
}

int AsciiRecv(unsigned char *payload, int max_payload_sz)
{
  int e, i, payload_sz;
  unsigned char b;
  unsigned char c[2];

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf("PCD > ");
#endif

  e = 0;
  for (;;)
  {
    if (SerialRecv(&b, 1) <= 0)
      return E_READER_MUTE;  /* No answer ? */

#ifdef WITH_DEBUG_CODE
    if (TraceReader)
      printf("%c", b);
#endif

    if (b == '-')
    {
      /* This is an error reply */
      e = S_BAD_RESP_CODE;
      break;
    } else
    if (b == '+')
    {
      /* This is a success reply */
      break;
    }
    /* Discard other characters */
  }

  /* Receive the command code */
  e = AsciiRecv2(c);
  if (e < 0) return e;

  /* Decode the command code */
  if ((payload != NULL) && (max_payload_sz >= 1))
    payload[0] = AsciiDecode2(c);

  /* Receive the length */
  e = AsciiRecv2(c);
  if (e < 0) return e;

  /* Decode the length */
  payload_sz = AsciiDecode2(c);

  /* Receive payload data in a loop */
  for (i=1; i<=payload_sz; i++)
  {
    /* Receive next byte */
    e = AsciiRecv2(c);
    if (e < 0) return e;

    /* Decode received data */
    if ((payload != NULL) && (max_payload_sz > i))
      payload[i] = AsciiDecode2(c);
  }

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf(" ...");
#endif

  /* CR / LF are now expected */
  SerialRecv(&b, 1);
  SerialRecv(&b, 1);

#ifdef WITH_DEBUG_CODE
  if (TraceReader)
    printf(" (%d)\r\n", i);
#endif

  /* See if the response has overflowed the buffer */
  if ((payload != NULL) && (i > max_payload_sz)) return E_OVERFLOW; 

  /* OK, return actual length */
  return i;
}
