/*
 * SpringProx or K531 SDK :: direct/askcts.c
 * -----------------------------------------
 *
 * (Partial) implementation of ASK CTS256B and CTS512B RFID chips
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

int AskCtsProc(void)
{
  unsigned char type[2];
  unsigned char uid[4];
  unsigned char cmd[3];
  unsigned char data[64];
  int r, t;
  unsigned char i;

  /* Enter ISO/IEC 14443-B mode */
  /* -------------------------- */
  cmd[0] = 0x58;                /* SPROX_CONTROL */
  cmd[1] = 0x0C;                /* SPROX_CONTROL_CONFIG */
  cmd[2] = 2;                   /* ISO/IEC 14443-B */  
  r = Send(cmd, 3);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  /* Configure the MfRC531 chipset to accept ASK CTS B specific frames */
  /* ----------------------------------------------------------------- */
  cmd[0] = 0xAC;                /* SPROX_PCDWRITERCREG */
  cmd[1] = 0x17;                /* RegTypeBFraming */
  cmd[2] = 0x00;
  r = Send(cmd, 3);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  cmd[0] = 0xAC;                /* SPROX_PCDWRITERCREG */
  cmd[1] = 0x1D;                /* RegBPSKDemControl */
  cmd[2] = 0xFE;
  r = Send(cmd, 3);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  /* Infinite loop, use CTRL+C to exit */
  printf("Waiting for a ST SR176 card\n");
  printf("Put your card on the reader, or hit CTRL+C to exit\n");

  for (;;)
  {
    r = AskCtsSelect(type, uid);
    if (r < 0) return r; /* Fatal error */
    if (r == 0)
      break;
  }

  switch (type[1] & 0xF0)
  {
    case 0x50 : t = 256; break;
    case 0x60 : t = 512; break;
    default   : t = 0;
  }

  /* Say hello */
  if (t)
    printf("Found a card that looks like an ASK CTS%dB\n", t);
  else
    printf("Found a card that looks like an unknown ASK CTS\n");
  printf("Manu=%02X Type=%02X UID=", type[0], type[1]);
  for (i=0; i<4; i++)
    printf("%02X", uid[i]);
  printf("\n");

  if (t == 0) return 0;

  /* And now read the card to display its whole content */
  /* -------------------------------------------------- */

  if (t == 256)
  {
    /* Read 16 blocks */
    printf("Reading 16 blocks...\n");
    for (i=0; i<16; i++)
    {
      r = AskCts256ReadBlock(i, &data[2*i]);
      if (r != 0)
      {
        printf("Failed to read at address %d !\n", i);
        return r;
      }
    }
    
    for (i=0; i<16; i++)
      printf("%02d : %02X%02X\n", i, data[2*i], data[2*i+1]);

  } else
  if (t == 512)
  {
    /* Read 32 blocks */
    printf("Reading 32 blocks...\n");
    for (i=0; i<32; i+=4)
    {
      r = AskCts512ReadBlocks(i, &data[2*i]);
      if (r != 0)
      {
        printf("Failed to read at address %d !\n", i);
        return r;
      }
    }

    for (i=0; i<32; i++)
      printf("%02d : %02X%02X\n", i, data[2*i], data[2*i+1]);

  }



  /* Done ! */
  return 0;
}

/*
 * AskCts256ReadBlock
 * ------------------
 * This is the CTS256B READ command
 */
int AskCts256ReadBlock(unsigned char addr, unsigned char data[2])
{
  unsigned char Cmd[14];
  unsigned char Rsp[64 + 3 + 1];
  int r;

  /* Send the READ(address) command as a type B frame */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x03;                /* Send length = 1 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x00;
  Cmd[7] = 0x6A;                /* Timeout = 1ms = 106 ETU */

  Cmd[8] = 0x10 | addr;         /* CTS256 READ command */

  Cmd[9] = 0xCC;
  Cmd[10] = 0xCC;               /* Room for CRC B (will be computed by reader) */
  Send(Cmd, 11);
  r = Recv(Rsp, sizeof(Rsp));

  if ((r < 5) || (Rsp[0] != 0x00))
    return S_NO_CARD;           /* No answer ? */

  /* Bytes are swaped in card */
  if (data != NULL)
    memcpy(data, &Rsp[3], 2);

  return 0;                     /* OK ! */
}

/*
 * AskCts256SetBlock
 * -----------------
 * This is the CTS256B WRITE command
 * The WRITE command is an OR operation. To write 0, use ERASE (Reset) !
 */
int AskCts256SetBlock(unsigned char addr, unsigned char mask[2])
{
  unsigned char Cmd[14];
  int r;

  if (mask == NULL)
    return E_INTERNAL;

  /* Send the WRITE(address) command as a type B frame together with data */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x05;                /* Send length = 3 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x00;
  Cmd[7] = 0xD3;                /* Timeout = 2ms = 212 ETU */

  Cmd[8] = 0x20 | addr;         /* CTS256 WRITE command */

  Cmd[9]  = mask[0];
  Cmd[10] = mask[1];

  Cmd[11] = 0xCC;
  Cmd[12] = 0xCC;               /* Room for CRC B (will be computed by reader) */

  Send(Cmd, 13);

  r = Recv(NULL, 0);            /* Don't expect any answer (card is mute after a WRITE) */
  if (r < 0) return r;

  return 0;                     /* OK ! */
}

/*
 * AskCts256ResetBlock
 * -------------------
 * This is the CTS256B ERASE command
 * The ERASE command puts 0000 at specified address
 */
int AskCts256ResetBlock(unsigned char addr)
{
  unsigned char Cmd[14];
  int r;

  /* Send the ERASE(address) command as a type B frame */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x03;                /* Send length = 1 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x00;
  Cmd[7] = 0xD3;                /* Timeout = 2ms = 212 ETU */

  Cmd[8] = 0x40 | addr;         /* CTS256 ERASE command */

  Cmd[9] = 0xCC;
  Cmd[10] = 0xCC;               /* Room for CRC B (will be computed by reader) */
  Send(Cmd, 11);

  r = Recv(NULL, 0);            /* Don't expect any answer (card is mute after a WRITE) */
  if (r < 0) return r;

  return 0;                     /* OK ! */
}

/*
 * AskCts512SelectAll
 * ------------------
 * This is the CTS512B SELECT command, with parameter 0xFFFF (= ALL)
 */
int AskCts512SelectAll(void)
{
  unsigned char Cmd[14];
  unsigned char Rsp[64 + 3 + 1];
  int r;

  /* Send the SELECT command as a type B frame */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x05;                /* Send length = 3 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x00;
  Cmd[7] = 0x6A;                /* Timeout = 1ms = 106 ETU */

  Cmd[8] = 0x9F;                /* CTS512 SELECT(ALL) command */
  Cmd[9] = 0xFF;                
  Cmd[10] = 0xFF;               

  Cmd[11] = 0xCC;
  Cmd[12] = 0xCC;               /* Room for CRC B (will be computed by reader) */
  Send(Cmd, 13);
  r = Recv(Rsp, sizeof(Rsp));
  
  if ((r < 3) || (Rsp[0] != 0x00))
    return S_NO_CARD;           /* No answer ? */

  return 0;                     /* OK ! */
}

/*
 * AskCts512ReadBlock
 * ------------------
 * This is the CTS512B READ command
 */
int AskCts512ReadBlock(unsigned char addr, unsigned char data[2])
{
  unsigned char Cmd[14];
  unsigned char Rsp[64 + 3 + 1];
  int r;

  /* Send the READ(address) command as a type B frame */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x03;                /* Send length = 1 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x00;
  Cmd[7] = 0x6A;                /* Timeout = 1ms = 106 ETU */

  Cmd[8] = 0xC0 | addr;         /* CTS512 READ command */

  Cmd[9] = 0xCC;
  Cmd[10] = 0xCC;               /* Room for CRC B (will be computed by reader) */
  Send(Cmd, 11);
  r = Recv(Rsp, sizeof(Rsp));

  if ((r < 5) || (Rsp[0] != 0x00))
    return S_NO_CARD;           /* No answer ? */

  if (data != NULL)
    memcpy(data, &Rsp[3], 2);

  return 0;                     /* OK ! */
}

/*
 * AskCts512ReadBlocks
 * -------------------
 * This is the CTS512B MULTI-READ command
 */
int AskCts512ReadBlocks(unsigned char addr, unsigned char data[8])
{
  unsigned char Cmd[14];
  unsigned char Rsp[64 + 3 + 1];
  int r;

  /* Send the READ(address) command as a type B frame */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x03;                /* Send length = 1 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x01;
  Cmd[7] = 0xA8;                /* Timeout = 4ms = 424 ETU */

  Cmd[8] = 0x20 | addr;         /* CTS512 MULTI-READ command */

  Cmd[9] = 0xCC;
  Cmd[10] = 0xCC;               /* Room for CRC B (will be computed by reader) */
  Send(Cmd, 11);
  r = Recv(Rsp, sizeof(Rsp));

  if ((r < 13) || (Rsp[0] != 0x00))
    return S_NO_CARD;           /* No answer ? */

  if (data != NULL)
    memcpy(data, &Rsp[3], 8);

  return 0;                     /* OK ! */
}

/*
 * AskCts512SetBlock
 * -----------------
 * This is the CTS512B WRITE command
 * Writen data is returned to check...
 */
int AskCts512SetBlock(unsigned char addr, unsigned char mask[2])
{
  unsigned char Cmd[14];
  unsigned char Rsp[64 + 3 + 1];
  int r;

  if (mask == NULL)
    return E_INTERNAL;

  /* Send the WRITE(address) command as a type B frame together with data */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x05;                /* Send length = 3 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x01;
  Cmd[7] = 0x72;                /* Timeout = 2,5ms + 1ms = 370 ETU */

  Cmd[8] = 0xA0 | addr;         /* CTS512 WRITE command */

  Cmd[9]  = mask[0];
  Cmd[10] = mask[1];

  Cmd[11] = 0xCC;
  Cmd[12] = 0xCC;               /* Room for CRC B (will be computed by reader) */

  Send(Cmd, 13);
  r = Recv(Rsp, sizeof(Rsp));

  if ((r < 5) || (Rsp[0] != 0x00))
    return S_NO_CARD;           /* No answer ? */

  /* Don't check data, they can be different */

  return 0;                     /* OK ! */
}

/*
 * AskCts512WriteBlock
 * -------------------
 * This is the CTS512B UPDATE command
 * Writen data is returned to check...
 */
int AskCts512WriteBlock(unsigned char addr, unsigned char data[2])
{
  unsigned char Cmd[14];
  unsigned char Rsp[64 + 3 + 1];
  int r;

  if (data == NULL)
    return E_INTERNAL;

  /* Send the WRITE(address) command as a type B frame together with data */
  Cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B */
  Cmd[1] = 0x00;
  Cmd[2] = 0x05;                /* Send length = 3 + 2 */
  Cmd[3] = 0x00;
  Cmd[4] = 0x42;                /* Expected receive length <= 64 + 2 */
  Cmd[5] = 0x01;                /* Compute CRC B */
  Cmd[6] = 0x02;
  Cmd[7] = 0x80;                /* Timeout = 5ms + 1ms = 640 ETU */

  Cmd[8] = 0x60 | addr;         /* CTS512 UPDATE command */

  Cmd[9]  = data[0];
  Cmd[10] = data[1];

  Cmd[11] = 0xCC;
  Cmd[12] = 0xCC;               /* Room for CRC B (will be computed by reader) */

  Send(Cmd, 13);
  r = Recv(Rsp, sizeof(Rsp));

  if ((r < 5) || (Rsp[0] != 0x00))
    return S_NO_CARD;           /* No answer ? */

  /* Check response */
  if ((data[0] != Rsp[3]) || (data[1] != Rsp[4]))
    return S_CARD_ERROR;

  return 0;                     /* OK ! */
}

/*
 * AskCtsSelect
 * ------------
 * This select either a CTS256B or CTS512B
 * UID of selected card is returned in uid[4]
 * Type of card is returned in type[2] :
 * - (type[1] & 0xF0) == 0x50 ==> CTS256B
 * - (type[1] & 0xF0) == 0x60 ==> CTS512B
 */
int AskCtsSelect(unsigned char type[2], unsigned char uid[4])
{
  int r, i, t = 0;
  unsigned char l_type[2];
  unsigned char l_uid[4];
 
  /* READ(0) to see whether there is a card or not */
  r = AskCts256ReadBlock(0, l_type);
  if (r != 0) return r;

  /* OK, there's a card */
  if (type != NULL)
  {
    /* Words are swapped in card */
    type[0] = l_type[1];
    type[1] = l_type[0];
  }

  switch (l_type[0] & 0xF0)
  {
    case 0x50 : t = 256; break;
    case 0x60 : t = 512; break;
    default   : return S_NO_CARD; /* Not a supported type */
  }

  /* Select the card (CTS512B only) */
  if (t == 512)
  {
    r = AskCts512SelectAll();
    if (r != 0) return r;
  }


  /* Get serial number */
  for (i=0; i<2; i++)
  {
    /* Swap words */
    switch (t)
    {
      case 256 : r = AskCts256ReadBlock((unsigned char) (4 - i), &l_uid[2*i]); break;
      case 512 : r = AskCts512ReadBlock((unsigned char) (4 - i), &l_uid[2*i]); break;
    }
    if (r != 0) return r;
  }  


  if (uid != NULL)
  {
    /* Words are swapped in card */
    uid[0] = l_uid[1];
    uid[1] = l_uid[0];
    uid[2] = l_uid[3];
    uid[3] = l_uid[2];
  }

  return 0;                     /* OK */
}