/*
 * SpringProx or K531 SDK :: direct/mifare.c
 * ------------------------------------------
 *
 * Implementation of Mifare classic read & write functions
 *
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

/*
 * MifareProc
 * ----------
 *
 * The function waits for a Mifare classic card, then read it
 * sector by sector, and display the data block by block.
 *
 * We rely on reader already knowing the right key to access the card
 * (one of the keys stored in EEPROM), overwise we must pass the key
 * in parameter of MifareReadSector function.
 *
 * Writing into the card is not different, but be careful not to
 * write into a sector trailer (security block) without appropriate
 * formatting.
 *
 * IMPORTANT NOTICE:
 *     As the communication layers are written for data length < 128,
 *     the MifareReadSector and MifareWriteSector offered here can't
 *     access the last 16 sectors of a Mifare 4k card (240-byte long)
 */

int MifareProc(void)
{
  int e, i;
  unsigned char s, sc, b;

  unsigned char atq[2];  /* Answer To Query for 144443 A     */
  unsigned char uid[12]; /* Serial number (UID) for 144443 A */
  unsigned char sak[1];  /* Select AcKnowledge for 144443 A  */

  unsigned char data[48];
  
  /* Infinite loop, use CTRL+C to exit */
  printf("Waiting for a Mifare Classic card\n");
  printf("Put your card on the reader, or hit CTRL+C to exit\n");

  for (;;)
  {
    e = SelectTypeA(atq, uid, sak);
    if (e < 0) return e; /* Fatal error */
    if (e == 0)
    {
      /* Found a type A card */
      /* ------------------- */

      /* What kind of card ? (See NXP litterature for reference) */
      if ((atq[1] == 0x00) && (atq[0] == 0x04))
      {
        /* Mifare Classic 1k (or compliant ?) */
        printf("Found a Mifare Classic 1k\n");
        sc = 16; /* 16 sectors can be read */
        break;
      } else
      if ((atq[1] == 0x00) && (atq[0] == 0x02))
      {
        /* Mifare Classic 1k (or compliant ?) */
        printf("Found a Mifare Classic 4k\n");
        sc = 32; /* 32 sectors can be read (sectors 32 to 39 are too big) */
        break;
      } else
      {
        /* Mifare Classic 1k (or compliant ?) */
        printf("Found a card, but this is not a Mifare Classic (ATQ=%02X%02X)\n", atq[1], atq[0]);
      }

      /* We HALT the card to allow anticollision (next loop will return next card, or none) */
      e = Halt();
      if (e < 0) return e;
    }
  }

  printf("Reading %d sectors...\n", sc);

  printf("SECT. BLOCK DATA\n");

  for (s=0; s<sc; s++)
  {
    e = MifareReadSector(s, data, NULL);
    if (e < 0) return e; /* Fatal error */

    if (e == 0)
    {
      /* Success, let's show the data */
      for (b=0; b<3; b++)
      {
        /* Show block number */
        if (b == 0)
          printf(" [%02d] [%03d]  ", s, 4 * s + b);
        else
          printf("      [%03d]  ", 4 * s + b);

        /* Show data in HEX */
        for (i=0; i<16; i++)
          printf("%02X ", data[16*b + i]);

        printf("  ");

        /* Show plain data */
        for (i=0; i<16; i++)
          printf("%c", data[16*b + i] >= ' ' ? data[16*b + i] : '.');

        printf("\n");
      }      
    } else
    {
      /* Failure, show the error code */
      printf(" [%02d]  Error %d\n\t(%s)\n\n", s, e, ErrorMessage(e));

      if (e == S_NO_CARD) return e; /* Exit if card removed, this is a permanent error */
    }
  }

  /* Done ! */
  return 0;
}

/*
 * MifareReadSector
 * ----------------
 * Read a 48-byte Mifare Classic sector
 * If key is NULL, reader's internal keys (EEPROM or pre-loaded into RAM) are tried
 */
int MifareReadSector(unsigned char sector, unsigned char data[48], unsigned char key[6])
{
  unsigned char cmd[8];
  unsigned char rsp[1 + 48];    /* 1 for status, 48 for data */
  int r;

  /* Build the command */
  cmd[0] = 0x48;   /* SPROX_CSB_READ_SECTOR */
  cmd[1] = sector;  /* Sector number */
  
  /* Append key if necessary, and send the command */
  if (key != NULL)
  {
    memcpy(&cmd[2], key, 6);
    r = Send(cmd, 8);
  } else
    r = Send(cmd, 2);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 48+1) return E_FRAME_BAD_LEN;

    /* Retrieve data */
    if (data != NULL)
      memcpy(data, &rsp[1], 48);

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;
  if (rsp[0] == 4)
    return S_MIFARE_AUTH;

  return S_CARD_ERROR;
}

/*
 * MifareReadBlock
 * ---------------
 * Read a 16-byte Mifare Classic block
 * If key is NULL, reader's internal keys (EEPROM or pre-loaded into RAM) are tried
 */
int MifareReadBlock(unsigned char block, unsigned char data[16], unsigned char key[6])
{
  unsigned char cmd[8];
  unsigned char rsp[1 + 16];    /* 1 for status, 16 for data */
  int r;

  /* Build the command */
  cmd[0] = 0x49;   /* SPROX_CSB_READ_BLOC */
  cmd[1] = block;  /* Absolute block address */
  
  /* Append key if necessary, and send the command */
  if (key != NULL)
  {
    memcpy(&cmd[2], key, 6);
    r = Send(cmd, 8);
  } else
    r = Send(cmd, 2);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 16+1) return E_FRAME_BAD_LEN;

    /* Retrieve data */
    if (data != NULL)
      memcpy(data, &rsp[1], 16);

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;
  if (rsp[0] == 4)
    return S_MIFARE_AUTH;

  return S_CARD_ERROR;
}

/*
 * MifareWriteSector
 * -----------------
 * Write a 48-byte Mifare Classic sector
 * If key is NULL, reader's internal keys (EEPROM or pre-loaded into RAM) are tried
 */
int MifareWriteSector(unsigned char sector, unsigned char data[48], unsigned char key[6])
{
  unsigned char cmd[56];
  unsigned char rsp[1];
  int r;

  if (data == NULL) return E_INTERNAL;

  /* Build the command */
  cmd[0] = 0x4A;   /* SPROX_CSB_WRITE_SECTOR */
  cmd[1] = sector; /* Sector number */

  /* Append data */
  memcpy(&cmd[2], data, 48);
  
  /* Append key if necessary, and send the command */
  if (key != NULL)
  {
    memcpy(&cmd[50], key, 6);
    r = Send(cmd, 56);
  } else
    r = Send(cmd, 50);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
    return 0; /* OK */

  /* Error ? */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;
  if (rsp[0] == 4)
    return S_MIFARE_AUTH;

  return S_CARD_ERROR;
}

/*
 * MifareWriteBlock
 * ----------------
 * Write a 16-byte Mifare Classic block
 * If key is NULL, reader's internal keys (EEPROM or pre-loaded into RAM) are tried
 */
int MifareWriteBlock(unsigned char block, unsigned char data[16], unsigned char key[6])
{
  unsigned char cmd[24];
  unsigned char rsp[1];
  int r;

  if (data == NULL) return E_INTERNAL;

  /* Build the command */
  cmd[0] = 0x4B;   /* SPROX_CSB_WRITE_BLOC */
  cmd[1] = block;  /* Absolute block address */

  /* Append data */
  memcpy(&cmd[2], data, 16);
  
  /* Append key if necessary, and send the command */
  if (key != NULL)
  {
    memcpy(&cmd[18], key, 6);
    r = Send(cmd, 24);
  } else
    r = Send(cmd, 18);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
    return 0; /* OK */

  /* Error ? */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;
  if (rsp[0] == 4)
    return S_MIFARE_AUTH;

  return S_CARD_ERROR;
}

