/*
 * SpringProx or K531 SDK :: direct/polling.c
 * ------------------------------------------
 *
 * Implementation of ISO 14443 A & B polling (e.g. card discovery)
 * for targets not using the SpringProxAPI.
 *
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

/*
 * PollingProc
 * -----------
 *
 * The function lookups for any ISO 14443 A or B compliant card,
 * and displays its data (serial number + type related informations)
 */

int PollingProc(void)
{
  unsigned char a_ATQ[2];  /* Answer To Query for 144443 A     */
  unsigned char a_UID[12]; /* Serial number (UID) for 144443 A */
  unsigned char a_SAK[1];  /* Select AcKnowledge for 144443 A  */

  unsigned char b_ATQ[11]; /* Answer To Query for 144443 B     */
                           /* Serial number (PUPI) is the 4    */
                           /* first bytes of ATQ               */
  int e, i;

  /* Infinite loop, use CTRL+C to exit */
  printf("Entering polling loop, hit CTRL+C to exit\n");

  for (;;)
  {
    e = SelectTypeA(a_ATQ, a_UID, a_SAK);
    if (e < 0) break; /* Fatal error */
    if (e == 0)
    {
      /* Found a type A card */
      /* ------------------- */

      unsigned char a_UID_sz;

      /* Retrieve size of UID in ATQ */
      switch (a_ATQ[0] & 0xC0)
      {
        case 0x00 : /* Single UID (4 bytes) */
                    a_UID_sz = 4;
                    break;
        case 0x40 : /* Double UID (7 bytes) */
                    a_UID_sz = 7;
                    break;
        case 0x80 : /* Triple UID (12 bytes) */
                    a_UID_sz = 12;
                    break;

        case 0xC0 : /* RFU, forbidden value */
        default   :
                    a_UID_sz = 0;
      }

      /* Display card's data */
      printf("[A] ");

      /* ATQ shown in reverse order according to early MIFARE litterature */
      printf("%02X%02X ", a_ATQ[1], a_ATQ[0]);

      if (a_UID_sz == 4)
      {
        /* Single UID shown in reverse order according to early MIFARE litterature */
        for (i=3; i>=0; i--)
          printf("%02X", a_UID[i]);
      } else
      {
        /* ISO says that first byte is manufacturer code (0x04 for NXP), and must come first ! */
        for (i=0; i<a_UID_sz; i++)
          printf("%02X", a_UID[i]);
      }

      /* And finally Select AcKnowledge */
      printf(" %02X\n", a_SAK[0]);

      /* We HALT the card to allow anticollision (next loop will return next card, or none) */
      e = Halt();
      if (e < 0) return e;
    }


    e = SelectTypeB(b_ATQ);
    if (e < 0) break; /* Fatal error */
    if (e == 0)
    {
      /* Found a type B card */
      /* ------------------- */

      /* Display card's data */
      printf("[B]      ");

      /* PUPI (serial number or random number) on 4 bytes */
      for (i=0; i<4; i++)
        printf("%02X", b_ATQ[i]);

      printf(" ");

      /* Remaining part of ATQ */
      for (   ; i<11; i++)
        printf("%02X", b_ATQ[i]);

      printf("\n");

      /* We HALT the card to allow anticollision (next loop will return next card, or none) */
      e = Halt();
      if (e < 0) return e;

      /* NOTE : this is correct in theory, but in fact the type A frames reset most of the         */
      /*        the type B cards, so they will answer again next time. Strong type B anticollision */
      /*        can be done only is reader stops "speaking type A" and offers slotted queries.     */
    }

  }

  /* Oups, communication error */
  return e;
}


/*
 * SelectTypeA
 * -----------
 *
 * Performs ISO 14443-A Request/Anticoll/Select loop to select a single
 * card.
 * We use the "Request Idle" query (REQA and not WUPA) so only "new" cards
 * may answer. Cards previously halted will remain quiet, thus allowing
 * to see every card in the field.
 */
int SelectTypeA(unsigned char atq[2], unsigned char uid[12], unsigned char sak[1])
{
  unsigned char cmd[3];
  unsigned char rsp[32 + 1];
  int r;

  /* Enter ISO/IEC 14443-A mode */
  /* -------------------------- */
  cmd[0] = 0x58;                /* SPROX_CONTROL */
  cmd[1] = 0x0C;                /* SPROX_CONTROL_CONFIG */
  cmd[2] = 1;                   /* ISO/IEC 14443-A */  
  r = Send(cmd, 3);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  /* Request/Anticoll/Select sequence */
  /* -------------------------------- */
  cmd[0] = 0x4D;                /* SPROX_CSB_SELECT_IDLE */
  r = Send(cmd, 1);
  if (r < 0)  return r;
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 4+3+1) return E_FRAME_BAD_LEN;

    /* Retrieve data */
    if (atq != NULL)
      memcpy(atq, &rsp[r-3], 2);
    if (sak != NULL)
      memcpy(sak, &rsp[r-1], 1);

    /* Maybe an overflow on actual rsp data, don't care as caller will retrieve length from sak */
    if (uid != NULL)
      memcpy(uid, &rsp[1], 12);


    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] >= 26)
    return E_READER_FATAL;

  /* No card */
  return S_NO_CARD;
}

/*
 * SelectTypeB
 * -----------
 *
 * Performs ISO 14443-B Request to discover a single card.
 * We use the "Request Idle" query (REQB and not WUPB) so only "new" cards
 * may answer. Cards previously halted will remain quiet, thus allowing
 * to see every card in the field.
 * NOTE : this is correct in theory, but in fact the type A frames reset most of the
 *        the type B cards, so they will answer again next time. Strong type B anticollision
 *        can be done only is reader stops "speaking type A" and offers slotted queries.
 */
int SelectTypeB(unsigned char atq[11])
{
  unsigned char cmd[3];
  unsigned char rsp[32 + 1];
  int r;

  /* Enter ISO/IEC 14443-B mode */
  /* -------------------------- */
  cmd[0] = 0x58;                /* SPROX_CONTROL */
  cmd[1] = 0x0C;                /* SPROX_CONTROL_CONFIG */
  cmd[2] = 2;                   /* ISO/IEC 14443-B */  
  r = Send(cmd, 3);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  /* Request/Anticoll/Select sequence */
  /* -------------------------------- */
  cmd[0] = 0x4D;                /* SPROX_CSB_SELECT_IDLE */
  Send(cmd, 1);
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 11+1) return E_FRAME_BAD_LEN;

    /* Retrieve data (skip first byte of response that just says OK) */
    if (atq != NULL)
      memcpy(atq, &rsp[2], 11);

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] >= 26) return E_READER_FATAL;

  /* No card */
  return S_NO_CARD;
}

/*
 * Halt
 * ----
 *
 * Halt current card. Don't care for type A or type B, reader already knows
 * what has been selected last.
 */
int Halt(void)
{
  unsigned char cmd[1];
  int r;

  /* Enter ISO/IEC 14443-A mode */
  /* -------------------------- */
  cmd[0] = 0x92;                /* SPROX_HALT */
  r = Send(cmd, 1);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  /* No card */
  return 0;
}
