/*
 * SpringProx or K531 SDK :: direct/stsr176.c
 * ------------------------------------------
 *
 * (Partial) implementation of ST MicroElectronics SR176 RFID chips
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

/*
 * Configure the reader for ST SR176 (not really ISO 14443-B, but looks like)
 */
int StSr176Proc(void)
{
  unsigned char chip_id[1];
  unsigned char uid[8];
  unsigned char data[2];

  unsigned char cmd[3];
  unsigned char b;
  int r, e, i;

  /* Enter ISO/IEC 14443-B mode */
  /* -------------------------- */
  cmd[0] = 0x58;                /* SPROX_CONTROL */
  cmd[1] = 0x0C;                /* SPROX_CONTROL_CONFIG */
  cmd[2] = 2;                   /* ISO/IEC 14443-B */  
  r = Send(cmd, 3);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  /* Configure the MfRC531 chipset to accept ST SR176 specific frames */
  /* ---------------------------------------------------------------- */
  cmd[0] = 0xAC;
  cmd[1] = 0x17;
  cmd[2] = 0x00;
  r = Send(cmd, 3);
  if (r < 0)  return r;
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  /* Infinite loop, use CTRL+C to exit */
  printf("Waiting for a ST SR176 card\n");
  printf("Put your card on the reader, or hit CTRL+C to exit\n");

  for (;;)
  {
    e = StSr176Select(chip_id, uid);
    if (e < 0) return e; /* Fatal error */
    if (e == 0)
      break;
  }

  /* Say hello */
  printf("Found a card that looks like a ST SR176\n");
  printf("ChipID=%02X, UID=", chip_id[0]);
  for (i=0; i<8; i++)
    printf("%02X", uid[i]);
  printf("\n");

  /* And now read the card to display its whole content */
  /* -------------------------------------------------- */

  printf("Reading 16 blocks...\n");
  printf("BLOCK  DATA\n");

  for (b=0; b<16; b++)
  {
    e = StSr176ReadBlock(b, data);
    if (e < 0) return e; /* Fatal error */

    /* Show block number */
    printf("[%02d]   ", b);

    if (e == 0)
    {
      /* Show data in HEX */
      for (i=0; i<2; i++)
        printf("%02X ", data[i]);

      printf("  ");

      /* Show plain data */
      for (i=0; i<2; i++)
        printf("%c", data[i] >= ' ' ? data[i] : '.');

      printf("\n");

    } else
    {
      /* Failure, show the error code */
      printf("Error %d\n\t(%s)\n\n", e, ErrorMessage(e));
      if (e == S_NO_CARD) return e; /* Exit if card removed, this is a permanent error */
    }
  }

  /* Done ! */
  return 0;
}


/*
 * StSr176Select
 * -----------
 * Request to discover a single card (Initiate/Select).
 * Read 4 first  blocks to get UID.
 */
int StSr176Select(unsigned char chip_id[1], unsigned char uid[8])
{
  unsigned char b, l_chip_id, l_uid[8];
  unsigned char cmd[12];
  unsigned char rsp[12];
  int r;

  /* Initiate command */
  /* ---------------- */
  cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B        */
  cmd[1] = 0x00;
  cmd[2] = 4;                   /* Length of command, including CRC */
  cmd[3] = 0x00; cmd[4] = 1;    /* Expect 1 byte in response        */
  cmd[5] = 0x01;                /* Compute CRC                      */
  cmd[6] = 0x00; cmd[7] = 0x80; /* Timeout 128 ETU                  */
  
  cmd[8] = 0x06;                /* ) ST SR176 initiate command      */
  cmd[9] = 0x00;                /* )                                */

  cmd[10] = 0xCC; cmd[11] = 0xCC; /* Room for CRC */

  Send(cmd, 12);
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] != 0)
    goto error;

  if (r < 5) return E_FRAME_BAD_LEN;

  /* Answer must be 0003 xx cccc                     */
  /* 0003 is the length of the answer, including CRC */
  /* xx the actual answer (3-2 = 1 byte !)           */
  /* cccc is a dummy CRC (already checked by reader) */  
  if ((rsp[1] != 0x00) || (rsp[2] != 0x03))
    return S_BAD_RESP_CODE;

  /* Retrieve data */
  l_chip_id = rsp[3];
  if (chip_id != NULL)
    chip_id[0] = rsp[3];

  /* Select command */
  /* -------------- */
  cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B        */
  cmd[1] = 0x00;
  cmd[2] = 4;                   /* Length of command, including CRC */
  cmd[3] = 0x00; cmd[4] = 1;    /* Expect 1 byte in response        */
  cmd[5] = 0x01;                /* Compute CRC                      */
  cmd[6] = 0x01; cmd[7] = 0x00; /* Timeout 256 ETU                  */
  
  cmd[8] = 0x0E;                /* ) ST SR176 select command        */
  cmd[9] = l_chip_id;           /* ) (id of the 'initiated' chip)   */

  cmd[10] = 0xCC; cmd[11] = 0xCC; /* Room for CRC */

  Send(cmd, 12);
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] != 0)
    goto error;


  /* Now read the 4 first blocks (e.g. 8 bytes), this is card's serial number (UID) */
  /* ------------------------------------------------------------------------------ */ 
  for (b=0; b<4; b++)
  {
    r = StSr176ReadBlock(b, &l_uid[2*b]);
    if (r != 0) return r;
  }

  if (uid != NULL)
    memcpy(uid, l_uid, 8);

  /* OK */
  return 0;

error:
  /* Error ? */
  if (rsp[0] >= 26) return E_READER_FATAL;

  /* No card */
  return S_NO_CARD;
}

/*
 * StSr176ReadBlock
 * ----------------
 * Read a 2-byte block from SR176 card
 */
int StSr176ReadBlock(unsigned char addr, unsigned char data[2])
{
  unsigned char cmd[12];
  unsigned char rsp[12];
  int r;

  /* Initiate command */
  /* ---------------- */
  cmd[0] = 0x97;                /* SPROX_PICCEXCHANGEBLOCK_B        */
  cmd[1] = 0x00;
  cmd[2] = 4;                   /* Length of command, including CRC */
  cmd[3] = 0x00; cmd[4] = 2;    /* Expect 2 bytes in response       */
  cmd[5] = 0x01;                /* Compute CRC                      */
  cmd[6] = 0x01; cmd[7] = 0x00; /* Timeout 256 ETU                  */
  
  cmd[8] = 0x08;                /* ) ST SR176 read                  */
  cmd[9] = addr;                /* ) block number as parameter      */

  cmd[10] = 0xCC; cmd[11] = 0xCC; /* Room for CRC */

  Send(cmd, 12);
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 6) return E_FRAME_BAD_LEN;

    /* Answer must be 0004 xxxx cccc                   */
    /* 0004 is the length of the answer, including CRC */
    /* xxxx the data we were looking for               */
    /* cccc is a dummy CRC (already checked by reader) */  
    if ((rsp[1] != 0x00) || (rsp[2] != 0x04))
      return S_BAD_RESP_CODE;

    if (data != NULL)
      memcpy(data, &rsp[3], 2);

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] >= 26) return E_READER_FATAL;

  /* No card */
  return S_NO_CARD;
}
