/*
 * SpringProx or K531 SDK :: direct/tcl.c
 * --------------------------------------
 *
 * Minimal implementation of T=CL stuff, and demo using a few
 * ISO 7816-4 APDUs
 *
 * Our SRK ship with two different kind of T=CL cards :
 * - Desfire (v6 or newer), that is ISO/IEC 14443-A
 * - JayCOS, that is ISO/IEC 14443-A
 *
 * Desfire is partially compliant with 7816-4 (Desfire v5 and earlier
 * isn't). We'll try only the following sequence :
 * - ISO SELECT APPLICATION (DESFIRE APPLICATION)
 * - ISO SELECT DIRECTORY (DESFIRE ROOT DIRECTORY = 000000)
 * - PROPRIETARY SELECT DIRECTORY (DESFIRE ROOT DIRECTORY = 000000)
 * 
 * JayCOS is compliant with 7816-4, but cards are shipped "empty",
 * without file system. We'll try the following sequence :
 * - PROPRIETARY GET ATR
 *   works only for JayCOS as this is a non-ISO APDU
 * - PROPRIETARY GET CHIP NUMBER
 *   works only for JayCOS as this is a non-ISO APDU
 * - ISO SELECT DIRECTORY (ISO ROOT DIRECTORY = 3F00)
 *   this one will fail if card is empty
 * 
 */

/*

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ref_direct.h"

/* List of APDUs */

static const unsigned char DesfireSelectAppAPDU[] =
  { 0x00, 0xA4, 0x04, 0x00, 0x07, 0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x00 };
static const unsigned char DesfireSelectMFAPDU[] =
  { 0x00, 0xA4, 0x04, 0x00, 0x03, 0x00, 0x00, 0x00 };
static const unsigned char DesfireGetVersion[] =
  { 0x90, 0x60, 0x00, 0x00, 0x00 };
static const unsigned char DesfireContinue[] =
  { 0x90, 0xAF, 0x00, 0x00, 0x00 };

static const unsigned char JaycosGetAtrAPDU[] =
  { 0x80, 0xEC, 0x00, 0x00, 0x0C };
static const unsigned char JaycosGetChipNumberAPDU[] =
  { 0xB0, 0xEE, 0x00, 0x00, 0x08 };
static const unsigned char JaycosSelectMFAPDU[] =
  { 0x00, 0xA4, 0x00, 0x00, 0x02, 0x3F, 0x00 };



unsigned short StatusWord(unsigned char iso_out[], unsigned char iso_out_length);
void ExplainStatusWord(unsigned short sw);

/*
 * TclProc
 * -------
 *
 * The function lookups for any ISO 14443 A or B T=CL card (compliant
 * up to level 4), then tries the test APDUs.
 *
 * Note that once connected to the card, knowing the actual type is 
 * not an issue anymore : TclTransmit works transparently for both
 */
int TclProc(void)
{
  unsigned char a_ATQ[2];  /* Answer To Query for 144443 A     */
  unsigned char a_UID[12]; /* Serial number (UID) for 144443 A */
  unsigned char a_SAK[1];  /* Select AcKnowledge for 144443 A  */

  unsigned char b_ATQ[11]; /* Answer To Query for 144443 B     */
                           /* Serial number (PUPI) is the 4    */
                           /* first bytes of ATQ               */

  unsigned char data[128];
  unsigned char data_length;
  unsigned short sw;
  int e, i;

  /* Infinite loop, use CTRL+C to exit */
  printf("Waiting for a T=CL card\n");
  printf("Put your card on the reader, or hit CTRL+C to exit\n");

  for (;;)
  {
    e = SelectTypeA(a_ATQ, a_UID, a_SAK);
    if (e < 0) return e; /* Fatal error */
    if (e == 0)
    {
      /* Found a type A card */
      /* ------------------- */

      /* Check that the card is T=CL compliant */
      if (a_SAK[0] & 0x20)
      {
        /* Yes ! Let's activate it */
        e = TclConnectTypeA();
        if (e < 0) return e; /* Fatal error */
        if (e == 0) break; /* Success */
      } else
      {
        /* Not compliant */
        printf("Found a type A card, but not T=CL compliant\n");
        /* We HALT the card to allow anticollision (next loop will return next card, or none) */
        e = Halt();
        if (e < 0) return e;
      }
    }

    e = SelectTypeB(b_ATQ);
    if (e < 0) return e; /* Fatal error */
    if (e == 0)
    {
      /* Found a type B card */
      /* ------------------- */

      /* Check that the card is T=CL compliant */
      if (b_ATQ[9] & 0x01)
      {
        /* Yes ! Let's activate it */
        e = TclConnectTypeB();
        if (e < 0) return e; /* Fatal error */
        if (e == 0) break; /* Success */
      } else
      {
        /* Not compliant */
        printf("Found a type B card, but not T=CL compliant\n");
        /* We HALT the card to allow anticollision (next loop will return next card, or none) */
        e = Halt();
        if (e < 0) return e;


        /* NOTE : this is correct in theory, but in fact the type A frames reset most of the         */
        /*        the type B cards, so they will answer again next time. Strong type B anticollision */
        /*        can be done only is reader stops "speaking type A" and offers slotted queries.     */
      }
    }
  }

  printf("T=CL card activated, now trying our test APDUs...\n");

  /* Try to select the Desfire application. Only Desfire version >=6 will return OK */
  /* ------------------------------------------------------------------------------ */

  data_length = sizeof(data);
  e = TclTransmit(DesfireSelectAppAPDU, sizeof(DesfireSelectAppAPDU), data, &data_length);
  if (e != 0) return e; /* Error ! */

  if ((data_length == 1) && (data[0] = 0xAC))
  {
    /* Older Desfire are not ISO... */
    /* ---------------------------- */
    printf("Card answered only 'AC' to our ISO APDU\n");
    printf("I think this is a pre-ISO Desfire, can't go further\n");
    return 0;
  }
  if (data_length < 2) return S_BAD_RESP_SIZE; /* Card didn't answer at least an ISO status word */
  sw = StatusWord(data, data_length); /* Get status word */

  if (sw == 0x9000)
  {
    /* OK, Desfire application selected in the card, this is a Desfire ! */
    /* ----------------------------------------------------------------- */
    printf("Card holds the Desfire application !\n");

    /* Select root application */ 
    data_length = sizeof(data);
    e = TclTransmit(DesfireSelectMFAPDU, sizeof(DesfireSelectMFAPDU), data, &data_length);
    if (e != 0) return e; /* Error ! */
    if (data_length < 2) return S_BAD_RESP_SIZE; /* Card didn't answer at least an ISO status word */
    sw = StatusWord(data, data_length); /* Get status word */

    if (sw == 0x9000)
    {
      printf("OK, selected Desfire root Directory (AID = 000000)\n");
    } else
    {
      printf("Error selecting Desfire root Directory\n");
      ExplainStatusWord(sw);
    }

    /* Now the GetVersion command .This isn't ISO, and more than that we must implement Desfire */
    /* specific chaining : send "continue" as long as SW says "more data to follow"             */
    data_length = sizeof(data);
    e = TclTransmit(DesfireGetVersion, sizeof(DesfireGetVersion), data, &data_length);
    if (e != 0) return e; /* Error ! */
    if (data_length < 2) return S_BAD_RESP_SIZE; /* Card didn't answer at least an ISO status word */
    sw = StatusWord(data, data_length); /* Get status word */

    while (sw == 0x91AF)
    {
      /* Desfire "continue" : we must loop while card says that there's something left */
      data_length = sizeof(data);
      e = TclTransmit(DesfireContinue, sizeof(DesfireContinue), data, &data_length);
      if (e != 0) return e; /* Error ! */
      if (data_length < 2) return S_BAD_RESP_SIZE; /* Card didn't answer at least an ISO status word */
      sw = StatusWord(data, data_length); /* Get status word */
    }

    /* Last but not least, Desfire status "OK" is not 9000 but 9100 for proprietary commands */
    if (sw == 0x9100)
    {
      printf("OK, we've receive the complete Desfire version buffer\n");
    } else
    {
      printf("Error retrieving Desfire version\n");
      ExplainStatusWord(sw);
    }

    /* Done with Desfire card, leave here */
    return 0;
  }

  /* Proprietary Get ATR command. JayCOS card will answer correctly (but other cards may also do) */
  /* -------------------------------------------------------------------------------------------- */

  data_length = sizeof(data);
  e = TclTransmit(JaycosGetAtrAPDU, sizeof(JaycosGetAtrAPDU), data, &data_length);
  if (e != 0) return e; /* Error ! */
  if (data_length < 2) return S_BAD_RESP_SIZE; /* Card didn't answer at least an ISO status word */
  sw = StatusWord(data, data_length); /* Get status word */

  if (sw == 0x9000)
  {
    /* OK */
    printf("Card answered positively to Get ATR !\n");
    printf("ATR  : ");
    for (i=0; i<data_length-2; i++)
      printf("%02X", data[i]);
    printf("\n");
  } else
  {
    printf("Get ATR failed error (card is not a JayCOS ?)\n");
    ExplainStatusWord(sw);
  }

  /* Proprietary Get Chip number command. Same as above */
  /* -------------------------------------------------- */

  data_length = sizeof(data);
  e = TclTransmit(JaycosGetChipNumberAPDU, sizeof(JaycosGetChipNumberAPDU), data, &data_length);
  if (e != 0) return e; /* Error ! */
  if (data_length < 2) return S_BAD_RESP_SIZE; /* Card didn't answer at least an ISO status word */
  sw = StatusWord(data, data_length); /* Get status word */

  if (sw == 0x9000)
  {
    /* OK */
    printf("Card answered positively to Get Chip Number !\n");
    printf("nbr.: ");
    for (i=0; i<data_length-2; i++)
      printf("%02X", data[i]);
    printf("\n");
  } else
  {
    printf("Get Chip Number failed error (card is not a JayCOS ?)\n");
    ExplainStatusWord(sw);
  }

  /* ISO Select Master File command. Any card will answer correctly. We focus on explaining the SW */
  /* --------------------------------------------------------------------------------------------- */

  data_length = sizeof(data);
  e = TclTransmit(JaycosSelectMFAPDU, sizeof(JaycosSelectMFAPDU), data, &data_length);
  if (e != 0) return e; /* Error ! */
  if (data_length < 2) return S_BAD_RESP_SIZE; /* Card didn't answer at least an ISO status word */
  sw = StatusWord(data, data_length); /* Get status word */

  if (sw == 0x9000)
  {
    /* OK */
    printf("Select Master File (3F00) success !\n");
    if (data_length > 2)
    {
      /* Card returned data (must be FCI of the MF) */
      printf("info: ");
      for (i=0; i<data_length-2; i++)
        printf("%02X", data[i]);
      printf("\n");
    }
  } else
  {
    printf("Select Master File (3F00) error !\n");
    ExplainStatusWord(sw);
  }

  /* Disconnect from the card */
  TclDisconnect();

  /* Done ! */
  return 0;
}

/*
 * TclConnectTypeA
 * ---------------
 *
 * Invoke "Get ATS" function to put type A card in level 4
 *
 * Note that in this implementation we don't care for returned ATS and
 * just drop it. In some application it may be interesting to see in the
 * ATS if card supports higher baudrates (then perform a PPS exchange).
 * Also some card applications (cardlets) expose themselves in the ATS
 * (historical bytes) so host application may recognize them quickly.
 */
int TclConnectTypeA(void)
{
  unsigned char cmd[3];
  unsigned char rsp[1 + 32];    /* 1 for status, up to 32 for ATS */
  int r;

  /* Build the command */
  cmd[0] = 0x81;    /* SPROX_TCL_FUNC */
  cmd[1] = 0x10;    /* SPROX_TCL_FUNC_RATS */
  cmd[2] = 0xFF;    /* Not using CID */
  
  /* Send the command */
  r = Send(cmd, 3);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 2) return E_FRAME_BAD_LEN;

#ifdef WITH_DEBUG_CODE
    if (TraceCard)
    {
      int i;
      printf("ATS: "); /* (Answer To Select) */
      for (i=1; i<r; i++)
        printf("%02X", rsp[i]);
      printf("\n");
    }
#endif

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] == 82)
    return S_NO_CARD; /* Card removed */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;

  return S_CARD_ERROR;
}

/*
 * TclConnectTypeB
 * ---------------
 *
 * Invoke "Attrib" function to put type B card in level 4
 *
 * Note that in this implementation we don't care for INFormation bytes
 * that may be returned by the card, and just drop them. Some card
 * applications (cardlets) are said to expose themselves in this field
 * (more or less like historical bytes in 7816 ATR or 14443-A ATS).
 */
int TclConnectTypeB(void)
{
  unsigned char cmd[3];
  unsigned char rsp[1 + 32];    /* 1 for status, up to 32 for higher layer INF */
  int r;

  /* Build the command */
  cmd[0] = 0x81;    /* SPROX_TCL_FUNC */
  cmd[1] = 0x20;    /* SPROX_TCL_FUNC_ATTRIB */
  cmd[2] = 0xFF;    /* Not using CID */
  
  /* Send the command */
  r = Send(cmd, 3);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0)  return r;

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (r < 1) return E_FRAME_BAD_LEN;

#ifdef WITH_DEBUG_CODE
    if (TraceCard)
    {
      int i;
      printf("ATA: "); /* (Answer To Attrib) */
      for (i=1; i<r; i++)
        printf("%02X", rsp[i]);
      printf("\n");
    }
#endif

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] == 82)
    return S_NO_CARD; /* Card removed */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;

  return S_CARD_ERROR;
}

/*
 * TclDisconnect
 * -------------
 * Tell card we're done - It can go to HALTed mode
 */
int TclDisconnect(void)
{
  unsigned char cmd[3];
  int r;

  /* Build the command */
  cmd[0] = 0x81;    /* SPROX_TCL_FUNC */
  cmd[1] = 0x12;    /* SPROX_TCL_FUNC_DESELECT */
  cmd[2] = 0xFF;    /* Not using CID */
  
  /* Send the command */
  r = Send(cmd, 3);
  if (r < 0) return r;

  /* Don't care for the answer */
  r = Recv(NULL, 0);
  if (r < 0)  return r;

  return 0;
}


int TclTransmit(const unsigned char iso_in[], unsigned char iso_in_len, unsigned char iso_out[], unsigned char *iso_out_len)
{
  unsigned char cmd[2 + 128];
  unsigned char rsp[1 + 128];    /* 1 for status, up to 128 for higher layer INF */
  int r;

#ifdef WITH_DEBUG_CODE
  if (TraceCard)
  {
    int i;
    printf("PICC< ");
    for (i=0; i<iso_in_len; i++)
      printf("%02X", iso_in[i]);
    printf("\n");
  }
#endif

  /* Build the command */
  cmd[0] = 0x82;    /* SPROX_TCL_EXCHANGE */
  cmd[1] = 0xFF;    /* Not using CID */

  memcpy(&cmd[2], iso_in, iso_in_len);
  
  /* Send the command */
  r = Send(cmd, 2+iso_in_len);
  if (r < 0) return r;

  /* Get the answer */
  r = Recv(rsp, sizeof(rsp));
  if (r < 0) return r;
  if (r < 1) return E_FRAME_BAD_LEN;

#ifdef WITH_DEBUG_CODE
  if (TraceCard)
  {
    int i;
    printf("PICC> ");
    for (i=1; i<r; i++)
      printf("%02X", rsp[i]);
    printf(" %d\n", rsp[0]);
  }
#endif

  /* Process answer */
  if (rsp[0] == 0)
  {
    if (iso_out == NULL) return 0;      /* Don't care for answer here */
    if (iso_out_len == NULL) return 0; /* Length not provided, can go further */

    /* Check that data will fit in buffer */
    if (*iso_out_len < r-1) return E_OVERFLOW; /* Overflow ! */

    /* Copy the data */
    memcpy(iso_out, &rsp[1], r-1);

    /* Update the length */
    *iso_out_len = r-1;

    /* OK */
    return 0;
  }

  /* Error ? */
  if (rsp[0] == 82)
    return S_NO_CARD; /* Card removed */
  if (rsp[0] >= 26)
    return E_READER_FATAL;
  if (rsp[0] == 1)
    return S_NO_CARD;

  return S_CARD_ERROR;
}

unsigned short StatusWord(unsigned char iso_out[], unsigned char iso_out_length)
{
  unsigned short sw;

  if (iso_out == NULL) return 0xFFFF;
  if (iso_out_length < 2) return 0xFFFF;

  sw  = iso_out[iso_out_length-2]; sw <<= 8;
  sw += iso_out[iso_out_length-1];

  return sw;
}

void ExplainStatusWord(unsigned short sw)
{
  unsigned char sw1 = (unsigned char) (sw / 0x0100);
  unsigned char sw2 = (unsigned char) (sw % 0x0100);

  if (sw1 == 0x90 && sw2 == 0x00)
  {
    printf("Normal processing\n");
    return;
  }

  if (sw1 == 0x61)
  {
    printf("%d bytes of response still available.\n", sw2);
    return;
  }

  if (sw1 == 0x62)
  {
    printf("State of non-volatile memory unchanged ");

    if (sw2 == 0x00)
      printf("No informatin given.");
    if (sw2 == 0x81)
      printf("Part of returned data my be corrupted.");
    if (sw2 == 0x82)
      printf("End of file/record reached before reading Le bytes.");
    if (sw2 == 0x83)
      printf("Selected file invalidated.");
    if (sw2 == 0x84)
      printf("FCI not formated according to 5.1.5.");

    printf("\n");
    return;
  }

  if (sw1 == 0x63)
  {
    printf("State of non-volatile memory changed. ");
    if (sw2 == 0x00)
      printf("No information given.");
    if (sw2 == 0x81)
      printf("File filled up by the last write.");
    if ((sw2 & 0xF0) == 0xC0)
      printf("Counter %d.", sw2 & 0x0F);

    printf("\n");
    return;
  }

  if (sw1 == 0x64 && sw2 == 0x00)
  {
    printf("State of non-volatile memory unchanged.\n");
    return;
  }

  if (sw1 == 0x65)
  {
    printf("State of non-volatile memory changed. ");
    if (sw2 == 0x81)
      printf("Memory failure.");
    printf("\n");
    return;
  }

  if (sw1 == 0x66)
  {
    printf("Reserved for security-related issues.\n");
    return;
  }

  if (sw1 == 0x67)
  {
    printf("Wrong length.\n");
    return;
  }

  if (sw1 == 0x68)
  {
    printf("Functions in CLA not supported. ");
    if (sw2 == 0x81)
      printf("Logical channel not supported.");
    if (sw2 == 0x82)
      printf("Secure messaging not supported.");
    printf("\n");
    return;
  }

  if (sw1 == 0x69)
  {
    printf("Command not allowed. ");
    if (sw2 == 0x81)
      printf("Command incompatible with file structure.");
    if (sw2 == 0x82)
      printf("Security status not satisfied.");
    if (sw2 == 0x83)
      printf("Authentication methon blocked.");
    if (sw2 == 0x84)
      printf("Referenced data invalidated.");
    if (sw2 == 0x85)
      printf("Conditions of use not satisfied.");
    if (sw2 == 0x86)
      printf("Command not allowed (no current EF).");
    if (sw2 == 0x87)
      printf("Expected SM data objects missing.");
    if (sw2 == 0x88)
      printf("SM dta objects incorrect.");
    printf("\n");
    return;
  }

  if (sw1 == 0x6A)
  {
    printf("Wrong parameter(s) P1-P2. ");
    if (sw2 == 0x80)
      printf("Incorrect parameters in the data field.");
    if (sw2 == 0x81)
      printf("Function not supported.");
    if (sw2 == 0x82)
      printf("File not found.");
    if (sw2 == 0x83)
      printf("Record not found.");
    if (sw2 == 0x84)
      printf("Not enough memory space in the file.");
    if (sw2 == 0x85)
      printf("Lc inconsistent with TLV structure.");
    if (sw2 == 0x86)
      printf("Incorrect parameters P1-P2.");
    if (sw2 == 0x87)
      printf("Lc inconsistent with P1-P2.");
    if (sw2 == 0x88)
      printf("Referenced data not found.");
    printf("\n");
    return;
  }

  if (sw1 == 0x6B)
  {
    printf("Wrong parameter(s) P1-P2.\n");
    return;
  }

  if (sw1 == 0x6C)
  {
    printf("Wrong length Le: should be %d\n", sw2);
    return;
  }

  if (sw1 == 0x6D)
  {
    printf("Instruction code not supported or invalid.\n");
    return;
  }

  if (sw1 == 0x6E)
  {
    printf("Class not supported.\n");
    return;
  }

  if (sw1 == 0x6F)
  {
    printf("No precise diagnosis.\n");
    return;
  }

  /* Default error message */
  printf("Not defined by 7816 (sw1=%02X, sw2=%02X).\n", sw1, sw2);
  return;
}
