#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "p2p_common.h"


static const BYTE NDEF_HELLO[] = 
{
  0xD1,       /* NDEF magic value */
  0x01,
  0x13,
  0x54,
    0x02,     /* Size of lang */
    'e', 'n', /* Lang is "en" */
  'H', 'e', 'l', 'l', 'o', ',', ' ', 'N', 'F', 'C', ' ', 'p', 'e', 'e', 'r', '!'
};

static const BYTE NDEF_THANKS[] = 
{
  0xD1,       /* NDEF magic value */
  0x01,
  0x14,
  0x54,
    0x02,     /* Size of lang */
    'e', 'n', /* Lang is "en" */
  'T', 'h', 'a', 'n', 'k', 's', ',', ' ', 'N', 'F', 'C', ' ', 'p', 'e', 'e', 'r', '!'
};

const BYTE NFC_ATR_MAGIC_NUMBER[3] = { 0x46, 0x66, 0x6D };
const BYTE NFC_ATR_GI_PARAMS[10]   = { 0x01, 0x01, 0x10, 0x03, 0x02, 0x00, 0x13, 0x04, 0x01, 0x96 };

static const char *COM_ANDROID_NPP = "com.android.npp";
static const char *URN_NFC_SN_SNEP = "urn:nfc:sn:snep";

#define ATR_PPI_HAS_GI    0x02
#define ATR_PPI_LRI_DEF   0x30 /* Compliant with NFC-Forum: LRI = 11b */


typedef struct _llcp_frame_stack_st
{
  LLCP_FRAME_ST frame;
  struct _llcp_frame_stack_st *next;

} LLCP_FRAME_STACK_ST;

LLCP_FRAME_STACK_ST *frame_stack = NULL;


/* SNEP server global variables */
static BYTE  *server_long_payload = NULL;
static DWORD  server_long_payload_length;
static DWORD  server_long_payload_offset;

/*
 * LCCP_SenderEnqueue
 * ------------------
 * Put a frame in sender's queue
 */
BOOL LLCP_SenderEnqueue(LLCP_FRAME_ST *frame)
{
  LLCP_FRAME_STACK_ST *p;

  if (frame == NULL)
    return FALSE;

  if (frame_stack == NULL)
  {
     /* Queue empty, let's create it */
    frame_stack = (LLCP_FRAME_STACK_ST *) malloc(sizeof(LLCP_FRAME_STACK_ST));

    if (frame_stack == NULL)
    {
      printf("Out of memory\n");
      exit(EXIT_FAILURE);
      return FALSE;
    }

    p = frame_stack;

  } else
  {
    /* Queue already existing */
    LLCP_FRAME_STACK_ST *p = frame_stack;

    /* Go to last frame */
    while (p->next != NULL)
      p = p->next;

    /* Create a new one */
    p->next = (LLCP_FRAME_STACK_ST *) malloc(sizeof(LLCP_FRAME_STACK_ST));

    if (p->next == NULL)
    {
      printf("Out of memory\n");
      exit(EXIT_FAILURE);
      return FALSE;
    }

    p = p->next;

  }

  /* Store the data into the last position in the queue */
  p->next = NULL;
  memcpy(&p->frame, frame, sizeof(LLCP_FRAME_ST));

  return TRUE;
}

/*
 * LCCP_SenderDequeue
 * ------------------
 * Get next frame from sender's queue
 */
BOOL LLCP_SenderDequeue(LLCP_FRAME_ST *frame)
{
  LLCP_FRAME_STACK_ST *p;

  if (frame == NULL)
    return FALSE;

  if (frame_stack == NULL)
  {
    /* Queue empty */
    memset(frame, 0, sizeof(LLCP_FRAME_ST));
    return TRUE;
  }

  /* Retrieve current frame */
  memcpy(frame, &frame_stack->frame, sizeof(LLCP_FRAME_ST));

  /* And let's remove it from the queue */
  p = frame_stack->next;
  free(frame_stack);
  frame_stack = p;

  return TRUE;
}

/*
 * LLCP_SerializeFrame
 * -------------------
 * Serialize a frame
 */
BOOL LLCP_SerializeFrame(BYTE buffer[], DWORD *length, const LLCP_FRAME_ST *frame)
{
  if ((frame == NULL) || (buffer == NULL) || (length == NULL))
    return FALSE;

  buffer[0] = ((frame->dsap << 2) & 0xFC) | ((frame->ptype >> 2) & 0x03);
  buffer[1] = ( frame->ssap       & 0x3F) | ((frame->ptype << 6) & 0xC0);

  if (frame->payload_length)
    memcpy(&buffer[2], frame->payload, frame->payload_length);

  *length = frame->payload_length + 2;

  return TRUE;
}

/*
 * LLCP_UnserializeFrame
 * ---------------------
 * Un-serialize a frame
 */
BOOL LLCP_UnserializeFrame(LLCP_FRAME_ST *frame, const BYTE buffer[], DWORD length)
{
  if ((frame == NULL) || (buffer == NULL) || (length < 2))
    return FALSE;

  frame->dsap  = (buffer[0] >> 2) & 0x3F;
  frame->ssap  =  buffer[1]       & 0x3F;
  frame->ptype = ((buffer[0] << 2) & 0x0C) | ((buffer[1] >> 6) & 0x03);

  frame->payload_length = length - 2;
  memcpy(frame->payload, &buffer[2], frame->payload_length);

  return TRUE;
}


/*
 * LLCP_Exchange
 * -------------
 * Exchange frames in NFC-DEP
 */
BOOL LLCP_Exchange(const LLCP_FRAME_ST *my_frame, LLCP_FRAME_ST *his_frame)
{
  BYTE buffer[258];
  DWORD send_len, recv_len;

  send_len = sizeof(buffer);
  if (!LLCP_SerializeFrame(buffer, &send_len, my_frame))
  {
    printf("LLCP_SerializeFrame failed\n");
    return FALSE;
  }

  recv_len = sizeof(buffer);
  if (!LLCP_ExchangeEx(buffer, send_len, buffer, &recv_len))
  {
    printf("LLCP_ExchangeEx failed\n");
    return FALSE;
  }

  if (!LLCP_UnserializeFrame(his_frame, buffer, recv_len))
  {
    printf("LLCP_UnserializeFrame failed\n");
    return FALSE;
  }

  return TRUE;
}

/*
 * LLCP_Create_PutNdef_Payload
 * ---------------------------
 * Create a SNEP PUT request
 */
BOOL LLCP_Create_PutNdef_Payload(BYTE message[], DWORD *length, const BYTE ndef[], DWORD ndef_size)
{
  message[0] = SNEP_VERSION; /* Version */
  message[1] = SNEP_REQ_PUT;
  message[2] = (BYTE) (ndef_size >> 24);
  message[3] = (BYTE) (ndef_size >> 16);
  message[4] = (BYTE) (ndef_size >> 8);
  message[5] = (BYTE) (ndef_size);

  memcpy(&message[6], ndef, ndef_size);

  *length = 6 + ndef_size;

  return TRUE;
}


/*
 * LLCP_SnepServer_Request
 * ------------------------
 * SNEP server core
 */
BOOL LLCP_SnepServer_Request(const BYTE his_message[], DWORD his_message_length, BYTE my_message[], DWORD *my_message_length)
{

  BYTE  his_request;
  BYTE  my_response = SNEP_RSP_REJECT;
  DWORD his_length;
  DWORD my_length = 0;
  const BYTE *his_data = &his_message[6];
  BYTE *my_data  = &my_message[6];

  *my_message_length = 0;

  if (server_long_payload != NULL)
  {
    printf("Receiving... %06ld / %06ld\n", server_long_payload_offset, server_long_payload_length);

    if ((his_message_length + server_long_payload_offset) > server_long_payload_length)
    {
      printf("Oups overflow\n"); // JDA TODO

    }

    memcpy(&server_long_payload[server_long_payload_offset], his_data, his_message_length-6);
    server_long_payload_offset += his_message_length;

    if (server_long_payload_offset >= server_long_payload_length)
    {
      /* This was the last chunk, let's process it as a complete message */
      his_request = SNEP_REQ_PUT;
      his_data    = server_long_payload;
      his_length  = server_long_payload_length;
    } else
    {
      my_response = SNEP_RSP_CONTINUE;
      goto send_response;
    }
  } else
  {
    his_request  = his_message[1];
    his_length   = his_message[2];
    his_length <<= 8;
    his_length  |= his_message[3];
    his_length <<= 8;
    his_length  |= his_message[4];
    his_length <<= 8;
    his_length  |= his_message[5];

  //  printf("SNEP> %02X %02X %ld\n", version, his_request, his_length);

    if ((his_length + 6) < his_message_length)
    {
      printf("SNEP message is too long\n");
      return FALSE;

    } else
    if ((his_length + 6) > his_message_length)
    {
      /* The total payload is longer than the one in this frame */
      server_long_payload_length = his_length;
      server_long_payload_offset = 0;

      server_long_payload = malloc(server_long_payload_length);
      if (server_long_payload == NULL)
      {
        printf("Out of memory\n");
        exit(EXIT_FAILURE);
        return FALSE;
      }

      memcpy(&server_long_payload[server_long_payload_offset], his_data, his_message_length-6);
      server_long_payload_offset += his_message_length-6;

      my_response = SNEP_RSP_CONTINUE;
      goto send_response;
    }
  }

  switch (his_request)
  {
    case SNEP_REQ_CONTINUE :
      /* Send remaining fragment */
      my_response = SNEP_RSP_REJECT;
      break;

    case SNEP_REQ_GET :
      /* Return an NDEF message */
      my_response = SNEP_RSP_NOT_IMPLEMENTED;
      break;

    case SNEP_REQ_PUT :
      /* Accept an NDEF message */
      printf("SNEP server: got a message from client\n\n");

      {
        DWORD i, j;

        for (i=0; i<his_length; i+=16)
        {
          printf("    ");
          for (j=0; j<16; j++)
          {
            if ((i+j) < his_length)
            {
              printf("%02X ", his_data[i+j]);
            } else
            {
              printf(".. ");
            }
          }
          printf("    ");
          for (j=0; j<16; j++)
          {
            if ((i+j) < his_length)
            {
              if ((his_data[i+j] >= ' ') && (his_data[i+j] <= 0x7F))
              {
                printf("%c", his_data[i+j]);
              } else
              {
                printf(".");
              }
            } else
            {
              printf(" ");
            }
          }
          printf("\n");
        }
      }
      printf("\n");

      if (server_long_payload != NULL)
      {
        free(server_long_payload);
        server_long_payload = NULL;
      }

      my_response = SNEP_RSP_SUCCESS;
      break;

    case SNEP_REQ_REJECT :
      /* Do not send remaining fragments */
      my_response = SNEP_RSP_REJECT;
      break;

    default :
      /* Unsupported request */
      my_response = SNEP_RSP_BAD_REQUEST;
  }


  /* Format the response */
send_response:
  my_message[0] = SNEP_VERSION;
  my_message[1] = my_response;
  my_message[2] = (BYTE) (my_length >> 24);
  my_message[3] = (BYTE) (my_length >> 16);
  my_message[4] = (BYTE) (my_length >> 8);
  my_message[5] = (BYTE) (my_length);

  *my_message_length = my_length + 6;

  /* OK */
  return TRUE;
}

/*
 * LLCP_SnepServer_Handle
 * ----------------------
 * Frame handler (and response provider) for SNEP server
 */
BOOL LLCP_SnepServer_Handle(const LLCP_FRAME_ST *his_frame, LLCP_FRAME_ST *my_frame)
{
  BYTE his_sequence;
  BYTE my_sequence;


  switch (his_frame->ptype)
  {
    case LLCP_PTYPE_CONNECT :
      /* The client wants to connect our SNEP server - anwser with a Connect Confirm */
      printf("SNEP server: got a new client\n");

      if (server_long_payload != NULL)
      {
        free(server_long_payload);
        server_long_payload = NULL;
      }
      server_long_payload_length = 0;
      server_long_payload_offset = 0;

      my_frame->ssap  = his_frame->dsap;
      my_frame->dsap  = his_frame->ssap;
      my_frame->ptype = LLCP_PTYPE_CC;
      return TRUE;

    case LLCP_PTYPE_I :
      /* The client is sending us something */

      /* Retrieve the sequence number */
      his_sequence  = (his_frame->payload[0] >> 4) & 0x0F;
      his_sequence++;
      his_sequence &= 0x0F;
      my_sequence   = (his_frame->payload[0] & 0x0F);

      if (LLCP_SnepServer_Request(&his_frame->payload[1], his_frame->payload_length-1, &my_frame->payload[1], &my_frame->payload_length))
      {
        /* Message handled */
        if (my_frame->payload_length)
        {
          /* Send an information in response */
          my_frame->ssap  = his_frame->dsap;
          my_frame->dsap  = his_frame->ssap;
          my_frame->ptype = LLCP_PTYPE_I;
          my_frame->payload[0] = ((my_sequence << 4) & 0xF0) | (his_sequence & 0x0F);
          my_frame->payload_length += 1;

          printf("<<<seq %02X\n", my_frame->payload[0]);
        } else
        {
          /* Send only a receive ready frame */
          my_frame->ssap  = his_frame->dsap;
          my_frame->dsap  = his_frame->ssap;
          my_frame->ptype = LLCP_PTYPE_RR;
          my_frame->payload[0] = (his_sequence) & 0x0F;
          my_frame->payload_length = 1;
        }
        return TRUE;
      } else
      {
        /* Receive not ready */
        my_frame->ssap  = his_frame->dsap;
        my_frame->dsap  = his_frame->ssap;
        my_frame->ptype = LLCP_PTYPE_RNR;
        my_frame->payload[0] = (his_sequence) & 0x0F;
        my_frame->payload_length = 1;
      }
      return TRUE;

    case LLCP_PTYPE_RR :
      /* Client has received something from us, nice to give this feedback, but... */
      return TRUE;

    case LLCP_PTYPE_DISC :
      /* Client is leaving, goodbye! */
      printf("SNEP server: remote client has gone\n");

      if (server_long_payload != NULL)
      {
        free(server_long_payload);
        server_long_payload = NULL;
      }

      if (snep_feature == SERVER_THEN_CLIENT)
      {
        /* We become client when the remote client leaves us */
        my_frame->dsap  = LLCP_ADDR_SNEP;
        my_frame->ssap  = LLCP_ADDR_CLIENT;
        my_frame->ptype = LLCP_PTYPE_CONNECT;
        my_frame->payload_length = 0;
      }

      return TRUE;

    default :
      break;
  }

  return FALSE;
}

/*
 * LLCP_SnepClient_Response
 * ------------------------
 * SNEP client core
 */
BOOL LLCP_SnepClient_Response(const BYTE his_message[], DWORD his_message_length, BYTE my_message[], DWORD *my_message_length)
{
  BYTE  his_request;
  BYTE  my_response = SNEP_REQ_REJECT;
  DWORD his_length;
  DWORD my_length = 0;
  const BYTE *his_data = &his_message[6];
  BYTE *my_data  = &my_message[6];

  *my_message_length = 0;

  his_request  = his_message[1];
  his_length   = his_message[2];
  his_length <<= 8;
  his_length  |= his_message[3];
  his_length <<= 8;
  his_length  |= his_message[4];
  his_length <<= 8;
  his_length  |= his_message[5];

//  printf("SNEP> %02X %02X %ld\n", version, his_request, his_length);

  if ((his_length + 6) < his_message_length)
  {
    printf("SNEP message is too long\n");
    return FALSE;

  } else
  if ((his_length + 6) > his_message_length)
  {
    printf("SNEP message is splitted, just ACK\n");
    return TRUE;
  }

  switch (his_request)
  {
    case SNEP_RSP_SUCCESS :
      printf("SNEP client: the server has received our message\n");
      return FALSE; /* This will DISC from the server */

    default :
      /* Unsupported request */
      my_response = SNEP_RSP_BAD_REQUEST;
  }


  /* Format the answer */
  my_message[0] = SNEP_VERSION;
  my_message[1] = my_response;
  my_message[2] = (BYTE) (my_length >> 24);
  my_message[3] = (BYTE) (my_length >> 16);
  my_message[4] = (BYTE) (my_length >> 8);
  my_message[5] = (BYTE) (my_length);

  *my_message_length = my_length + 6;

  /* OK */
  return TRUE;
}

/*
 * LLCP_SnepClient_Handle
 * ----------------------
 * Frame handler (and query provider) for SNEP client
 */
BOOL LLCP_SnepClient_Handle(const LLCP_FRAME_ST *his_frame, LLCP_FRAME_ST *my_frame)
{
  BYTE his_sequence;
  BYTE my_sequence;

  switch (his_frame->ptype)
  {
    case LLCP_PTYPE_CC :
      /* The server has accepted us - Send him our message */
      printf("SNEP client: connected to remote server, sending our message\n");

      /* Send our NDEF right now within a PUT request */
      if (snep_feature != SERVER_THEN_CLIENT)
      {
        /* We are only client --> just say "Hello" */
        if (!LLCP_Create_PutNdef_Payload(&my_frame->payload[1], &my_frame->payload_length, NDEF_HELLO, sizeof(NDEF_HELLO)))
        {
          printf("Failed to create NDEF PUT request (client)\n");
          return FALSE; 
        }
      } else
      {
        /* We were server --> say "Thanks" */
        if (!LLCP_Create_PutNdef_Payload(&my_frame->payload[1], &my_frame->payload_length, NDEF_THANKS, sizeof(NDEF_THANKS)))
        {
          printf("Failed to create NDEF PUT request (client)\n");
          return FALSE; 
        }
      }

      my_frame->ssap  = his_frame->dsap;
      my_frame->dsap  = his_frame->ssap;
      my_frame->ptype = LLCP_PTYPE_I;
      my_frame->payload_length += 1;
      return TRUE;

    case LLCP_PTYPE_I :
      /* The server is sending us something */

      /* Retrieve the sequence number */
      his_sequence = (his_frame->payload[0] >> 4) & 0x3F;
      my_sequence  = (his_frame->payload[0] & 0x3F) - 1;

      if (LLCP_SnepClient_Response(&his_frame->payload[1], his_frame->payload_length-1, &my_frame->payload[1], &my_frame->payload_length))
      {
        /* Message handled */
        if (my_frame->payload_length)
        {
          /* Send an information in response */
          my_frame->ssap  = his_frame->dsap;
          my_frame->dsap  = his_frame->ssap;
          my_frame->ptype = LLCP_PTYPE_I;
          my_frame->payload[0] = (((my_sequence + 1) << 4) & 0xF0) | ((his_sequence + 1) & 0x0F);
          my_frame->payload_length += 1;
        } else
        {
          /* Send only a receive ready frame */
          my_frame->ssap  = his_frame->dsap;
          my_frame->dsap  = his_frame->ssap;
          my_frame->ptype = LLCP_PTYPE_RR;
          my_frame->payload[0] = (his_sequence) & 0x0F;
          my_frame->payload_length = 1;
        }
        return TRUE;
      } else
      {
        /* DISC */
        my_frame->ssap  = his_frame->dsap;
        my_frame->dsap  = his_frame->ssap;
        my_frame->ptype = LLCP_PTYPE_DISC;
        my_frame->payload_length = 0;
      }
      return TRUE;

    case LLCP_PTYPE_RR :
      /* Server has received something from us, nice to give this feedback, but... */
      printf("SNEP client: remote server has received our message\n");
      return TRUE;

    case LLCP_PTYPE_DM :
      /* Server acknowledges our DISC */
      printf("SNEP client: disconnected from remote server\n");
      return TRUE;

    default :
      break;
  }

  return FALSE;
}

/*
 * LLCP_Handle
 * -----------
 * Frame handler (and response provider)
 */
BOOL LLCP_Handle(const LLCP_FRAME_ST *his_frame)
{
  DWORD i;
  LLCP_FRAME_ST my_frame;

  memset(&my_frame, 0, sizeof(LLCP_FRAME_ST));

/*
  printf(">>> DSAP=%02X PTYPE=%02X SSAP=%02X  ", his_frame->dsap, his_frame->ptype, his_frame->ssap);
  for (i=0; i<his_frame->payload_length; i++)
    printf("%02X", his_frame->payload[i]);
  printf("\n");
*/

  if (his_frame->ptype == LLCP_PTYPE_SYMM)
  {
    /* Empty (SYMM) frame */
    goto frame_ready;
  }

  if (his_frame->ptype == LLCP_PTYPE_AGF)
  {
    /* Aggregated frame, to be split into smaller frames */
    printf("Aggregated frame not implemented yet\n");
    return FALSE;
  }

  if (his_frame->dsap == LLCP_ADDR_SNEP)
  {
    /* (Remote) SNEP client talking to the (local) SNEP server */
    if (LLCP_SnepServer_Handle(his_frame, &my_frame))
      goto frame_ready;
  }

  if (his_frame->ssap == LLCP_ADDR_SNEP)
  {
    /* (Remote) SNEP server talking to the (local) SNEP client */
    if (LLCP_SnepClient_Handle(his_frame, &my_frame))
      goto frame_ready;
  }

  printf("Not handled LLCP frame?\n");

  printf(">>> DSAP=%02X PTYPE=%02X SSAP=%02X  ", his_frame->dsap, his_frame->ptype, his_frame->ssap);
  for (i=0; i<his_frame->payload_length; i++)
    printf("%02X", his_frame->payload[i]);
  printf("\n");

frame_ready:

  /*
  printf("<<< DSAP=%02X PTYPE=%02X SSAP=%02X  ", my_frame.dsap, my_frame.ptype, my_frame.ssap);
  for (i=0; i<my_frame.payload_length; i++)
    printf("%02X", my_frame.payload[i]);
  printf("\n");
  */

  return LLCP_SenderEnqueue(&my_frame);
}


