/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2003-2014 PRO ACTIVE SAS, FRANCE - www.springcard.com
   
  ref_desfire_benchmark.c
  -----------------------

  This reference application test the overall communication speed
  with a NXP DESFire EV1 cards.
  
  JDA 30/10/2012 : initial release
  JDA 13/08/2014 : moved to Visual C++ Express 2010, added the 'A'
                   suffix to all text-related functions  

*/
#include "products/springprox/springprox.h"
#include "cardware/desfire/sprox_desfire.h"

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef WIN32
int gettimeofday(struct timeval* p, void* tz)
{
  ULARGE_INTEGER ul; // As specified on MSDN.
  FILETIME ft;

  // Returns a 64-bit value representing the number of
  // 100-nanosecond intervals since January 1, 1601 (UTC).
  GetSystemTimeAsFileTime(&ft);

  // Fill ULARGE_INTEGER low and high parts.
  ul.LowPart = ft.dwLowDateTime;
  ul.HighPart = ft.dwHighDateTime;
  // Convert to microseconds.
  ul.QuadPart /= 10UL;
  // Remove Windows to UNIX Epoch delta.
  ul.QuadPart -= 11644473600000000UL;
  // Modulo to retrieve the microseconds.
  p->tv_usec = (long) (ul.QuadPart % 1000000UL);
  // Divide to retrieve the seconds.
  p->tv_sec = (long) (ul.QuadPart / 1000000UL);

  return 0;
}
#endif

#ifdef __linux__
#include <inttypes.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif

#ifdef WIN32
#define CHECK_RC() { if (rc!=0) { printf("\nline %d - failed (%d) %s\n", __LINE__-1, rc, SPROX_Desfire_GetErrorMessage(rc)); return FALSE; } }
#endif

#ifdef __linux__
#define CHECK_RC() { if (rc!=0) { printf("\nline %d - failed (%d) %s\n", __LINE__-1, rc, SPROX_Desfire_GetErrorMessage(rc)); return FALSE; } }
#endif

static BOOL sprox_desfire_benchmark(void);

const BYTE abNullKey[24]         = { 0 };

const BYTE abRootKey1K[24]       = { "ABCDEFGHABCDEFGHABCDEFGH" };
const BYTE abRootKey2K[24]       = { "Card Master Key!Card Mas" };
const BYTE abRootKey3K[24]       = { "Card Master Key!        " };

int main(int argc, char **argv)
{
  SWORD rc;
  int i;
  char s_buffer[64];
  BYTE atq[2];
  BYTE sak[1];
  BYTE snr[12];
  BYTE snrlen = 12;
  BYTE ats[32];
  BYTE atslen = 32;

  BYTE card_identifier  = 0xFF;  /* Default 0xFF means 'no CID' */
	BOOL use_iso_wrapping = FALSE;
	BYTE want_dsi_dri     = 0;

  char *device = NULL;

  printf("SpringCard SDK for K663/K632/K531/SpringProx-CF/CSB Legacy\n");
  printf("\n");
  printf("ref_desfire_benchmark : NXP MIFARE DESFire EV0/EV1 benchmarking\n");
  printf("---------------------\n");
  printf("This is a free and unsupported sample provided by www.springcard.com\n");
  printf("Please read LICENSE.txt for details\n");
  printf("\n");

  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-b0"))
		{
		  want_dsi_dri = 0;
		} else
    if (!strcmp(argv[i], "-b1"))
		{
		  want_dsi_dri = 1;
		} else
    if (!strcmp(argv[i], "-b2"))
		{
		  want_dsi_dri = 2;
		} else
    if (!strcmp(argv[i], "-b3"))
		{
		  want_dsi_dri = 3;
		} else
    if (!strcmp(argv[i], "-iw"))
		{
		  use_iso_wrapping = TRUE;
		} else
    if (!strcmp(argv[i], "-uc"))
		{
		  card_identifier = 1;
		} else
    if (i == argc-1)
    {
      device = argv[i];
      printf("Device : '%s'\n", device);
    } else
    {
      printf("usage: ref_desfire_benchmark [-v] [-b0|-b1|-b2|-b3] [-iw] [-uc] [device]\n");
      printf("       -b0 : set baudrate to 106kbit/s\n");
      printf("       -b1 : set baudrate to 212kbit/s\n");
      printf("       -b2 : set baudrate to 424kbit/s\n");
      printf("       -b3 : set baudrate to 848kbit/s\n");
			printf("       -iw : use ISO 7816-4 wrapping of Desfire commands\n");
			printf("       -uc : use a card identifier (CID)\n");
      printf("\n");
      return 0;
    }
  }

  rc = SPROX_GetLibrary(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");


  SPROX_Desfire_SelectCid(card_identifier);

	if (use_iso_wrapping)
	  SPROX_Desfire_IsoWrapping(DF_ISO_WRAPPING_CARD);
  else
		SPROX_Desfire_IsoWrapping(DF_ISO_WRAPPING_OFF);

  /* Open reader */
  /* ----------- */

  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  rc = SPROX_ReaderGetDevice(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s\n", s_buffer);

  rc = SPROX_ReaderGetFirmware(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", s_buffer);

  /* Configure reader in ISO 14443-A mode */
  /* ------------------------------------ */
  rc = SPROX_SetConfig(CFG_MODE_ISO_14443_A);
  if (rc != MI_OK)
  {
    printf("Failed to configure the reader\n");
    goto close;
  }

  /* RF field ON */
  SPROX_ControlRF(TRUE);

    /* Select any tag in the ISO 14443-A RF field */
    /* ------------------------------------------ */

  printf("Looking for a tag...\n");

  rc = SPROX_TclA_ActivateIdle(atq, snr, &snrlen, sak);
  if (rc == MI_CARD_NOT_TCL)
  {
    /* On a bien trouve un badge, mais il s'arrete au type 3 */
    printf("The selected tag is not ISO 14443-4 (T=CL)\n");
    goto tag_halt;
  }
  if (rc == MI_NOTAGERR)
  {
    /* Pas de badge dans le champ */
    printf("No available tag in RF field\n");
    goto close;
  }
  if (rc != MI_OK)
  {
    /* Erreur */
    printf("Failed to select tag\n");
    goto close;
  }

  printf("T=CL tag found, SNR=");
  for (i = 0; i < snrlen; i++)
    printf("%02X", snr[i]);
  printf(" ATQ=%02X%02X SAK=%02X\n", atq[1], atq[0], sak[0]);

  /* Is the tag a DESFire ? */
  /* ---------------------- */
  if ( (atq[1] != 0x03) || ( (atq[0] != 0x44) && (atq[0] != 0x04) ))
  {
    printf("This is not a DESFire tag\n");
    goto tag_halt;
  }

  /* Open a T=CL session on the tag */
  /* ------------------------------ */
  rc = SPROX_TclA_GetAts(card_identifier, ats, &atslen);
  if (rc != MI_OK)
  {
    /* Erreur */
    printf("Failed to activate the tag (%d)\n", rc);
    goto tag_halt;
  }

  printf("T=CL tag activated, ATS=");
  for (i = 0; i < atslen; i++)
    printf("%02X", ats[i]);
  printf(" (TA1=%02X TA2=%02X TA3=%02X)\n", ats[2], ats[3], ats[4]);

  /* PPS is not really interesting since we stay at 106kbps */
  rc = SPROX_TclA_Pps(card_identifier, want_dsi_dri, want_dsi_dri);
  if (rc == MI_OK)
  {
    printf("PPS OK, DSI=DRI=%d\n", want_dsi_dri);
  } else
  if ((rc == MI_UNKNOWN_FUNCTION) || (rc == MI_WRONG_PARAMETER))
  {
    printf("PPS not supported by this reader\n");
  } else
  {
    /* Erreur */
    printf("Failed to negociate PPS (DSI=DRI=%d)\n", want_dsi_dri);
    goto tag_stop;
  }


  /* Check DESFire library functions */
  /* ------------------------------- */

  sprox_desfire_benchmark();

  goto tag_stop;

tag_stop:

  /* T=CL (ISO 14443-4) tag stop */
  /* --------------------------- */

  rc = SPROX_TclA_Deselect(card_identifier);
  if (rc != MI_OK)
  {
    printf("T=CL DESELECT failed (%d)\n", rc);
    goto tag_halt;
  }

tag_halt:

  /* ISO 14443-3 tag stop */
  /* -------------------- */

  rc = SPROX_TclA_Halt();
  if (rc != MI_OK)
  {
    printf("HALT failed\n");
    goto close;
  }
  
  printf("SUCCESS!\n");

close:
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc != MI_OK)
    printf("%s (%d)\n", SPROX_GetErrorMessageA(rc), rc);

  return 0;
}

static void clock_increment_diff(struct timeval *to, const struct timeval *inc, const struct timeval *dec)
{
  to->tv_sec += inc->tv_sec;
  to->tv_sec -= dec->tv_sec;

  to->tv_usec += inc->tv_usec;
  to->tv_usec -= dec->tv_usec;
  
  while (to->tv_usec > 1000000)
  {
    to->tv_usec -= 1000000;
	to->tv_sec += 1;
  }
}

static unsigned long clock_to_ms(const struct timeval *c)
{
  unsigned long r;
  
  r  = c->tv_sec * 1000;
  r += c->tv_usec / 1000;
  
  return r;
}

static float clock_bitrate(const struct timeval *c, unsigned long nBytes)
{
  float r, t;
  
  t  = (float) c->tv_sec;
  t += (float) c->tv_usec / (float) 1000000.0;
  
  r  = (float) nBytes / t;
  r *= 8.0; /* bytes -> bits */
  r /= 1024.0;
  
  return r;
}


static BOOL sprox_desfire_benchmark(void)
{
	DF_VERSION_INFO stVersionInfo;
	WORD wDesFireSoftwareVersion;
	SWORD rc;
	BYTE bKeyVersion;
	BYTE bNApplicationsFound;
  BYTE bStdDataFileID = 1;
	DWORD dwNbBytesRead;
	BYTE abDataBuffer[2048];

  int i;
  DWORD nBytes = 0;
  struct timeval t_read;
  struct timeval t_write;
  struct timeval t0, t1;
  
  t_read.tv_sec = t_read.tv_usec = 0;
  t_write.tv_sec = t_write.tv_usec = 0;
  

	//  Get the card's version information.
	rc = SPROX_Desfire_GetVersion(&stVersionInfo);
	CHECK_RC();
	wDesFireSoftwareVersion = (unsigned int) ((unsigned int) stVersionInfo.bSwMajorVersion << 8 | stVersionInfo.bSwMinorVersion);
  printf("\n");

  if (wDesFireSoftwareVersion < 0x0100)
  {
	  printf("This is a Desfire EV0!\n");

  } else
	{
    printf("This is a Desfire EV1!\n");
  }
   
	//  After activating a DesFire card the currently selected 'application' is the card
	//  iTestMode (AID = 0). Therefore the next command is redundant.
	//  We use the command to check that the card is responding correctly.
	rc = SPROX_Desfire_SelectApplication(0);
	CHECK_RC();

	//  Which Card Master Key is currently in use ?
	//  Beginning with DesFire version 0.1 the GetKeyVersion command can help determining
	//  which key is currently effective.
	rc = SPROX_Desfire_GetKeyVersion(0, &bKeyVersion);
	CHECK_RC();

	//  Authenticate with the key which corresponds to the retieved key version.
	if (bKeyVersion == 0)
	{
		rc = SPROX_Desfire_Authenticate(0, abNullKey);
		if (rc == (DFCARD_ERROR-DF_AUTHENTICATION_ERROR))
		{
	    rc = SPROX_Desfire_AuthenticateIso(0, abNullKey);
			CHECK_RC();

      //  The key is configured for 3DES3K
			//  for obvious reasons we cannot revert the config to 3DES2K, but a hook by AES permits it
			rc = SPROX_Desfire_ChangeKeyAes(DF_APPLSETTING2_AES, 0xAE, abRootKey2K, NULL);
			CHECK_RC();
			rc = SPROX_Desfire_AuthenticateAes(0, abRootKey2K);
		}
		CHECK_RC();


	} else
	if (bKeyVersion == 0xAA)
	{
		rc = SPROX_Desfire_Authenticate(0, abRootKey1K);
		if (rc == (DFCARD_ERROR-DF_AUTHENTICATION_ERROR))
		{
	    rc = SPROX_Desfire_AuthenticateIso(0, abRootKey1K);

      //  The key is configured for 3DES3K
			//  for obvious reasons we cannot revert the config to 3DES2K, but a hook by AES permits it
			rc = SPROX_Desfire_ChangeKeyAes(DF_APPLSETTING2_AES, 0xAE, abRootKey2K, NULL);
			CHECK_RC();
			rc = SPROX_Desfire_AuthenticateAes(0, abRootKey2K);
		}
	  CHECK_RC();

	} else
	if (bKeyVersion == 0xC7)
	{
		rc = SPROX_Desfire_Authenticate(0, abRootKey2K);
		if (rc == (DFCARD_ERROR-DF_AUTHENTICATION_ERROR))
		{
			rc = SPROX_Desfire_AuthenticateIso(0, abRootKey2K);

      //  The key is configured for 3DES3K
			//  for obviously we cannot revert the config to 3DES2K, but a hook by AES permits it
			rc = SPROX_Desfire_ChangeKeyAes(DF_APPLSETTING2_AES, 0xAE, abRootKey2K, NULL);
			CHECK_RC();
			rc = SPROX_Desfire_AuthenticateAes(0, abRootKey2K);
		}
	  CHECK_RC();

	} else
	if (bKeyVersion == 0xAE)
	{
    rc = SPROX_Desfire_AuthenticateAes(0, abRootKey2K);
	  CHECK_RC();

	} else
	{
		//  Who knows the key ?
		printf("Sorry, the master key is unknown (version = %02X)\n", bKeyVersion);
		return FALSE;
	}

  //  Back to default (null) key
  rc = SPROX_Desfire_ChangeKey(0, abNullKey, NULL);
  CHECK_RC();

	//  Authenticate again
	rc = SPROX_Desfire_Authenticate(0, abNullKey);
	CHECK_RC();

	//  Try to allow to change the key
	rc = SPROX_Desfire_ChangeKeySettings(0xF);
	CHECK_RC();

	//  List the existing applications
	rc = SPROX_Desfire_GetApplicationIDs(0, NULL, &bNApplicationsFound);
	CHECK_RC();

  if (bNApplicationsFound)
	{
		//  Startup with a blank card !!!
		rc = SPROX_Desfire_FormatPICC();
		CHECK_RC();
	}

  //  Card is ready...
	printf("\n");

	//  Authenticate again
	rc = SPROX_Desfire_Authenticate(0, abNullKey);
	CHECK_RC();

	//  Create one application
	//  It has no keys and its key settings are least restrictive.
	//  Note: since there are no keys, the Change Keys key settting must be assigned
	//        the value 0xE or 0xF.
  rc = SPROX_Desfire_CreateApplication(0xAAAAAA, 0xFF, 0);
	CHECK_RC();

	//  Create one file in application A:
	//  Since the application does not have any keys associated with it, the file access rights
	//  settings 0xE (public access granted) and 0xF (access denied) are the only permissable
	//  ones.
	rc = SPROX_Desfire_SelectApplication(0xAAAAAA);
	CHECK_RC();

	//  Create a the Standard Data file.
	rc = SPROX_Desfire_CreateStdDataFile(bStdDataFileID, 0, 0xEEEE, 2048);
	CHECK_RC();

	//  Fill the Standard Data file with some data.
	memset(abDataBuffer, 0xDA, sizeof(abDataBuffer));

  printf("Performing benchmark, please wait...\n");

  for (i=0; i<16; i++)
  {
    //  Write the file content at once
    gettimeofday(&t0, NULL);
	rc = SPROX_Desfire_WriteData(bStdDataFileID, DF_COMM_MODE_PLAIN, 0, 2048, abDataBuffer);
    gettimeofday(&t1, NULL);
	CHECK_RC();
    clock_increment_diff(&t_write, &t1, &t0);
    printf("."); fflush(NULL);
  
	  //  Read the file content at once
    gettimeofday(&t0, NULL);
	rc = SPROX_Desfire_ReadData(bStdDataFileID, DF_COMM_MODE_PLAIN, 0, 2048, abDataBuffer, &dwNbBytesRead);
    gettimeofday(&t1, NULL);
	CHECK_RC();
    clock_increment_diff(&t_read, &t1, &t0);
    printf("."); fflush(NULL);

    nBytes += 2048;
  }

  printf("\nCleanup...\n");
  
	//  Back to the root application
	rc = SPROX_Desfire_SelectApplication(0);
	CHECK_RC();

	//  Get authenticated
	rc = SPROX_Desfire_Authenticate(0, abNullKey);
	CHECK_RC();

	//  Delete all demonstration data from the card.
	rc = SPROX_Desfire_FormatPICC();
	CHECK_RC();

  printf("Read : %dB received in %ldms (%fkbit/s)\n", nBytes, clock_to_ms(&t_read), clock_bitrate(&t_read, nBytes));
  printf("Write: %dB sent in %ldms (%fkbit/s)\n", nBytes, clock_to_ms(&t_write), clock_bitrate(&t_write, nBytes));
  
#ifdef __linux__
  {
    struct rusage usage;
    getrusage(RUSAGE_SELF, &usage);
    printf("user CPU time:                %ld.%06ld\n", usage.ru_utime.tv_sec, usage.ru_utime.tv_usec);
	printf("system CPU time:              %ld.%06ld\n", usage.ru_stime.tv_sec, usage.ru_stime.tv_usec);
	printf("voluntary context switches:   %ld\n", usage.ru_nvcsw);
	printf("involuntary context switches: %ld\n", usage.ru_nivcsw);
  }
#endif

  printf("\n");
	return TRUE;
}

