/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2000-2014 PRO ACTIVE SAS, FRANCE - www.springcard.com
   
  ref_find.c
  ----------

  This is the reference applications that shows how to poll the reader,
  showing each incoming tag identifier
  
  LTX 25/08/2009 : initial release
  JDA 24/01/2012 : minor changes to adapt to release 1.6x of the SDK
  JDA 04/02/2013 : minor changes to adapt to release 1.7x of the SDK
  JDA 13/08/2014 : moved to Visual C++ Express 2010, added the 'A'
                   suffix to all text-related functions  

*/
#include "products/springprox/springprox.h"
#include "products/springprox/sprox_capas.h"

#ifdef WIN32
#include <conio.h>              /* For kbhit and getch */
#endif
#ifdef __linux
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

const char *device = NULL;
WORD find_protos = 0;
BOOL find_once = FALSE;

void usage(void)
{
  printf("usage: ref_find [PROTOCOL LIST] [-once] [device]\n\n");
  printf("Write PROTOCOL LIST as follow:\n");
  printf(" -* : all protocols\n");
  printf(" -a : for ISO 14443-A\n");
  printf(" -b : for ISO 14443-B\n");
  printf(" -c : for Innovatron (old Calypso cards)\n");
  printf(" -v : for ISO 15693\n");
  printf(" -i : for ICODE1\n");
  printf(" -f : for Felica\n");  
  printf(" -z : for Topaz/Jewel (NFC Forum type 1)\n");
  printf(" -k : for Kovio RF barcode\n");  
  printf(" -p : for Inside PicoTag (also HID iClass)\n");
  printf(" -m : for other memory cards (ST SR..., ASK CTS)\n");
  printf("If PROTOCOL LIST is empty, all protocols are tried (same as -*)\n\n");
  printf("If -once is set, the software exits as soon as a card is found.\n\n");
  printf("The last parameter shall be the name of the communication device.\n");
  printf("The default device is read from Registry or from /etc/springprox.cfg\n\n");
}

BOOL parse_args(int argc, char **argv)
{
  int i;

  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-*"))
    {
      find_protos = 0xFFFF;
    } else
    if (!strcmp(argv[i], "-a"))
    {
      find_protos |= PROTO_14443_A;
    } else
    if (!strcmp(argv[i], "-b"))
    {
      find_protos |= PROTO_14443_B;
    } else
    if (!strcmp(argv[i], "-c"))
    {
      find_protos |= PROTO_INNOVATRON;
    } else
    if (!strcmp(argv[i], "-v"))
    {
      find_protos |= PROTO_15693;
    } else
    if (!strcmp(argv[i], "-i"))
    {
      find_protos |= PROTO_ICODE1;
    } else
    if (!strcmp(argv[i], "-f"))
    {
      find_protos |= PROTO_FELICA;
    } else
    if (!strcmp(argv[i], "-z"))
    {
      find_protos |= PROTO_INNOVISION_JEWEL;
    } else
    if (!strcmp(argv[i], "-k"))
    {
      find_protos |= PROTO_KOVIO_BARCODE;
    } else
    if (!strcmp(argv[i], "-p"))
    {
      find_protos |= PROTO_INSIDE_PICO;
    } else
    if (!strcmp(argv[i], "-m"))
    {
      find_protos |= PROTO_ST_SR|PROTO_ASK_CTS;
    } else
    if (!strcmp(argv[i], "-once"))
    {
      find_once = TRUE;
    } else
    if (!strcmp(argv[i], "-debug"))
    {
      SPROX_SetVerbose(0xFF, NULL);
    } else
    if (!strcmp(argv[i], "-?"))
    {
      usage();
      return FALSE;
    } else
    if (i == (argc-1))
    {
      device = argv[i];
    } else
    {
      printf("Unrecognized parameter '%s'\n", argv[i]);
      return FALSE;
    }
  }

  if (find_protos == 0)
    find_protos = 0xFFFF;

  return TRUE;
}

int main(int argc, char **argv)
{
  SWORD rc;
  int i;
  char s_buffer[64];
  BYTE uid[32];
  BYTE uid_len;
  BYTE info[32];
  BYTE info_len;

  DWORD features;
  WORD proto = 0;

  /* Display the informations and check the command line */
  /* --------------------------------------------------- */

  printf("SpringCard SDK for K663/K632/K531/SpringProx-CF/CSB Legacy\n");
  printf("\n");
  printf("ref_find : multi-protocol card polling sample\n");
  printf("--------\n");
  printf("This is a free and unsupported sample provided by www.springcard.com\n");
  printf("Please read LICENSE.txt for details\n");
  printf("\n");

  if (!parse_args(argc, argv))
    return EXIT_FAILURE;

  rc = SPROX_GetLibraryA(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");

  /* Open the reader */
  /* --------------- */

  rc = SPROX_ReaderOpenA(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  /* Display reader's information */
  printf("Reader found\n");
  rc = SPROX_ReaderGetDeviceA(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s\n", s_buffer);

  rc = SPROX_ReaderGetFirmwareA(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", s_buffer);

  /* Does the reader support the protocol we want to use ? */
  rc = SPROX_ReaderGetFeatures(&features);
  if (rc == MI_OK)
  {
    if (!(features & SPROX_WITH_ISO_14443))
    {
      if (find_protos & (PROTO_14443_A|PROTO_14443_B|PROTO_INSIDE_PICO|PROTO_ST_SR|PROTO_ASK_CTS|PROTO_INNOVATRON|PROTO_INNOVISION_JEWEL|PROTO_KOVIO_BARCODE))
      {
        printf("Warning: the reader doesn't support ISO 14443 and related protocols\n");
      }
    }
    if (!(features & SPROX_WITH_ISO_15693))
    {
      if (find_protos & (PROTO_15693|PROTO_ICODE1))
      {
        printf("Warning: the reader doesn't support ISO 15693 and related protocols\n");
      }
    }
    if (!(features & SPROX_WITH_FELICA))
    {
      if (find_protos & (PROTO_FELICA))
      {
        printf("Warning: the reader doesn't support Felica and related protocols\n");
      }
    }
  }

  /* Now perform the polling */
  /* ----------------------- */

  printf("\nWaiting for cards... (Press <Ctrl>+C to exit)\n\n");

  for (;;)
  {
    rc = MI_OK;
    /* Not too often... */
#ifdef WIN32
    if (kbhit())
      break;
    Sleep(50);
#endif
#ifdef __linux
    usleep(5000);
#endif

    for (;;)
    {
      /* Invoke reader's SPROX_FIND function (with subsequent SPROX_FIND_INFO) */
      /* --------------------------------------------------------------------- */

      uid_len = sizeof(uid);
      info_len = sizeof(info);
      rc = SPROX_FindEx(find_protos, &proto, uid, &uid_len, info, &info_len);


      if (rc != MI_OK)
      {
	      if (rc == MI_NOTAGERR)
          continue; /* No card */

        printf("\n[Failed]\n");
		    printf("  %s (%d)\n", SPROX_GetErrorMessageA(rc), rc);
        if (rc < -128)
          goto close; /* Fatal error */

      } else
      {
        /* A card has been found */
        /* --------------------- */

        printf("\nTag found [OK]\n");
        printf("Family=");

        switch (proto)
        {
          case PROTO_14443_A :
            printf("ISO/IEC 14443-A (or NXP Mifare)\n");
            break;
          case PROTO_14443_B :
            printf("ISO/IEC 14443-B\n");
            break;
          case PROTO_15693 :
            printf("ISO/IEC 15693\n");
            break;
          case PROTO_ICODE1 :
            printf("NXP ICODE-1\n");
            break;
          case PROTO_INSIDE_PICO :
            printf("Inside Secure PicoTag\n");
            break;
          case PROTO_ST_SR :
            printf("STMicroElectronics SR\n");
            break;
          case PROTO_ASK_CTS :
            printf("ASK CTS256B/CTS512B\n");
            break;
          case PROTO_INNOVATRON :
            printf("Calypso (Innovatron protocol)\n");
            break;
          case PROTO_INNOVISION_JEWEL :
            printf("Innovision Jewel\n");
            break;
          case PROTO_KOVIO_BARCODE :
            printf("Kovio RF barcode\n");
            break;
          case PROTO_FELICA :
            printf("Felica\n");
            break;
          default :
            printf("%04X\n", proto);
        }
    
        printf("UID=");
        for (i=0; i<uid_len; i++)
          printf("%02X", uid[i]);
        printf("\n");

        /* Leave the loop to show the card's details */
        break;
      }
    }

    switch (proto)
    {
      case PROTO_14443_A :

        /* INFO returns the ATQA (2 bytes) then the SAK (1 byte) */
        /* ----------------------------------------------------- */

        {

          WORD atq;
          BYTE sak;

          atq = info[0] | info[1] << 8;
          sak = info[2];

          printf("ATQ=%04X, SAK=%02X\n", atq, sak);

          /* See NXP AN10834 and AN10833 */
          printf("Based on the SAK, this card is ");
          if (sak & 0x02)
          {
            printf("a NFC object (peer-to-peer capable)\n");
          } else
          {
            if (sak & 0x08)
            {
              if (sak & 0x10)
              {
                printf("a Mifare Classic 4K\n");
              } else
              if (sak & 0x01)
              {
                printf("a Mifare Mini\n");
              } else
              {
                printf("a Mifare Classic 1K\n");
              }
            } else
            {
              if (sak & 0x10)
              {
                if (sak & 0x01)
                {
                  printf("a Mifare Plus 4K in SL2\n");
                } else
                {
                  printf("a Mifare Plus 2K in SL2\n");
                }
              } else
              {
                if (sak & 0x20)
                {
                  printf("T=CL compliant\n");
                } else
                {
                  printf("a Mifare UltraLight\n");
                }
              }
            }
          }

          /* Halt the card */
          SPROX_A_Halt();
        }
        break;

      case PROTO_14443_B :

        /* INFO returns the ATQB (11 or maybe 12 bytes somedays) */
        /* ----------------------------------------------------- */

        printf("ATQB=");
		    for (i = 0; i < info_len; i++)
	          printf("%02X", info[i]);
		    printf("\n");

	      /* Halt the card */
  	    SPROX_B_Halt(uid);
        break;

      case PROTO_15693 :

        /* The UID starts with the vendor name (ISO 7816-6) */
        /* ------------------------------------------------ */

        if (uid[0] != 0xE0)
        {
          printf("Strange! UID doesn't start with E0\n");
        }

        switch (uid[1])
        {
          case 0x02 :
            /* ST */
            printf("Manufacturer is STMicroElectronics\n");

				    switch ((uid[2] & 0xFC) >> 2)
				    {
				      case 0x05 : /* LRI64 */
					      printf("This is a LRI64\n");
					      break;
				      case 0x16 : /* M24LR16E */
					      printf("This is a M24LR16E\n");
					      break;
				      default   : /* ST card */
                printf("Un-referenced product type: %02X\n", (uid[2] & 0xFC) >> 2);
					      break;
				    }

            break;

          case 0x04 :
            /* NXP */
            printf("Manufacturer is NXP\n");
				    switch (uid[2])
				    {
					    case 0x01 : /* ICODE SLI */
								printf("This is a ICODE SLI\n");
								break;
					    case 0x03 : /* ICODE SLI-L */
								printf("This is a ICODE SLI-L\n");
								break;
					    case 0x02 : /* ICODE SLI-S */
								printf("This is a ICODE SLI-S\n");
								break;
					    default   : /* NXP card */
					      printf("Un-referenced product type: %02X\n", uid[2]);
								break;
				    }
            break;

          case 0x07 :
            /* Texas */
            printf("Manufacturer is Texas Instruments\n");
					  switch (uid[2] & 0xFE)
					  {
					    case 0x00 : /* Tag-it HF-I Plus Inlay */
								printf("This is a Tag-it HF-I Plus Inlay\n");
								break;
					    case 0x80 : /* Tag-it HF-I Plus Chip */
								printf("This is a Tag-it HF-I Plus Chip\n");
								break;
					    case 0xC0 : /* Tag-it HF-I Standard Chip/Inlay */
								printf("This is a Tag-it HF-I Standard\n");
								break;
					    case 0xC4 : /* Tag-it HF-I Pro Chip/Inlay */
								printf("This is a Tag-it HF-I Pro\n");
								break;
					    default   : /* Tag-it card */
					      printf("Un-referenced product type: %02X\n", uid[2] & 0xFE);
								break;
					  }
            break;

          case 0x16 :
            /* EM */
            printf("Manufacturer is EM Marin (or Legic)\n");
            break;

          default :
            printf("Un-referenced manufacturer: %02X\n", uid[1]);
        }

	      /* Halt the card */
	      SPROX_Iso15693_Halt();
        break;
        
      default :
        break;
	  }

    if (find_once)
      break;
  }

close:

  /* Close the reader */
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc != MI_OK)
    printf("%s (%d)\n", SPROX_GetErrorMessageA(rc), rc);

    
  return EXIT_SUCCESS;
}
