#include "products/springprox/springprox.h"
#include "cardware/hidsio/sprox_hidsio.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static const char *PROGRAM = "ref_hidsio";
static const char *DESCR = "Demo program for HID SIO with SpringCard API & reader";
static const char *COPYRIGHT = "Copyright (c) 2013 PRO ACTIVE SAS, FRANCE - www.springcard.com";

#define NB_BOOKS 8
#define NB_PAGES 32
#define NB_BLOCKS 128

static const BYTE TEST_DATA[8] = { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08 };
static const BYTE BLANK_BLOCK[8] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };

const char *device = NULL;
BYTE sam_slot = 0;
BOOL test_auth = TRUE;
BOOL test_loop = FALSE;
BOOL use_sam_anticoll_instead_of_find = FALSE;

BOOL parse_args(int argc, char **argv)
{
  int i;

  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-l"))
    {
      test_loop = TRUE;
    } else
    if (!strcmp(argv[i], "-q"))
    {
      test_auth = FALSE;
    } else	
    if (!strcmp(argv[i], "-a"))
    {
      use_sam_anticoll_instead_of_find = TRUE;
    } else	
    if (!strcmp(argv[i], "-s") && ((i+1)<argc))
    {
      sam_slot = (BYTE) atoi(argv[++i]);
    } else
    if (!strcmp(argv[i], "-d") && ((i+1)<argc))
    {
      device = argv[++i];
    } else
    if (i == (argc-1))
    {
      device = argv[i];
    } else
    {
      printf("Unrecognized parameter '%s'\n", argv[i]);
      printf("Try %s -h for help\n\n", PROGRAM);
      return FALSE;
    }
  }

  return TRUE;
}

#ifdef WIN32
int gettimeofday(struct timeval* p, void* tz)
{
  ULARGE_INTEGER ul; // As specified on MSDN.
  FILETIME ft;

  // Returns a 64-bit value representing the number of
  // 100-nanosecond intervals since January 1, 1601 (UTC).
  GetSystemTimeAsFileTime(&ft);

  // Fill ULARGE_INTEGER low and high parts.
  ul.LowPart = ft.dwLowDateTime;
  ul.HighPart = ft.dwHighDateTime;
  // Convert to microseconds.
  ul.QuadPart /= 10UL;
  // Remove Windows to UNIX Epoch delta.
  ul.QuadPart -= 11644473600000000UL;
  // Modulo to retrieve the microseconds.
  p->tv_usec = (long) (ul.QuadPart % 1000000UL);
  // Divide to retrieve the seconds.
  p->tv_sec = (long) (ul.QuadPart / 1000000UL);

  return 0;
}
#endif

#ifdef __linux__
#include <inttypes.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif

static void clock_diff(struct timeval *t1, const struct timeval *t0)
{
  t1->tv_sec  -= t0->tv_sec;
  
  if (t1->tv_usec >= t0->tv_usec)
  {
    t1->tv_usec -= t0->tv_usec;
  } else
  {
    t1->tv_sec  -= 1;
    t1->tv_usec += 1000000;
    t1->tv_usec	-= t0->tv_usec;
  }
}

static unsigned long clock_to_ms(const struct timeval *c)
{
  unsigned long r;
  
  r  = c->tv_sec * 1000;
  r += c->tv_usec / 1000;
  
  return r;
}

int main(int argc, char **argv)
{
  BYTE version[2];
  BYTE firmware[32];
  BYTE uid[32];
  BYTE uid_len;
  WORD protocol;	  
  BYTE usn[12];
  BYTE pacsbits[256];
  WORD pacsbits_len;
  BYTE track_result;
  WORD book;
  WORD page;
  WORD block;
  BYTE data[4 * 8];
  DWORD i;
  SWORD rc;
  struct timeval t0, t1;  

  printf("SpringCard :: SpringProx API SDK\n");
  printf("\n");
  printf("%s : %s\n", PROGRAM, DESCR);
  printf("%s\n\n", COPYRIGHT);

  if (!parse_args(argc, argv))
    return EXIT_FAILURE;

  /* Tell the HID SIO library to use the SPROX interface */
  /* --------------------------------------------------- */
  
  rc = SPROX_HidSio_RF_Assign_Sprox();
  if (rc)
  {
    printf("SPROX_HidSio_RF_Assign_Sprox: %d\n", rc);
    goto failed;
  }

  /* Connect to the contactless reader */
  /* --------------------------------- */

  rc = SPROX_ReaderOpen(device);
  if (rc)
  {
    printf("SPROX_ReaderOpen: %d\n", rc);
    goto failed;
  }
  
  rc = SPROX_ReaderGetFirmware(firmware, sizeof(firmware));
  if (rc)
  {
    printf("SPROX_ReaderGetFirmware: %d\n", rc);
    goto failed;
  }
  
  printf("Reader firmware: %s\n", firmware);

  /* Does this reader have a card interface (for the SAM) ? */
  /* ------------------------------------------------------ */
  
  rc = SPROX_Card_GetFirmware(firmware, sizeof(firmware));
  if (rc)
  {
    printf("SPROX_Card_GetFirmware: %d\n", rc);
    goto failed;
  }
  
  printf("Card interface: %s\n", firmware);
  
  /* Connect to the SAM */
  /* ------------------ */

  rc = SPROX_HidSio_Sam_Connect_Sprox(sam_slot);
  if (rc)
  {
    printf("SPROX_HidSio_Sam_Connect_Sprox(%d): %d\n", sam_slot, rc);
    goto failed;
  }

  /* Get SAM's data */
  /* -------------- */

  rc = SPROX_HidSio_Sam_GetVersionInfo(version, firmware);
  if (rc)
  {
    printf("SPROX_HidSio_GetVersionInfo: %d\n", rc);
	goto failed;
  }
  
  printf("Connected to HID SIO:\n");
  printf("\tVersion=%02X.%02X\n", version[0], version[1]);
  printf("\tFirmware=");
  for (i=0; i<6; i++)
    printf("%02X", firmware[i]);
  printf("\n");

  rc = SPROX_HidSio_Sam_GetSerialNumber(usn);
  if (rc)
  {
    printf("SPROX_HidSio_GetSerialNumber: %d\n", rc);
    goto failed;
  }

  printf("\tSerial number=");
  for (i=0; i<12; i++)
    printf("%02X", usn[i]);
  printf("\n");
  
  /* Now wait for a card */
  /* ------------------- */

wait_for_a_card:
  
  printf("Please place the test card on the antenna...\n");  
  
  if (!use_sam_anticoll_instead_of_find)
  {
    /* Use the reader's FIND method - We try all protocols to be representative */
    /* ------------------------------------------------------------------------ */
    
    for (;;)
    {
      uid_len = sizeof(uid);
      rc = SPROX_Find(0xFFFF, &protocol, uid, &uid_len);
      if (rc == MI_OK)
        break; /* Card found */
      if (rc != MI_NOTAGERR)
      {
        /* Error */
        printf("SPROX_Find: %d\n", rc);
        goto failed;
      }
    }

    printf("SPROX_Find got a card, proto=%04X, UID=", protocol);
    for (i=0; i<uid_len; i++)
      printf("%02X", uid[i]);
    printf("\n");

    if ((protocol != PROTO_INSIDE_PICO) || (uid_len != 8))
    {
      printf("This is not an Inside Pico!!!\n");
      goto failed;
    }

    /* Init the library with the CSN of the card */
    rc = SPROX_HidSio_SetDetectedCardInfo_Picopass(uid);
    if (rc)
    {
      printf("SPROX_HidSio_SetDetectedCardInfo_Picopass: %d\n", rc);
      goto failed;
    }
    
    /* Just to check this works! */
    rc = SPROX_HidSio_TrackCard(&track_result);
    if (rc)
    {
      printf("SPROX_HidSio_TrackCard (%d): %ld\n", 1, rc);
      goto failed;
    }

    switch (track_result)
    {
      case HIDSIO_RSP_NO_VALID_CARDINFO :
        printf("SPROX_HidSio_TrackCard (%d) returned 'NO VALID CARD INFO'\n", 1);
        goto failed;
      case HIDSIO_RSP_NO_CARD_IN_FIELD :
        printf("SPROX_HidSio_TrackCard (%d) returned 'CARD ABSENT'\n", 1);
        goto failed;
      case HIDSIO_RSP_CARD_IN_FIELD :
        break;
      default :
        printf("SPROX_HidSio_TrackCard (%d) returned '%02X'\n", 1, track_result);
        goto failed;
    }
	
  } else  
  {
    /* SAM method (slower of course) */
    /* ----------------------------- */

    BYTE cardinfo[256];
    WORD cardinfo_len;	
    BYTE csn[32];
    WORD csn_len;
    BYTE atsatqatr[32];
    WORD atsatqatr_len;
    BYTE sak;
    BYTE cid;
    BYTE fwi;
	
    /* Set RF Field ON */
    rc = SPROX_HidSio_RF_SetFieldOn();
    if (rc)
    {
      printf("SPROX_HidSio_RF_SetFieldOn: %d\n", rc);
      goto failed;
    }
  
    for (;;)
    {	
      /* Ask the SAM to find a iClass card on the RF field */
      cardinfo_len = sizeof(cardinfo);
      rc = SPROX_HidSio_PerformAnticollision(PROTO_14443_B, 0, cardinfo, &cardinfo_len);
	    if (rc == MI_OK)
	      break; /* Card found */
	    if (rc != HIDSIO_ERROR_SAM_CARD_NOT_FOUND)
	    {
	      /* Error! */
        printf("SPROX_HidSio_PerformAnticollisionEx: %d\n", rc);
	      goto failed;
      }
	  }

	  printf("Card found! (using the SAM)\n");

    printf("CardInfo=");
    for (i=0; i<cardinfo_len; i++)
      printf("%02X", cardinfo[i]);
    printf("\n");

    /* Explain the returned data */
    csn_len = sizeof(csn);
    atsatqatr_len = sizeof(atsatqatr);

    rc = SPROX_HidSio_DecodeCardInfo(cardinfo, cardinfo_len, &protocol, csn, &csn_len, atsatqatr, &atsatqatr_len, &sak, &cid, &fwi);
    if (rc)
    {
      printf("SPROX_HidSio_DecodeCardInfo: %d\n", rc);
      return EXIT_FAILURE;
    }

    printf("\tProtocol=%04X, CSN=", protocol);
    for (i=0; i<csn_len; i++)
      printf("%02X", csn[i]);
    printf("\n");
  }
  
  /* Read the PAC */
  /* ------------ */
  
  gettimeofday(&t0, NULL);

  pacsbits_len = sizeof(pacsbits);
  rc = SPROX_HidSio_GetContentElement(NULL, 0, pacsbits, &pacsbits_len);
  if (rc)
  {
    printf("SPROX_HidSio_GetContentElement: %d\n", rc);
    return EXIT_FAILURE;
  }
  
  gettimeofday(&t1, NULL);  

  printf("PAC read (%d unused bits)=", pacsbits[0]);
  for (i=1; i<pacsbits_len; i++)
    printf("%02X", pacsbits[i]);
  printf("\n");
  
  clock_diff(&t1, &t0);
  printf("(read PAC took %ldms)\n", clock_to_ms(&t1));

  /* After reading the PAC, the SAM stops the RF field. Silly, isn't it ? Let's try to find the card again. This time using only FIND */
  /* -------------------------------------------------------------------------------------------------------------------------------- */
  
  uid_len = sizeof(uid);
  rc = SPROX_Find(PROTO_INSIDE_PICO, &protocol, uid, &uid_len);
  if (rc == MI_NOTAGERR)
  {
	  /* Card lost ? */
	  printf("Card lost after reading the PAC\n");
	  goto failed;
  }
  if (rc)
  {
    /* Other error */
    printf("SPROX_Find: %d\n", rc);
    goto failed;
  }

  /* Init the library with the CSN of the card */
  rc = SPROX_HidSio_SetDetectedCardInfo_Picopass(uid);
  if (rc)
  {
    printf("SPROX_HidSio_SetDetectedCardInfo_Picopass: %d\n", rc);
    goto failed;
  }
  
  /* Select book 0, page 2 */
  /* --------------------- */

  book = 0;
  page = 2;
  
  gettimeofday(&t0, NULL);  

  rc = SPROX_HidSio_Picopass_Select((BYTE) book, (BYTE) page);
  if (rc)
  {
    printf("SPROX_HidSio_Picopass_Select (book %d, page %d): %d\n", book, page, rc);
    goto failed;
  }

  /* Read 128 blocks (i.e. 1024 bytes) */
  /* --------------------------------- */

  for (block=0; block<NB_BLOCKS; block+=4)
  {
    rc = SPROX_HidSio_Picopass_Read4Blocks((BYTE) block, data);
    if (rc)
    {
      printf("SPROX_HidSio_Picopass_Read4Block (book %d, page %d, block %d): %d\n", book, page, block, rc);
      goto failed;
    }
  }
  
  if (!test_auth)
  {
    /* Stop here */
	  /* --------- */
	
    gettimeofday(&t1, NULL);  
    clock_diff(&t1, &t0);
    printf("(read %d  blocks (%dB) took %ldms)\n", NB_BLOCKS, 8*NB_BLOCKS, clock_to_ms(&t1));
	
  } else
  {
    /* Authentication on book 0, page 2, application 1, using Kmd2 key */
    /* --------------------------------------------------------------- */

    rc = SPROX_HidSio_Picopass_MutualAuth(TRUE, 0x030026);
    if (rc)
    {
      printf("SPROX_HidSio_Picopass_MutualAuth (debit, 030026): %d\n", rc);
      goto failed;
    }
  
    gettimeofday(&t1, NULL);  
    clock_diff(&t1, &t0);
    printf("(read %d  blocks (%dB) + mutual authentication took %ldms)\n", NB_BLOCKS, 8*NB_BLOCKS, clock_to_ms(&t1));

    /* Write some data into block 7 */
    /* ---------------------------- */

    block = 7;
    memcpy(data, TEST_DATA, 8);

    rc = SPROX_HidSio_Picopass_WriteBlock((BYTE) block, data);
    if (rc)
    {
      printf("SPROX_HidSio_Picopass_WriteBlock (book %d, page %d, block %d): %d\n", book, page, block, rc);
      goto failed;
    }

    /* Read back block 7 */
    /* ----------------- */

    rc = SPROX_HidSio_Picopass_ReadBlock((BYTE) block, data);
    if (rc)
    {
      printf("SPROX_HidSio_Picopass_ReadBlock (book %d, page %d, block %d): %d\n", book, page, block, rc);
      goto failed;
    }

    /* Check the data - and restore an empty block */
    /* ------------------------------------------- */

    if (memcmp(data, TEST_DATA, 8))
    {
      printf("Read/write data error (book %d, page %d, block %d)!\n", book, page, block);
      goto failed;
    }

    rc = SPROX_HidSio_Picopass_WriteBlock((BYTE) block, BLANK_BLOCK);
    if (rc)
    {
      printf("SPROX_HidSio_Picopass_WriteBlock (book %d, page %d, block %d): %d\n", book, page, block, rc);
      goto failed;
    }
  }

  /* Everything is OK! */
  /* ----------------- */
  
  printf("Please remove the test card...\n");

wait_removal:

  /* Wait until card is removed */
  /* -------------------------- */
  rc = SPROX_HidSio_TrackCard(&track_result);
  if (rc)
  {
    printf("SPROX_HidSio_TrackCard (%d): %ld\n", 1, rc);
    goto failed;
  }

  switch (track_result)
  {
    case HIDSIO_RSP_NO_VALID_CARDINFO :
      printf("SPROX_HidSio_TrackCard (%d) returned 'NO VALID CARD INFO'\n", 1);
      goto failed;
    case HIDSIO_RSP_NO_CARD_IN_FIELD :
      /* Card removed */
      break;
    case HIDSIO_RSP_CARD_IN_FIELD :
	  /* Card still present */
	  goto wait_removal;
    default :
	  printf("SPROX_HidSio_TrackCard (%d) returned '%02X'\n", 1, track_result);
      goto failed;
  }

  if (test_loop)
    goto wait_for_a_card;

  /* Drop the connection to the SAM */
  /* ------------------------------ */
  rc = SPROX_HidSio_Sam_Disconnect();
  if (rc)
  {
    printf("SPROX_HidSio_Sam_Disconnect: %d\n", rc);
    goto failed;
  }
  
  rc = SPROX_ReaderClose();
  if (rc)
  {
    printf("SPROX_ReaderClose: %d\n", rc);
    goto failed;
  }	
  
  printf("SUCCESS!\n");
  return EXIT_SUCCESS;
  
failed:
  printf("FAILED!\n");
  SPROX_HidSio_Sam_Disconnect();
  SPROX_ReaderClose();
  return EXIT_FAILURE;  
}

