/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
	
  Copyright (c) 2009, 2010 PRO ACTIVE SAS - www.proactive-group.com
   
  Ref_15693.c
  -----------

  This is the reference applications that shows how to work with
  ISO/IEC 15693 cards.
  
  DRI 25/04/2014 Add features for ICODE SLI cards EAS management.

*/
#include "products/springprox/springprox.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Select bit 37 (XNP documentation), please take caution when 				*/
/* considering the byte order. Here UID 0 is UID 7 for NXP. 				  */ 
#define SLIX_MASK 							0x10

/* Program command set.																								*/
#define CMD_NONE									 0
#define CMD_SET_EAS 							 1
#define CMD_RESET_EAS 						 2
#define CMD_LOCK_EAS 							 3
#define CMD_PASSWORD_PROTECTED_EAS 4
#define CMD_SET_PASSWORD					 5
#define CMD_WRITE_PASSWORD				 6
#define CMD_LOCK_PASSWORD					 7
#define CMD_GET_RANDOM_NB					 8

#define CMD_HELP								110
#define CMD_UNDEFINED						111

/* ICODE Command set. 																								*/
#define SET_EAS									0xA2
#define RESET_EAS								0xA3
#define LOCK_EAS								0xA4
#define PASSWORD_PROTECTED_EAS  0xA6
#define SET_PASSWORD						0xB3
#define WRITE_PASSWORD					0xB4
#define LOCK_PASSWORD						0xB5
#define GET_RANDOM_NB						0xB2
#define PASSWORD_ID_EAS					0x10
				
/* Globals. 																													*/
WORD block_size  = 0;
WORD block_count = 0;
BOOL has_read_multiple = TRUE;
BOOL is_writable = TRUE;

/* Prototypes. 																												*/
BYTE get_command (char * command);
void set_EAS(BYTE snr[8]);
void reset_EAS(BYTE snr[8]);
void lock_EAS(BYTE snr[8]);
void protect_EAS(BYTE snr[8]);
void set_password(BYTE snr[8], BYTE password[4]);
void write_password(BYTE snr[8], BYTE password[4]);
void lock_password(BYTE snr[8]);
short get_random_nb(BYTE snr[8], BYTE * rd_nb);
int hex_to_bytes(BYTE* buffer, size_t buffer_len, char* ascii_string, size_t ascii_string_len);
void DecodeUid(const BYTE uid[8]);
void GetSystemInformation(void);
void display_help(void);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
/* 								ICODE SLI COMMUNICATION REFERENCE CODE						  */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
int main(int argc, char **argv)
{
  SWORD rc;
  int i = 0;
	char *device = NULL;
	
	BYTE uid[8];
	BYTE rd_nb[2];
	BYTE password[4] = {0,0,0,0};
	BYTE command = CMD_NONE;
	BOOL success = TRUE;
  DWORD features;
  TCHAR s_buffer[64];
	
	/* Command interpreter. 																						*/
	if ((argc > 1) && ( argc < 5))
	{
		device = argv[1];
		
		if (argc > 2)
		{
			command = get_command(argv[2]);
			if (command == CMD_UNDEFINED)
			{
				printf("Unrecognized command type.\n");
				display_help();
				goto done;
			}				
		}
	}
	
	/* Concerning --set_password and --write_password. 									*/
	if ((command == CMD_SET_PASSWORD)  || (command == CMD_WRITE_PASSWORD))
	{
		if (argv[3] != NULL)
		{
			if (strlen(argv[3]) != 8)
			{
				printf("Incorrect password size (4 bytes required).\n");
				printf("%d", strlen(argv[3]));
				goto done;
			} 
			else	
				hex_to_bytes(password, 4, argv[3], 8);
			
		} else
		{
			printf("You must specify a password with this command.\n");
			goto done;
		}
	}

  printf("SpringCard SDK for K531/SpringProx-CF/CSB Legacy\n");
  printf("\n");
  printf("ISO 15693 reference demo\n");
  printf("-------------------------------------\n");
  printf("www.springcard.com\n\n");

  rc = SPROX_GetLibrary(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
  /* 												OPEN READER 															*/
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found.\n");
		printf("Error code : %d.\n", rc);
    goto done;
  }
  printf("Reader found.\n");
  rc = SPROX_ReaderGetDevice(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s.\n", s_buffer);

  rc = SPROX_ReaderGetFirmware(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s.\n", s_buffer);

  rc = SPROX_ReaderGetFeatures(&features);
  if ((rc == MI_OK) && !(features & 0x00000002))
  {
    printf("Warning: this reader is not 15693-enabled\n");
    printf("         all subsequent tests will fail !\n");
  }

  printf("\n------------------------------\n");

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
  /* 					 Configure reader in ISO 15693 standard mode						*/
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
	
  rc = SPROX_SetConfig(CFG_MODE_ISO_15693);
  if (rc != MI_OK)
  {
    printf("Failed to configure the reader for ISO 15693 operation.\n");
    goto close;
  }

  /* RF field ON */
  SPROX_ControlRF(TRUE);

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
  /* 					 	Select any ISO 15693 tag in the RF field							*/
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
  
  printf("Put your ISO 15693 tag on the reader...\n");

  for (;;)
  {
	  rc = SPROX_Iso15693_SelectAny(0, uid);
	  if (rc == MI_OK)
	  {
      printf("Found ISO 15693 tag in RF field.\n");
      break;
	  }
  }
	
  printf("Tag found: UID=");
  for (i = 0; i < 8; i++)
    printf("%02X", uid[i]);
  printf("\n\n");

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
  /* 					 						Command interpreter													*/
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *  */
	switch(command)
	{
		case CMD_NONE :
			DecodeUid(uid);
		break;
		
		case CMD_SET_EAS :
			set_EAS(uid);
			break;
		
		case CMD_RESET_EAS :
			reset_EAS(uid);
			break;
		
		case CMD_LOCK_EAS :
			printf("Uncomment if you really want to unlock this feature.\n");
			//lock_EAS(uid);
			break;
		
		case CMD_PASSWORD_PROTECTED_EAS :
			protect_EAS(uid);
			break;
		
		case CMD_SET_PASSWORD :
			set_password(uid, password);
			break;
		
		case CMD_WRITE_PASSWORD :
			write_password(uid, password);
			break;
			
		case CMD_LOCK_PASSWORD :
			printf("Uncomment if you really want to unlock this feature.\n");
			//lock_password(uid);
			break;
		
		case CMD_GET_RANDOM_NB :
			get_random_nb(uid,rd_nb);
			break;

		case CMD_HELP :
			display_help();
			break;
			
		default:
			printf("Unimplemented...\n");
			display_help();
			goto close;
			break;
	}
	
	
close:
  /* Close the reader */
  SPROX_ReaderClose();

done:
  printf("Done.\n");
	
  return 0;
}

void set_EAS(BYTE snr[8])
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	SHORT rc = -1;
	
	printf("Setting EAS...\n");
	rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, SET_EAS, NULL, 0, buffer, &recv_length, 1024);
	
	if (rc == MI_OK)
		printf("EAS set successfuly. \n");
	else
	{
		printf("Failed setting EAS. \n");
		printf("Command failed, error code : %d\n",rc);
	}
}

void reset_EAS(BYTE snr[8])
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	SHORT rc = -1;
	
	printf("Resetting EAS...\n");
	rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, RESET_EAS, NULL, 0, buffer, &recv_length, 1024);
	
	if (rc == MI_OK)
		printf("EAS reset successfuly. \n");
	else
	{
		printf("Failed resetting EAS. \n");
		printf("Command failed, error code : %d\n",rc);
	}
}

void lock_EAS(BYTE snr[8])
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	SHORT rc = -1;
	
	printf("Locking EAS...\n");
	rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, LOCK_EAS, NULL, 0, buffer, &recv_length, 1024);
	
	if (rc == MI_OK)
		printf("EAS locked successfuly. \n");
	else
	{
		printf("Failed locking EAS. \n");
		printf("Command failed, error code : %d\n",rc);
	}
}

void protect_EAS(BYTE snr[8])
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	SHORT rc = -1;
	
	printf("Enable password protected EAS...\n");
	rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, PASSWORD_PROTECTED_EAS, NULL, 0, buffer, &recv_length, 1024);
	
	if (rc == MI_OK)
		printf("EAS protected successfuly. \n");
	else
	{
		printf("Failed protecting EAS. \n");
		printf("Command failed, error code : %d\n",rc);
	}
}

short get_random_nb(BYTE snr[8], BYTE * rd_nb)
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	BYTE i = 0;
	SHORT rc = 0;
	
	printf("Getting random number: ");
	rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, GET_RANDOM_NB, NULL, 0, buffer, &recv_length, 0);
	if (rc == MI_OK )
	{
	
		rd_nb[0] = buffer[0];
		rd_nb[1] = buffer[1];
	
		/* Two last byte dedicated to CRC. 																*/
		for (i=0;i<(recv_length-2);i++)
			printf("%02X ",buffer[i]);
			
		printf("\n");
	} else
	{ 
		printf("Command failed, error code : %d\n",rc);
	}
	
	return rc;
}

void set_password(BYTE snr[8], BYTE password[4])
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	SHORT rc = 0;
	BYTE rd_nb[2] = {0,0};
	BYTE cmd_params[5] = {0,0,0,0,0};
	
	printf("Setting password...\n");
	rc = get_random_nb(snr, rd_nb);
	if ( rc == MI_OK)
	{
		password[0] = password[0] ^ rd_nb[0];
		password[1] = password[1] ^ rd_nb[1];
		password[2] = password[2] ^ rd_nb[0];
		password[3] = password[3] ^ rd_nb[1];

		cmd_params[0] = PASSWORD_ID_EAS;
		memcpy(&cmd_params[1],password,sizeof(password));

		rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, SET_PASSWORD, cmd_params, 5, buffer, &recv_length, 0);
		
		if (rc == MI_OK)
			printf("Set password for EAS successfuly.\n");
		else
		{
			printf("Failed to set password for EAS.\n");
			printf("Command failed, error code : %d\n",rc);
		}
			
	} else
	{
		printf("Command failed, error code : %d\n",rc);
		printf("Failed to set password for EAS.\n");
	}
}

void write_password(BYTE snr[8], BYTE password[4])
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	SHORT rc = 0;
	BYTE cmd_params[5] = {0,0,0,0,0};
	
	printf("Writting password...\n");
	cmd_params[0] = PASSWORD_ID_EAS;
	memcpy(&cmd_params[1],password,sizeof(password));
	
	rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, WRITE_PASSWORD, cmd_params, 5, buffer, &recv_length, 0);
	
	if (rc == MI_OK)
		printf("Wrote password for EAS successfuly.\n");
	else
	{
		printf("Failed to write password for EAS.\n");
		printf("Command failed, error code : %d\n",rc);
	}		
	 
}

void lock_password(BYTE snr[8])
{
	BYTE buffer[256];
	WORD recv_length = sizeof(buffer);
	BYTE cmd_params = PASSWORD_ID_EAS;
	SHORT rc = 0;
	
	rc = SPROX_Iso15693_ExchangeCustomCommand(FALSE, 0x04, snr, LOCK_PASSWORD, &cmd_params, 1, buffer, &recv_length, 0);
	
	if (rc == MI_OK)
		printf("Locked password for EAS successfuly.\n");
	else
	{
		printf("Failed to lock password for EAS.\n");
		printf("Command failed, error code : %d\n",rc);
	}		
} 

/* Display ICODE SLI label type.																			*/ 
void DecodeUid(const BYTE uid[8])
{
	/* Check if the card is from NXP. 																	*/
	if ( uid[1] == 0x04 )
	{
		printf("Manufacturer:   NXP.\n");
		
		switch (uid[2])
		{
			/* ICODE SLI : 28 blocks, 4 bytes in each block.								*/
			case 0x01 : 
									/* If Password protected AFI bit is cleared, SLI    */
									if ( !(uid[3] & SLIX_MASK))
										printf("Chip:           ICODE-SLI.\n");
									else
									/* Else, SLI-X. 																		*/
										printf("Chip:           ICODE-SLI-X.\n");
									break;
			
			/* ICODE SLI-S : 40 blocks, 4 bytes in each block.							*/
			case 0x02 : 
									/* If Password protected AFI bit is cleared, SLI    */
									if ( !(uid[3] & SLIX_MASK))
										printf("Chip:           ICODE-SLI-S.\n");
									else
									/* Else, SLI-X. 																		*/
										printf("Chip:           ICODE-SLI-X-S.\n");
										
									has_read_multiple = FALSE;
									break;
			/* ICODE SLI-L : 8 blocks, 4 bytes in each block .							*/						
			case 0x03 : 
									/* If Password protected AFI bit is cleared, SLI    */
									if ( !(uid[3] & SLIX_MASK))
										printf("Chip:           ICODE-SLI-L.\n");
									else
									/* Else, SLI-X. 																		*/
										printf("Chip:           ICODE-SLI-X-L.\n");
										
									has_read_multiple = FALSE;
									break;
									
			default   : /* Other NXP card. 																	*/
									printf("Unknown chip!\n");
									break;
		}
		
	} else
	{
		printf("Unmanaged manufacturer/chip.\n");
	}

}

void GetSystemInformation(void)
{
  SWORD rc;
  int i = 0;
  BYTE data[256];
  WORD len = sizeof(data);
  
  printf("GetSystemInformation ");
  rc = SPROX_Iso15693_GetSystemInformation(NULL, data, &len);
  if (rc == MI_OK) 
  {
	  printf("[OK]\n");

    if (len > 9)
    {
      printf("UID=");
	    for (i = 1; i < 9; i++)
		    printf("%02X", data[9-i]);
      printf("\n");

      if (data[0] & 0x01)
      {
        printf("DSFID=%02X\n", data[i++]);
      } 
                  
      if (data[0] & 0x02)
      {
        printf("AFI=%02X\n", data[i++]);
      } 
                  
      if (data[0] & 0x04)
      {
			  block_count = data[i++] + 1;
        printf("Number of Blocks=%d\n", block_count);
				block_size = (data[i++] & 0x1F) + 1;
        printf("Block Size=%d\n", block_size);
      }
                  
      if (data[0] & 0x08)
      {
        printf("IC-Ref=%02X\n", data[i++]);
      }
    }
  } else
  {
	  printf("[Failed]\n");
    printf("  %s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }
}

/* Return a command specific ID. 																			*/
BYTE get_command (char * command)
{
	BYTE cmd = 0;
	
	if (!strcmp(command,"--set_eas"))
	{
		cmd = CMD_SET_EAS;
		goto end;
	}
	
	if (!strcmp(command,"--reset_eas"))
	{
		cmd = CMD_RESET_EAS;
		goto end;
	}
	
	if (!strcmp(command,"--lock_eas"))
	{
		cmd = CMD_LOCK_EAS;
		goto end;
	}
	
	if (!strcmp(command,"--protect_eas"))
	{
		cmd = CMD_PASSWORD_PROTECTED_EAS;
		goto end;
	}
	
	if (!strcmp(command,"--get_random_nb"))
	{
		cmd = CMD_GET_RANDOM_NB;
		goto end;
	}
	
	if (!strcmp(command,"--set_password"))
	{
		cmd = CMD_SET_PASSWORD;
		goto end;
	}
	
	if (!strcmp(command,"--write_password"))
	{
		cmd = CMD_WRITE_PASSWORD;
		goto end;
	}
	
	if (!strcmp(command,"--lock_password"))
	{
		cmd = CMD_LOCK_PASSWORD;
		goto end;
	}
	
	if (!strcmp(command,"--help"))
	{
		cmd = CMD_HELP;
		goto end;
	}

	cmd = CMD_UNDEFINED;
		
	end :
	return cmd;
}

BOOL is_hex(BYTE c)
{
  if ((c >= '0') && (c <= '9')) return TRUE;
 if ((c >= 'A') && (c <= 'F')) return TRUE;
 if ((c >= 'a') && (c <= 'f')) return TRUE;
 return FALSE;
}

BYTE hex_to_byte(BYTE c)
{
  if ((c >= '0') && (c <= '9')) return c - '0';
 if ((c >= 'A') && (c <= 'F')) return c - 'A' + 10;
 if ((c >= 'a') && (c <= 'f')) return c - 'a' + 10;
 return 0xFF;
}

int hex_to_bytes(BYTE* buffer, size_t buffer_len, char* ascii_string, size_t ascii_string_len)
{

 size_t i, j;
 char b;

 if (ascii_string_len % 2 != 0)
  return -1;

 j=0;
 for (i = 0; i< ascii_string_len; i++)
 {
  b = ascii_string[i];
  if (!is_hex(b))
   return -1;

  if ((i % 2) == 0)
  {
   buffer[j] = (hex_to_byte(b) << 4) & 0xF0;
  } else
  {
   buffer[j] += 0x0F & hex_to_byte(b);
   j++;
  }
 }

 return (int) j;
}

void display_help(void)
{
	printf("\n");																
	printf("COMMAND STRUCTURE\n");
	printf("[executable_name.exe] [COM] [COMMAND] [OPTION]\n\n");
	
	printf("[COM]\n");
	printf("Serial port used with reader (ex: COM1...).\n\n");

	printf("[COMMANDS]\n");
	printf("*** Available for all ICODE SLI types. ***\n");	
	printf("--set_eas\n");
	printf("--reset_eas\n");
	printf("--lock_eas\n");
	printf("--protect_eas\n\n");
	printf("*** For SLI-S and SLI-L. ***\n");
	printf("--get_random_nb\n"); 
	printf("--set_password\n");
	printf("--write_password\n");
	printf("--lock_password\n");
	printf("--help\n\n");
	
	printf("[OPTIONS]\n"); 
	printf("4 BYTES password for --set_password \n"); 
	printf("and --write_password commands.\n");

}


