/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2012 PRO ACTIVE SAS - www.proactive-group.com
   
  ref_p2p_initiator.c
  -------------------

  This is the reference application that shows how to work with an
  NFC object in peer-to-peer mode, the SpringCard reader being
  an initator.

  This test scenario relies on the Simple NDEF Exchange Protocol
  (SNEP), as defined by NFC Forum. SNEP is implemented in Android 4.0
  and newer.

  There are 3 possible execution modes:

  1. We (the initiator) behave as a SNEP server. We wait for a NDEF
     message pushed by the target's SNEP client (PUT request).
     For instance, use the free application 'NFCShare' (available
     on Google Play Store) to send us a Text message.

  2. We (the initiator) behave as a SNEP client. We push our own Text
     message ("Hello, NFC peer!") to the target's SNEP server in a
     PUT request.

  3. We (the initiator) behave as a SNEP server. Once a NDEF message
     has been received from the target's SNEP client, we become a
     SNEP client and push our own Text message ("Thanks, NFC peer!").
  
*/

#include "products/springprox/springprox.h"

#ifdef WIN32
#include <conio.h>              /* For kbhit and getch */
#endif
#ifdef __linux
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "p2p_common.h"

int snep_feature = SERVER_THEN_CLIENT;



int main(int argc, char **argv)
{
  SWORD rc;
  int i;
  char s_buffer[64];
  BYTE uid[12];
  BYTE uid_len = 12;
  DWORD features;
  WORD proto = 0;

  LLCP_FRAME_ST my_frame;
  LLCP_FRAME_ST his_frame;

  char *device = NULL;

  if (argc > 1)
    device = argv[1];

  printf("SpringCard SDK for K531/SpringProx-CF/CSB Legacy\n");
  printf("\n");
  printf("ref_p2p_initiator\n");
  printf("-----------------\n");
  printf("www.springcard.com\n\n");

  /* Open reader */
  /* ----------- */

  rc = SPROX_GetLibrary(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");

  /* Open reader */
  /* ----------- */

  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  printf("Reader found\n");
  rc = SPROX_ReaderGetDevice(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s\n", s_buffer);

  rc = SPROX_ReaderGetFirmware(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", s_buffer);

  rc = SPROX_ReaderGetFeatures(&features);
  if ((rc == MI_OK) && !(features & 0x00000008))
  {
    printf("Warning: this reader is not 18092-enabled\n");
  }

  printf("\n------------------------------\n");


  /* Set RF field to ON */
  SPROX_ControlRF(TRUE);

  printf("Waiting for target... (Press <Ctrl>+C to exit)\n\n");

  for (;;)
  {
    rc = MI_OK;
    /* Not too often... */
#ifdef WIN32
    if (kbhit())
      break;
    Sleep(50);
#endif
#ifdef __linux
    usleep(5000);
#endif

    for (;;)
    {
      uid_len = sizeof(uid);
      rc = SPROX_Find(PROTO_FELICA|PROTO_14443_A, &proto, uid, &uid_len);
	    if (rc == MI_NOTAGERR)
        continue;                  /* No more tag (or no tag at all) */

      if (rc != MI_OK)
      {
        printf("\n[Failed]\n");
		     printf("  %s (%d)\n", SPROX_GetErrorMessage(rc), rc);
        if (rc < -128)
          goto close;           /* Fatal error */
      }

      if (rc == MI_OK)
      {
        printf("\nTag found [OK]\n");
		    printf("Family=");

		    switch (proto)
        {
		      case PROTO_14443_A	: printf("Nfc-A\n"); break;
		      case PROTO_FELICA   : printf("Nfc-F\n"); break;
		      default             : printf("%04X\n", proto);
        }
    
		    printf("NFCID=");
		    for (i=0; i<uid_len; i++)
		      printf("%02X", uid[i]);
		    printf("\n");

		    break;
	    }
	  }

	  if (proto == PROTO_14443_A)
	  {
      BYTE atq[2];
      BYTE snr[12];
      BYTE snrlen = 12;
      BYTE sak[1];
      WORD w_atq;

      rc = SPROX_A_SelectAny(atq, snr, &snrlen, sak);
      if (rc == MI_OK)
      {
		    w_atq = (atq[1] << 8) | atq[0];
		    printf("ATQ of Nfc-A device is %04X\n", w_atq);
        SPROX_A_Halt();
      }
    }

	  if (proto == PROTO_FELICA)
    {
      /* Got a Felica target */
      /* ------------------- */

      if (!LLCP_Activate())
      {
        printf("LLCP_Activate failed\n");

      } else
      {
        /* Connected to LLCP target */
        /* ------------------------ */

        if (snep_feature == CLIENT)
        {
          /* We are a demanding client and try to connect to the SNEP server immediately */
          my_frame.dsap  = LLCP_ADDR_SNEP;
          my_frame.ssap  = LLCP_ADDR_CLIENT;
          my_frame.ptype = LLCP_PTYPE_CONNECT;
          my_frame.payload_length = 0;
          if (!LLCP_SenderEnqueue(&my_frame))
          {
            printf("Failed to enqueue 1st SNEP client request\n");
            return EXIT_FAILURE;
          }
        }

        for (;;)
        {
          /* Get the next message we have to send */
          if (!LLCP_SenderDequeue(&my_frame))
          {
            printf("LLCP_SenderDequeue failed\n");
            return EXIT_FAILURE;
          }

          /* Perform the exchange: send our frame, receive peer's frame */
          if (!LLCP_Exchange(&my_frame, &his_frame))
          {
            /* Peer has been lost */
            break;
          }

          /* Handle peer's frame */
          if (!LLCP_Handle(&his_frame))
          {
            printf("LLCP_Handle failed\n");
            break;
          }
        }
      }

      /* Disconnected from LLCP target, or connection failed, reset the field */
      SPROX_ControlRF(FALSE);
      Sleep(100);
      SPROX_ControlRF(TRUE);
    }
  }

close:
  /* Close the reader */
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc == MI_OK)
  {
    printf("Done\n");
  } else
  {
    printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }
  return 0;
}




/* LCCP says we shall not use DID in underlying NFC-DEP MAC layer */
static const BYTE DID = 0x00;

/*
 * LLCP_ExchangeEx
 * ----------------
 * Exchange frames in NFC-DEP
 * This function uses SPROX_NfcI_Exchange as we are in initiator mode
 */
BOOL LLCP_ExchangeEx(const BYTE send_buffer[], DWORD send_len, BYTE recv_buffer[], DWORD *recv_len)
{
  SWORD rc;
  WORD l_recv_len;
  
  if ((send_buffer == NULL) && (send_len != 0))
    return FALSE;
  
  l_recv_len = (WORD) *recv_len;

  /* Actual exchange */  
  rc = SPROX_NfcI_Exchange(DID, send_buffer, (WORD) send_len, recv_buffer, &l_recv_len);
  if (rc != MI_OK)
  {
    printf("Exchange failed, rc=%d\n", rc);
    return FALSE;
  }

  *recv_len = l_recv_len;

  return TRUE;
}

/*
 * LLCP_Activate
 * -------------
 * Activate the LLCP layer in initiator mode
 * This functions uses SPROX_NfcI_Atr as we are in initiator mode
 */
BOOL LLCP_Activate(void)
{
  BYTE buffer[128];
  WORD slen = 0;
  WORD rlen = sizeof(buffer) - 10;
  SWORD rc;

  memcpy(&buffer[slen], NFC_ATR_MAGIC_NUMBER, sizeof(NFC_ATR_MAGIC_NUMBER));
  slen += sizeof(NFC_ATR_MAGIC_NUMBER);
  memcpy(&buffer[slen], NFC_ATR_GI_PARAMS, sizeof(NFC_ATR_GI_PARAMS));
  slen += sizeof(NFC_ATR_GI_PARAMS);

  rc = SPROX_NfcI_Atr(DID, buffer, slen, &buffer[0], &buffer[10], &rlen);
  if (rc != MI_OK)
  {
    printf("Atr failed, rc=%d\n", rc);
    return FALSE;
  }

  {
    WORD i;
    printf("Nfc Target activated, ID3t=");
    for (i=0; i<10; i++)
      printf("%02X", buffer[i]);
    printf(" Gt=");
    for (i=0; i<rlen-10; i++)
      printf("%02X", buffer[10+i]);
    printf("\n");
  }

  /* Check that we retrieve the NFC-Forum magic number */
  if (memcmp(&buffer[10], NFC_ATR_MAGIC_NUMBER, sizeof(NFC_ATR_MAGIC_NUMBER)))
  {
    printf("Target doesn't announce compliance to NFC-Forum LLCP\n");
    return FALSE;
  }

  printf("NFC-Forum LLCP layer activated\n");
  return TRUE;
}
