/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2000-2013 PRO ACTIVE SAS, FRANCE - www.springcard.com
   
  ref_tcl.c
  ---------

  This is the reference applications that shows how to exchange APDUs
  with cards compliant with ISO 14443 layer 4.

  The test uses either Desfire "GET VERSION" APDU or ISO 7816-4
  "SELECT MASTER FILE" APDU.
  
  JDA 04/02/2013 : initial release
  JDA 14/02/2013 : corrected value of cid_supported returned by Test_TCL_A_Ex

*/
#include "products/springprox/springprox.h"

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static const char *PROGRAM = "ref_tcl";
static const char *DESCR = "ISO 14443-4 'T=CL' reference";
static const char *COPYRIGHT = "Copyright (c) 2000-2013 PRO ACTIVE SAS, FRANCE - www.springcard.com";

const char *device = NULL;
WORD test_protos = 0;
BYTE test_dsi = 0;
BYTE test_dri = 0;

void usage(void)
{
  printf("Usage: %s [protocol] [baudrate] [-once] [device]\n\n", PROGRAM);
  printf("Choose the protocol, either:\n");
  printf(" -a : for ISO 14443-A\n");
  printf(" -b : for ISO 14443-B\n");
  printf(" -c : for Innovatron (old Calypso cards)\n");
  printf(" -* : for all\n");
  printf("If no protocol is specified, all protocols are tried (same as -*)\n\n");
  printf("Choose the baudrates:\n");
  printf(" -dr0, -dr1, -dr2 or -dr3 for PCD->PICC baudrate\n");
  printf(" -ds0, -ds1, -ds2 or -ds3 for PICC->PCD baudrate\n");
  printf("Default is -dr0 -ds0 (106kbit/s both directions)\n\n");
  printf("If -once is set, the software exits as soon as a card is found.\n\n");
  printf("The last parameter shall be the name of the communication device.\n");
  printf("The default device is read from Registry or from /etc/springprox.cfg\n\n");
}

BOOL parse_args(int argc, char **argv)
{
  int i;

  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-*"))
    {
      test_protos = PROTO_14443_A | PROTO_14443_B | PROTO_INNOVATRON;
    } else
    if (!strcmp(argv[i], "-a"))
    {
      test_protos |= PROTO_14443_A;
    } else
    if (!strcmp(argv[i], "-b"))
    {
      test_protos |= PROTO_14443_B;
    } else
    if (!strcmp(argv[i], "-c"))
    {
      test_protos |= PROTO_INNOVATRON;
    } else
    if (!strcmp(argv[i], "-ds0"))
    {
      test_dsi = 0;
    } else
    if (!strcmp(argv[i], "-ds1"))
    {
      test_dsi = 1;
    } else
    if (!strcmp(argv[i], "-ds2"))
    {
      test_dsi = 2;
    } else
    if (!strcmp(argv[i], "-ds3"))
    {
      test_dsi = 3;
    } else
    if (!strcmp(argv[i], "-dr0"))
    {
      test_dri = 0;
    } else
    if (!strcmp(argv[i], "-dr1"))
    {
      test_dri = 1;
    } else
    if (!strcmp(argv[i], "-dr2"))
    {
      test_dri = 2;
    } else
    if (!strcmp(argv[i], "-dr3"))
    {
      test_dri = 3;
    } else
    if (!strcmp(argv[i], "-debug"))
    {
      SPROX_SetVerbose(0xFF, NULL);
    } else
    if (!strcmp(argv[i], "-?"))
    {
      usage();
      return FALSE;
    } else
    if (i == (argc-1))
    {
      device = argv[i];
    } else
    {
      printf("Unrecognized parameter '%s'\n", argv[i]);
      printf("Try %s -h for help\n\n", PROGRAM);
      return FALSE;
    }
  }

  if (test_protos == 0)
    test_protos = PROTO_14443_A | PROTO_14443_B | PROTO_INNOVATRON;

  return TRUE;
}


SWORD Test_TCL_A(const BYTE uid[], BYTE uidlen);
SWORD Test_TCL_B(const BYTE atqb[11]);

int apdu_count = 5;

int main(int argc, char **argv)
{
  SWORD rc;
  int i;
  char s_buffer[64];
  
  BYTE uid_or_pupi[32];
  BYTE uid_or_pupi_len = sizeof(uid_or_pupi);
  BYTE info[32];
  BYTE info_len = sizeof(info);

  WORD got_proto;
	BOOL verbose = FALSE;

  /* Display the informations and check the command line */
  /* --------------------------------------------------- */

  printf("SpringCard :: SpringProx API SDK\n");
  printf("\n");
  printf("%s : %s\n", PROGRAM, DESCR);
  printf("%s\n\n", COPYRIGHT);

  if (!parse_args(argc, argv))
    return EXIT_FAILURE;

  rc = SPROX_GetLibrary(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");

  /* Open the reader */
  /* --------------- */

  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  /* Display reader's information */
  printf("Reader found\n");
  rc = SPROX_ReaderGetDevice(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s\n", s_buffer);

  rc = SPROX_ReaderGetFirmware(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", s_buffer);

  /* Now wait for a card */
  /* ------------------- */
  
  rc = SPROX_FindEx(test_protos, &got_proto, uid_or_pupi, &uid_or_pupi_len, info, &info_len);
  if (rc == MI_NOTAGERR)
  {
    printf("No card on the reader!\n");
    goto close;
  }
  if (rc != MI_OK)
  {
    printf("Failed to select the card!\n");
    goto close;
  }
  
  if (got_proto == PROTO_14443_A)
  {
    /* Test a ISO 14443 type A card */
    /* ---------------------------- */

    printf("Found a ISO 14443-A card, UID=");
    for (i=0; i<uid_or_pupi_len; i++)
      printf("%02X", uid_or_pupi[i]);
    printf("\n");
  
    rc = Test_TCL_A(uid_or_pupi, uid_or_pupi_len);
    if (rc != MI_OK)
      goto close;
  
  } else
  if (got_proto == PROTO_14443_B)
  {
    /* Test a ISO 14443 type A card */
    /* ---------------------------- */

    printf("Found a ISO 14443-B card, PUPI=");
    for (i=0; i<uid_or_pupi_len; i++)
      printf("%02X", uid_or_pupi[i]);
    printf("\n");
    
    printf("                          ATQB=");
    for (i=0; i<info_len; i++)
      printf("%02X", info[i]);
    printf("\n");

    rc = Test_TCL_B(info);
    if (rc != MI_OK)
      goto close;
  
  } else
  {
    printf("Invalid protocol\n");
    goto close;  
  }    

  printf("\n");
  printf("SUCCESS!\n");
  printf("All tests have been successfully passed\n");

close:

  /* Close the reader */
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc == MI_OK)
  {
    printf("Done\n");
  } else
  {
    printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }

  return 0;
}

SWORD Test_TCL_A_Ex(const BYTE uid[], BYTE uidlen, BYTE cid, BYTE dsi, BYTE dri, BOOL skip_pps, BOOL *cid_supported)
{
  static const BYTE send_buffer[] = { 0x90, 0x60, 0x00, 0x00, 0x00 };
  BYTE recv_buffer[256];
  WORD recv_length;
  BYTE ats[32];
  BYTE atslen = sizeof(ats);
  BYTE ta1 = 0x00;
  BYTE tc1 = 0x00;
  int i;
  SWORD rc;
  
  rc = SPROX_TclA_GetAts(cid, ats, &atslen);
  if (rc != MI_OK)
  {
    printf("Get ATS failed!\n");
    return rc;
  }

  /* Understand the ATS */
  if (ats[0] & 0x10)
  {
    ta1 = ats[1];
  }
  if (ats[0] & 0x40)
  {
    if ((ats[0] & 0x10) && (ats[0] & 0x20))
    {
      tc1 = ats[3];
    } else
    if ((ats[0] & 0x10) || (ats[0] & 0x20))
    {
      tc1 = ats[2];
    } else
    {
      tc1 = ats[1];
    }
  }

  if (cid_supported != NULL)
    *cid_supported = TRUE;

  if (!(tc1 & 0x02))
  {
    /* The card doesn't support the CID feature */
    if (cid_supported != NULL)
      *cid_supported = FALSE;
  }

  if ((test_dsi == 1) && !(ta1 & 0x01))
  {
    printf("Bad option: the card doesn't support DSI=1\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dsi == 2) && !(ta1 & 0x02))
  {
    printf("Bad option: the card doesn't support DSI=2\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dsi == 3) && !(ta1 & 0x04))
  {
    printf("Bad option: the card doesn't support DSI=3\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dri == 1) && !(ta1 & 0x10))
  {
    printf("Bad option: the card doesn't support DRI=1\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dri == 2) && !(ta1 & 0x20))
  {
    printf("Bad option: the card doesn't support DRI=2\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dri == 3) && !(ta1 & 0x40))
  {
    printf("Bad option: the card doesn't support DRI=3\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dsi != test_dri) && (ta1 & 0x80))
  {
    printf("Bad option: the card wants DSI=DRI\n");
    return MI_LIB_CALL_ERROR;
  }
  
  if (!skip_pps)
  {
    rc = SPROX_TclA_Pps(cid, dsi, dri);
    if (rc != MI_OK)
    {
      printf("PPS failed!\n");
      return rc;
    }
  }
  
  for (i=0; i<apdu_count; i++)
  {
    recv_length = sizeof(recv_buffer);
    rc = SPROX_Tcl_Exchange(cid, send_buffer, sizeof(send_buffer), recv_buffer, &recv_length);
    if (rc != MI_OK)
    {
      printf("Exchange APDU %d/%d failed!\n", i, apdu_count);
      return rc;
    }
  }
  
  rc = SPROX_Tcl_Deselect(cid);
  if (rc != MI_OK)
  {
    printf("DESELECT failed!\n");
    return rc;
  }
  
  rc = SPROX_A_SelectAgain(uid, uidlen);
  if (rc != MI_OK)
  {
    printf("SelectAgain (WUPA+select) failed!\n");
    return rc;
  }

  return MI_OK;
}

SWORD Test_TCL_A(const BYTE uid[], BYTE uidlen)
{
  SWORD rc;
  BOOL cid_supported;
  
  if ((test_dsi == 0) && (test_dri == 0))
  {  
    /* The PPS is not mandatory */
    /* ------------------------ */

    printf("Test T=CL A, no CID, no PPS\n");
    rc = Test_TCL_A_Ex(uid, uidlen, 0xFF, 0, 0, TRUE, &cid_supported);
    if (rc != MI_OK) return rc;

    printf("Test T=CL A, CID=0, no PPS\n");
    rc = Test_TCL_A_Ex(uid, uidlen, 0, 0, 0, TRUE, NULL);
    if (rc != MI_OK) return rc;

    if (cid_supported)
    {
      printf("Test T=CL A, CID=1, no PPS\n");
      rc = Test_TCL_A_Ex(uid, uidlen, 1, 0, 0, TRUE, NULL);
      if (rc != MI_OK) return rc;

      printf("Test T=CL A, CID=13, no PPS\n");
      rc = Test_TCL_A_Ex(uid, uidlen, 13, 0, 0, TRUE, NULL);
      if (rc != MI_OK) return rc;

      printf("Test T=CL A, CID=14, no PPS\n");
      rc = Test_TCL_A_Ex(uid, uidlen, 14, 0, 0, TRUE, NULL);
      if (rc != MI_OK) return rc;
    }
  }

  /* We must use the PPS */
  /* ------------------- */
  
  printf("Test T=CL A, no CID, PPS dsi=%d, dri=%d\n", test_dsi, test_dri);
  rc = Test_TCL_A_Ex(uid, uidlen, 0xFF, test_dsi, test_dri, FALSE, &cid_supported);
  if (rc != MI_OK) return rc;

  printf("Test T=CL A, CID=0, PPS dsi=%d, dri=%d\n", test_dsi, test_dri);
  rc = Test_TCL_A_Ex(uid, uidlen, 0, test_dsi, test_dri, FALSE, NULL);
  if (rc != MI_OK) return rc;

  if (cid_supported)
  {
    printf("Test T=CL A, CID=1, PPS dsi=%d, dri=%d\n", test_dsi, test_dri);
    rc = Test_TCL_A_Ex(uid, uidlen, 1, test_dsi, test_dri, FALSE, NULL);
    if (rc != MI_OK) return rc;

    printf("Test T=CL A, CID=13, PPS dsi=%d, dri=%d\n", test_dsi, test_dri);
    rc = Test_TCL_A_Ex(uid, uidlen, 13, test_dsi, test_dri, FALSE, NULL);
    if (rc != MI_OK) return rc;

    printf("Test T=CL A, CID=14, PPS dsi=%d, dri=%d\n", test_dsi, test_dri);
    rc = Test_TCL_A_Ex(uid, uidlen, 14, test_dsi, test_dri, FALSE, NULL);
    if (rc != MI_OK) return rc;
  }
  
  return MI_OK;  
}

SWORD Test_TCL_B_Ex(const BYTE pupi[4], BYTE cid, BYTE dsi, BYTE dri)
{
  static const BYTE send_buffer[] = { 0x00, 0xA4, 0x00, 0x00, 0x02, 0x3F, 0x00, 0x00 };
  BYTE recv_buffer[256];
  WORD recv_length;
  int i;
  SWORD rc;
  
  rc = SPROX_TclB_AttribEx(pupi, cid, dsi, dri);
  if (rc != MI_OK)
  {
    printf("ATTRIB failed!\n");
    return rc;
  }
  
  for (i=0; i<apdu_count; i++)
  {
    recv_length = sizeof(recv_buffer);
    rc = SPROX_Tcl_Exchange(cid, send_buffer, sizeof(send_buffer), recv_buffer, &recv_length);
    if (rc != MI_OK)
    {
      printf("Exchange APDU %d/%d failed!\n", i, apdu_count);
      return rc;
    }
  }
  
  rc = SPROX_Tcl_Deselect(cid);
  if (rc != MI_OK)
  {
    printf("DESELECT failed!\n");
    return rc;
  }
  
  rc = SPROX_B_SelectAny(0, NULL);
  if (rc != MI_OK)
  {
    printf("SelectAny (WUPB) failed!\n");
    return rc;
  }

  return MI_OK;
}

SWORD Test_TCL_B(const BYTE atqb[11])
{
  SWORD rc;

  if ((test_dsi == 1) && !(atqb[8] & 0x01))
  {
    printf("Bad option: the card doesn't support DSI=1\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dsi == 2) && !(atqb[8] & 0x02))
  {
    printf("Bad option: the card doesn't support DSI=2\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dsi == 3) && !(atqb[8] & 0x04))
  {
    printf("Bad option: the card doesn't support DSI=3\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dri == 1) && !(atqb[8] & 0x10))
  {
    printf("Bad option: the card doesn't support DRI=1\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dri == 2) && !(atqb[8] & 0x20))
  {
    printf("Bad option: the card doesn't support DRI=2\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dri == 3) && !(atqb[8] & 0x40))
  {
    printf("Bad option: the card doesn't support DRI=3\n");
    return MI_LIB_CALL_ERROR;
  }
  if ((test_dsi != test_dri) && (atqb[8] & 0x80))
  {
    printf("Bad option: the card wants DSI=DRI\n");
    return MI_LIB_CALL_ERROR;
  }

  printf("Test T=CL B, no CID, dsi=%d, dri=%d\n", test_dsi, test_dri);
  rc = Test_TCL_B_Ex(atqb, 0xFF, test_dsi, test_dri);
  if (rc != MI_OK) return rc;

  printf("Test T=CL B, CID=0, dsi=%d, dri=%d\n", test_dsi, test_dri);
  rc = Test_TCL_B_Ex(atqb, 0, test_dsi, test_dri);
  if (rc != MI_OK) return rc;

  if (atqb[10] & 0x01)
  {
    /* Only possible if the card supports the CID */
    printf("Test T=CL B, CID=1, dsi=%d, dri=%d\n", test_dsi, test_dri);
    rc = Test_TCL_B_Ex(atqb, 1, test_dsi, test_dri);
    if (rc != MI_OK) return rc;

    printf("Test T=CL B, CID=13, dsi=%d, dri=%d\n", test_dsi, test_dri);
    rc = Test_TCL_B_Ex(atqb, 13, test_dsi, test_dri);
    if (rc != MI_OK) return rc;

    printf("Test T=CL B, CID=14, dsi=%d, dri=%d\n", test_dsi, test_dri);
    rc = Test_TCL_B_Ex(atqb, 14, test_dsi, test_dri);
    if (rc != MI_OK) return rc;
  }
  
  return MI_OK;
}


