﻿/*
 * ******************************************************************
 * Springcard - CsbApduScriptor
 * Scripting application using SpringProx API
 * Copyright (c) 2004-2014 Pro Active SAS, France
 * Go to www.springcard.com for information and updates
 * ******************************************************************
 */
using System;
using SpringProxAPI;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Drawing;

namespace CsbApduScriptor.Classes
{
	
	public class FormatException : ApplicationException
	{
		public FormatException(string message) : base(message) { }
	}
	
	/// <summary>
	/// A class used to dialog with the UI when using the reader's thread
	/// </summary>
	public class UiAction
	{
		public enum ActionCode
		{
			SetCardUid,				// Display card'UID (or PUPI) on main's form status bar
			AddTextToOutput,		// Add a text to the richtextbox used for ouput
			SetSelectionColor,		// Set color of text of the richtextbox used for ouput
			JobFinished,			// Inform main form that job is finished
		};
		
		/// <summary>
		/// Action to run
		/// </summary>
		public ActionCode Action { get; set; }

		/// <summary>
		/// Action parameters
		/// </summary>
		public byte[] ActionBytes { get; set; }
		public string ActionString { get; set; }
		public string SecondActionString { get; set; }
		public Color ActionColor { get; set; }
	}
	
	/// <summary>
	/// A class used to talk with reader and cards
	/// </summary>
	public class Reader
	{
		#region Constants
		public const short ERROR_CONTINUE = -1;
		public const short ERROR_STOP = -2;
		public const short GO_ON = 0;
		#endregion
		
		#region Class private's vars
		private bool isConnected = false;		// Is reader connected ?
		private string readerComPort = "";		// Reader's COM Port
		public bool runThread = false;			// Used to start and stop the thread
		#endregion

		#region Properties
		public string LastError { get; private set; }	// Last error code from this class
		public string Script { get; set; }				// Content of user script
		public bool StopOnError  { get; set; }			// Stop execution upon error
		public bool AsciiFormat   { get; set; }			// Output format in Ascii
		public bool Answer78164Format   { get; set; }	// Answer is in ISO 7816-4 format
		public bool HexFormat   { get; set; }			// Output format in Hexa.
		#endregion
		
		
		public Reader(string ComPort)
		{
			this.isConnected = false;
			this.readerComPort = ComPort;
		}
		
		/// <summary>
		/// Set reader's COM Port
		/// </summary>
		public string ComPort
		{
			get {
				return this.readerComPort;
			}
			set {
				this.ComPort = value;
			}
		}

		/// <summary>
		/// Connect to the reader
		/// </summary>
		/// <returns></returns>
		public bool Connect()
		{
			this.isConnected = false;
			if(this.readerComPort.Trim().Equals(""))
			{
				LastError = "Error, COM port not set";
				return false;
			}
			
			short res = SPROX.ReaderOpen(readerComPort);
			if(res != SPROX.MI_OK)
			{
				LastError = "Error while opening the COM port " + readerComPort + ", error code : " + res.ToString();
				return false;
			}
			this.isConnected = true;
			return true;
		}
		
		/// <summary>
		/// Close connection with reader
		/// </summary>
		~Reader()
		{
			UnConnect();
		}
		
		
		/// <summary>
		/// Unconnect from current reader
		/// </summary>
		/// <returns></returns>
		public void UnConnect()
		{
			this.runThread = false;
			if(this.isConnected)
			{
				this.isConnected = false;
				SPROX.ReaderClose();
			}
		}
		
		/// <summary>
		/// Enumerate available COM ports on current system
		/// </summary>
		/// <returns></returns>
		public static List<Win32DeviceMgmt.DeviceInfo> enumerateComPorts()
		{
			return Win32DeviceMgmt.GetAllCOMPorts();
		}
		
		/// <summary>
		/// Get library's version
		/// </summary>
		/// <returns></returns>
		public string getDllVersion()
		{
			return SPROX.GetLibrary();
		}

		
		/// <summary>
		/// Returns firmware's version
		/// </summary>
		/// <param name="FirmwareVersion"></param>
		/// <returns></returns>
		public bool getFirmwareVersion(out string FirmwareVersion)
		{
			FirmwareVersion = "";
			if(!this.isConnected)
			{
				LastError = "Error, reader connection not set";
				return false;
			}
			FirmwareVersion = SPROX.ReaderGetFirmware();
			return true;
		}
		
		/// <summary>
		/// Search for card's family
		/// </summary>
		/// <param name="cardFamily"></param>
		/// <returns></returns>
		private string getFamilyName(ushort cardFamily)
		{
			Dictionary<ushort, string> families = new Dictionary<ushort, string>();
			families.Add(0x0000, "No family");
			families.Add(0x0001, "ISO 14443-A (including Mifare) / ISO 18092 @ 106kbit/s / NFC Forum Type 2 and Type 4-A Tags");
			families.Add(0x0002, "ISO 14443-B / NFC Forum Type 4-B Tags");
			families.Add(0x0004, "ISO 15693");
			families.Add(0x0008, "NXP ICODE1");
			families.Add(0x0010, "Inside Contactless PicoPass (also HID iClass)");
			families.Add(0x0020, "ST MicroElectronics SRxxx");
			families.Add(0x0040, "ASK CTS256B or CTS512B");
			families.Add(0x0080, "Innovatron / (legacy Calypso cards – sometimes called 14443-B’)");
			families.Add(0x0100 , "RFU");
			families.Add(0x0200 , "RFU");
			families.Add(0x0400 , "NFC Forum Type 1 Tags (Innovision/Broadcomm chips)");
			families.Add(0x0800 , "Kovio RF barcode");
			families.Add(0x1000 , "JIS:X6319-4 (Felica) / ISO 18092 @ 212 kbit/s and 424 kbit/s / NFC Forum Type 3 Tags");
			families.Add(0x2000 , "RFU");
			families.Add(0x4000 , "RFU");
			families.Add(0x8000 , "RFU");
			foreach(var pair in families)
			{
				if( (cardFamily & pair.Key) == cardFamily)
					return pair.Value;
			}
			return "";
		}
		
		/// <summary>
		/// Activate card's exchange protocol according to card's family
		/// </summary>
		/// <param name="cardFamily"></param>
		/// <returns></returns>
		private bool ActivateCardExchangeAccordingToFamily(ushort cardFamily)
		{
			bool goOn = true;
			short rc;
			switch (cardFamily)
			{
				case 0x0001:	// ISO14443-A
					byte[] ats = new byte[32];
					byte atslen = 32;
					rc = SPROX.TclA_GetAts(0xFF, ats, ref atslen);
					if (rc != SPROX.MI_OK)
					{
						goOn = false;
						this.LastError = "Error, " + SPROX.GetErrorMessage(rc)+". Error code :  "+rc;
					}
					break;
					
				case 0x0002:	// ISO14443-B
					byte[] atq = new byte[11];
					rc = SPROX.B_SelectAny(0x00, atq);
					if (rc == SPROX.MI_OK)
					{
						rc = SPROX.TclB_Attrib(atq, 0xFF);
						if (rc != SPROX.MI_OK)
						{
							goOn = false;
							this.LastError = "Error, " + SPROX.GetErrorMessage(rc)+". Error code :  "+rc;
						}
					}
					else
					{
						goOn = false;
						this.LastError = "Error, " + SPROX.GetErrorMessage(rc)+". Error code :  "+rc;
					}
					break;
					
				default:
					this.LastError = "Error, unknown card family";
					goOn = false;
					break;
			}
			return goOn;
		}
		
		/// <summary>
		/// Dump an array of bytes to the editor's (VS ou SharpDevelop) output
		/// </summary>
		/// <param name="data"></param>
		/// <param name="length"></param>
		private void TraceDebugByteArray(byte[] data, byte length)
		{
			Trace.WriteLine("---------------------------------");
			for(byte i=0; i<length; i++)
				Trace.Write(string.Format("0x{0:X} ", data[i]));
			Trace.Write('\n');
		}
		
		/// <summary>
		/// Send an APDU to the card and get its response
		/// </summary>
		/// <param name="APDU"></param>
		/// <param name="answerCode"></param>
		/// <param name="answerLength"></param>
		/// <returns></returns>
		private byte[] SendApduToCard(byte[] APDU, ref short answerCode, ref ushort answerLength)
		{
			byte[] recv_buffer = new byte[256];
			ushort recv_len = (ushort) recv_buffer.Length;
			TraceDebugByteArray(APDU, (byte) APDU.Length);
			answerCode = SPROX.Tcl_Exchange(0xFF, APDU, (ushort) APDU.Length, recv_buffer, ref answerLength);
			return recv_buffer;
		}
		
		/// <summary>
		/// Set the selection's color of the main's form output richtextbox (from the thread)
		/// </summary>
		/// <param name="Form"></param>
		/// <param name="colorToSet"></param>
		private void SetSelectionColor(MainForm Form, Color colorToSet)
		{
			UiAction uiAction = new UiAction();
			uiAction.Action = UiAction.ActionCode.SetSelectionColor;
			uiAction.ActionColor = colorToSet;
			Form.setInformationFromThread(uiAction);
		}
		
		/// <summary>
		/// Send text to the main's form output richtextbox (from the thread)
		/// </summary>
		/// <param name="Form"></param>
		/// <param name="text"></param>
		private void SendTextToOuptut(MainForm Form, string text)
		{
			UiAction uiAction = new UiAction();
			uiAction.Action = UiAction.ActionCode.AddTextToOutput;
			uiAction.ActionString = text;
			Form.setInformationFromThread(uiAction);
		}
		
		/// <summary>
		/// Used to tell to the caller that the job is finished (from the thread)
		/// </summary>
		/// <param name="Form"></param>
		private void JobDone(MainForm Form)
		{
			UiAction uiAction = new UiAction();
			uiAction.Action = UiAction.ActionCode.JobFinished;
			Form.setInformationFromThread(uiAction);
		}

		/// <summary>
		/// Convert a string representing a command to an array of bytes containing the APDU
		/// </summary>
		/// <param name="command"></param>
		/// <returns></returns>
		private byte[] ConvertCommandToApdu(string command)
		{
			byte[] send_buffer = new byte[command.Length / 2];
			ushort send_len = (ushort) (send_buffer.Length);
			
			if(!HexToArray(ref send_buffer, command, send_len, false))
				throw new FormatException("Command's format is incorrect");
			return send_buffer;
		}
		
		/// <summary>
		/// Send card's answer to ouput
		/// </summary>
		/// <param name="Form"></param>
		/// <param name="CardAnswer"></param>
		/// <param name="cardAnswerCode"></param>
		/// <param name="answerLength"></param>
		/// <returns>A numerical value to identify what to do on return see class's constants</returns>
		private int SendCardAnswerToOutput(MainForm Form, byte[] CardAnswer, short cardAnswerCode, ushort answerLength, string cmd)
		{
			if (answerLength < 2 && Answer78164Format)
			{
				SendTextToOuptut(Form, string.Format("\nCard answer is too short ({0:d}) '" + cmd.ToUpper() + "\n"));
				if (StopOnError)
					return ERROR_STOP;
				return ERROR_CONTINUE;
			}			
			
			Color resultColor;
			string output = "";
			bool errorOccured = cardAnswerCode == SPROX.MI_OK ? false : true;
			if (errorOccured)
			{
				resultColor = Color.Red;
				SetSelectionColor(Form, resultColor);
				SendTextToOuptut(Form, "\n => Error : " + SPROX.GetErrorMessage(cardAnswerCode));
			}
			else
			{
				resultColor = Color.Green;
				output = ArrayToHex(CardAnswer, answerLength);
				SetSelectionColor(Form, resultColor);
				SendTextToOuptut(Form, "\n => Success\n");
			}			
			
			if (HexFormat)
			{
				SetSelectionColor(Form, Color.Black);
				SendTextToOuptut(Form, "\nReceived HEX: ");
				SetSelectionColor(Form, resultColor);
				SendTextToOuptut(Form, HexFormatWithSpace(output));
			}
			
			if (AsciiFormat)
			{
				SetSelectionColor(Form, Color.Black);
				SendTextToOuptut(Form, "\nReceived ASCII: ");
				SetSelectionColor(Form, resultColor);
				SendTextToOuptut(Form, ConvertToString(CardAnswer));
			}

			SendTextToOuptut(Form, "\n");
			
			if (errorOccured)
			{
				if(StopOnError)
					return ERROR_STOP;
				else
					return ERROR_CONTINUE;
			}
			return GO_ON;
		}
		
		/// <summary>
		/// Inform user of the script's start
		/// </summary>
		/// <param name="Form"></param>
		private void InformScriptStart(MainForm Form)
		{
			SetSelectionColor(Form, Color.Black);
			SendTextToOuptut(Form, "-----------------------------" + "\n");
			SendTextToOuptut(Form, "Beginning Script Execution..." + "\n");
			SendTextToOuptut(Form, "-----------------------------" + "\n");			
		}
		
		
		/// <summary>
		/// Code in charge to interpret and send user commands to card
		/// </summary>
		/// <param name="Form"></param>
		/// <returns></returns>
		private bool CommandsInterpreter(MainForm Form)
		{
			char[] delimiter = new char[1];
			delimiter[0] = '\n';
			
			string[] lines = Script.Split(delimiter);		// seperating apdus (delimiter = '\n')
			
			InformScriptStart(Form);
		
			string cmd = "";

			foreach (string line in lines)
			{
				string apdu = line.Trim().Replace(" ","");
				if (apdu.StartsWith("#") || apdu.Equals(""))	// skip comment and blanks lines
					continue;
				
				if (apdu.EndsWith("\\"))	// if line ends with \, concatenation with next line
				{
					cmd = cmd + apdu.Remove(apdu.Length - 1);
					continue;
				}

				cmd = cmd + apdu;
				
				SetSelectionColor(Form, Color.Black);
				SendTextToOuptut(Form, "\nSending : ");
				SetSelectionColor(Form, Color.Blue);
				if(cmd.Length % 2 != 0)
				{
					SetSelectionColor(Form, Color.Red);
					SendTextToOuptut(Form, "\nError, problem with the length of your APDU, verify your code : " + cmd.ToUpper() + "\n");
					if (StopOnError)
						return false;
					else
						continue;
				}
				
				SendTextToOuptut(Form, HexFormatWithSpace(cmd.ToUpper()) );
				
				try {
					byte[] APDU = ConvertCommandToApdu(cmd);
					short cardAnswerCode = 0;		// Answer received by card
					ushort answerLength = 256;		// Answer's length
					byte[] CardAnswer = SendApduToCard(APDU, ref cardAnswerCode, ref answerLength);
					if(cardAnswerCode != SPROX.MI_OK)
					{
						SetSelectionColor(Form, Color.Red);
						SendTextToOuptut(Form, "\n" + "Error, " + SPROX.GetErrorMessage(cardAnswerCode)+". Error code :  " + cardAnswerCode + "\n");
						if (StopOnError)
							return false;
						else
							continue;
					}
					
					int next = SendCardAnswerToOutput(Form, CardAnswer, cardAnswerCode, answerLength, cmd);
					switch (next) {
						case ERROR_CONTINUE:
							continue;
							break;
							
						case ERROR_STOP:
							return false;
							break;
					}
				} catch (FormatException FE) {
					SetSelectionColor(Form, Color.Red);
					SendTextToOuptut(Form, "\n" + FE.Message + " " + cmd.ToUpper() + "\n");
					if (StopOnError)
						return false;
					else
						continue;
				}
				
				cmd = "";
				
			}
			SendTextToOuptut(Form, "\n" + "\n");
			return true;
		}
		

		/// <summary>
		/// Thread to manage the reader
		/// </summary>
		/// <param name="Form"></param>
		public void ReaderDialog(MainForm Form)
		{
			while(runThread)
			{
				if(this.isConnected)
				{
					UiAction uiAction = new UiAction();
					DateTime now = DateTime.Now;
					Trace.WriteLine("Loop in the thread - "+now.ToString());
					
					short rc;																	// Result of calls to DLL
					ushort protocols = 0xFFFF;													// Protocols we want to find
					ushort cardFamily = 0;														// Find's result, on exit, found contactless family (one bit only is set, or none)
					byte[] uid = new byte[16];													// Card's UID (when it is found)
					byte uidLength = 16;														// Length of card's UID
					rc = SPROX.Find(protocols, ref cardFamily, uid, ref uidLength);

					uiAction.ActionString = "";
					uiAction.SecondActionString = "";
					uiAction.ActionBytes = null;
					
					if(rc == SPROX.MI_OK)
					{
						uiAction.Action = UiAction.ActionCode.SetCardUid;
						uiAction.ActionString = ArrayToHex(uid, uidLength);
						uiAction.SecondActionString = HexFormatWithSpace(getFamilyName(cardFamily));
						Form.setInformationFromThread(uiAction);
						
						bool goOn = ActivateCardExchangeAccordingToFamily(cardFamily);
						if(goOn)
						{
							CommandsInterpreter(Form);
							this.runThread = false;
							JobDone(Form);
						}
						else	// There was an error
						{
							SendTextToOuptut(Form, this.LastError);
						}
						SPROX.Tcl_Deselect(0xff);
					}
					else	// Problem while communication
					{
						uiAction.Action = UiAction.ActionCode.SetCardUid;
						Form.setInformationFromThread(uiAction);
					}
				}
			}
		}
		
		
		#region Utilities not directly in relation with the reader, the card or the API
		private bool isASCIIChar(byte b)
		{
			if ((b < 0x20) || (b>= 0x7F))
				return false;
			
			return true;
		}
		
		private string getASCIIChar(byte b)
		{
			byte[] ab = new byte[1];
			ab[0] = b;
			return System.Text.Encoding.ASCII.GetString(ab);
		}
		
		/// <summary>
		/// "Convert" an array of bytes to its ASCII (string) representation
		/// </summary>
		/// <param name="array"></param>
		/// <returns></returns>
		private string ConvertToString(byte[] array)
		{
			string s = "";
			for (int i=0; i<array.Length; i++)
			{
				if (isASCIIChar(array[i]))
				{
					s += getASCIIChar(array[i]);
				} else
				{
					s += " ";
				}
			}
			return s;
		}
		
		byte h2b(char h1, char h2)
		{
			return (byte)((h2q(h1) << 4) + h2q(h2));
		}
		
		byte h2q(char h)
		{
			if ((h >= '0') && (h <= '9')) return (byte)(h - '0');
			if ((h >= 'A') && (h <= 'F')) return (byte)(h - 'A' + 10);
			if ((h >= 'a') && (h <= 'f')) return (byte)(h - 'a' + 10);
			return 0x00;
		}

		
		/// <summary>
		/// Converts a string composed of hexadecimal commands in a bytes array
		/// </summary>
		/// <param name="array"></param>
		/// <param name="str"></param>
		/// <param name="size"></param>
		/// <param name="reverse"></param>
		/// <returns></returns>
		private bool HexToArray(ref byte[] array, string str, int size, bool reverse = false)
		{
			int i, l;
			if ((str == null) || (array == null))
				return false;

			l = str.Length/2;
			for (i=0; i<size; i++)
			{
				if (reverse)
				{
					if (i<l)
						array[size-i-1] = h2b(str[2*i], str[(2*i) + 1]);
					else
						array[size-i-1] = 0x00;
				}
				else
				{
					if (i<l)
						array[i] = h2b(str[2 * i], str[(2 * i) + 1]);
					else
						array[i] = 0x00;
				}
			}
			return true;
		}

		/// <summary>
		/// Convert an array of bytes to its hexadecimal representation
		/// </summary>
		/// <param name="array"></param>
		/// <param name="len"></param>
		/// <param name="reverse"></param>
		/// <returns></returns>
		private string ArrayToHex(byte[] array, int len, bool reverse = false)
		{
			string str = "";
			int i;

			if (reverse)
			{
				for (i = len - 1; i >= 0; i--)
					str += String.Format("{0:X02}", array[i]);
			}
			else
			{
				for (i = 0; i < len; i++)
					str += String.Format("{0:X02}", array[i]);
			}

			return str;
		}
		
		
		/// <summary>
		/// Format a string from 010203AAFF to 01 02 03 AA FF
		/// </summary>
		/// <param name="text"></param>
		/// <returns></returns>
		private string HexFormatWithSpace(string text)
		{
			return Regex.Replace(String.Format("{0:X8}", text), "([0-9A-F]{2})(?!$)","$1 ");
		}
		#endregion
	}
	
}
