﻿/*
 * ******************************************************************
 * Springcard - CsbApduScriptor
 * Scripting application using SpringProx API
 * Copyright (c) 2004-2014 Pro Active SAS, France
 * Go to www.springcard.com for information and updates
 * ******************************************************************
 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Collections.ObjectModel;
using System.Runtime.InteropServices;
using CsbApduScriptor.Classes;
using System.Threading;
using Microsoft.Win32;
using System.Reflection;
using System.Diagnostics;

namespace CsbApduScriptor
{
	/// <summary>
	/// Main form...
	/// </summary>
	public partial class MainForm : Form
	{
		public string ScriptFileName = "";		// Current script's name
		private bool ASCIIFormat = true;		// Diplay result in ASCII format
		private bool HEXFormat = true;			// Display result in Hex format
		
		// For Open and Save dialogs
		private const string filters = "Script files(*.txt)|*.txt|All files (*.*)|*.*";
		private const string defaultExt = ".txt";
		
		private PersistWindowState persistWindowState;		// To persist window's position
		private Reader reader = null;						// Reader object
		Thread ReaderThread = null;							// Thread to communicate with the card
		SynchronizationContext ctx;							// To update User Interface from the Thread
		private string FormTitle = "";						// Backuped to add current filename
		AppMRU appMRU;										// Application most Recent Used Items
		private string ComPort = "";						// Reader's (COM) port
		private ShadowForm shade = null;					// Shadow form for opened windows
		private Action<object, EventArgs> OnComPortClick;	// To manage clicks on COM/USB/TCP ports
		
		public MainForm()
		{
			persistWindowState = new PersistWindowState();
			persistWindowState.Parent = this;
			persistWindowState.RegistryPath = @"Software\SpringCard\"+Application.ProductName;
			ctx = SynchronizationContext.Current;	// To update GUI from the readrer's thread
			InitializeComponent();
			FormTitle = this.Text;
			this.Text = FormTitle + " - [[Unamed new project]]";
			appMRU = new AppMRU(this.miRecent, Application.ProductName, this.myOwnRecentFileGotClicked_handler);
		}
		
		#region Gestion interface
		/// <summary>
		/// Wrap lines
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void CheckBoxWrapCheckedChanged(object sender, EventArgs e)
		{
			richTextBoxScript.WordWrap = checkBoxWrap.Checked;
		}
		
		/// <summary>
		/// Manage output format
		/// </summary>
		private void SetOutputFormat()
		{
			if (radioButtonAscii.Checked)
			{
				ASCIIFormat = true;
				HEXFormat = false;
			}
			else if (radioButtonHex.Checked)
			{
				ASCIIFormat = false;
				HEXFormat = true;
			}
			else if (radioButtonBoth.Checked)
			{
				ASCIIFormat = true;
				HEXFormat = true;
			}
		}
		
		/// <summary>
		/// When user clicks on the button used to display output in Ascii format
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void RadioButtonAsciiCheckedChanged(object sender, EventArgs e)
		{
			SetOutputFormat();
		}
		
		/// <summary>
		/// Clear output result
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void ButClearResClick(object sender, EventArgs e)
		{
			richTextBoxResult.Clear();
		}
		
		/// <summary>
		/// Ask for closing
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void QuitToolStripMenuItemClick(object sender, EventArgs e)
		{
			Close();
		}
		
		/// <summary>
		/// Create a new blank script
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void NewToolStripMenuItemClick(object sender, EventArgs e)
		{
			/* Erasing default file name	*/
			ScriptFileName = "";

			/* Erasing script area	*/
			richTextBoxScript.Text = "";
			
		}
		#endregion
		
		/// <summary>
		/// Write current script to a text file
		/// </summary>
		public void writeFile()
		{
			/* Getting user entry	*/
			string Script = richTextBoxScript.Text.Trim();
			char[] delimiter = new char[1];
			delimiter[0] = '\n';
			
			/* seperating apdus (delimiter = '\n')	*/
			string[] lines = Script.Split(delimiter);

			Stream writeFileStream = new FileStream(ScriptFileName, FileMode.OpenOrCreate, FileAccess.Write);
			StreamWriter writeFileStreamWriter = new StreamWriter(writeFileStream);
			
			foreach (string line in lines)
				writeFileStreamWriter.WriteLine(line);
			
			writeFileStreamWriter.Close();
			writeFileStream.Close();
		}
		
		/// <summary>
		/// Open an existing script
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void OpenToolStripMenuItemClick(object sender, EventArgs e)
		{
			openFileDialog.Filter = filters;
			openFileDialog.FilterIndex = 1;
			openFileDialog.RestoreDirectory = true;
			openFileDialog.DefaultExt= defaultExt;
			openFileDialog.ShowDialog();
		}
		
		/// <summary>
		/// Save script to a text file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void SaveToolStripMenuItemClick(object sender, EventArgs e)
		{
			string Script = richTextBoxScript.Text.Trim();
			
			/* Is there something to save?	*/
			if (!Script.Equals(""))
			{
				if (ScriptFileName.Trim().Equals(""))
				{
					saveFileDialog.Filter = filters;
					saveFileDialog.FilterIndex = 1;
					saveFileDialog.RestoreDirectory = true;
					saveFileDialog.DefaultExt= defaultExt;
					saveFileDialog.ShowDialog();
				}
				else
				{
					writeFile();
				}
			}
			
		}
		
		/// <summary>
		/// Save script to a (new) file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void SaveAsToolStripMenuItemClick(object sender, EventArgs e)
		{
			string Script = richTextBoxScript.Text.Trim();
			
			/* Is there something to save?	*/
			if (Script != "")
			{
				saveFileDialog.Filter = filters;
				saveFileDialog.FilterIndex = 1;
				saveFileDialog.RestoreDirectory = true;
				saveFileDialog.DefaultExt= defaultExt;
				saveFileDialog.ShowDialog();
			}
		}
		
		/// <summary>
		/// Ask for confirmation before to close application
		/// </summary>
		/// <returns></returns>
		private bool AskConfirmQuitAppli()
		{
			if (MessageBox.Show("Are you sure you want to quit?", "Confirmation", MessageBoxButtons.OKCancel) == DialogResult.OK)
				return true;
			return false;
		}
		

		/// <summary>
		/// Try to close reader before to quit
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void MainFormFormClosing(object sender, FormClosingEventArgs e)
		{
			if (richTextBoxScript.Text.Trim() != "" && AskConfirmQuitAppli() == false)
			{
				e.Cancel = true;
				return;
			}

			if(reader != null)
			{
				reader.UnConnect();
				reader.runThread = false;
				Thread.Sleep(100);	// To give thread the necessary time to close itself
				if(ReaderThread != null && ReaderThread.IsAlive)
					ReaderThread.Abort();
			}
		}
		
		/// <summary>
		/// Open choosen file by user
		/// </summary>
		/// <param name="filename"></param>
		private bool OpenFile(string filename)
		{
			if(!File.Exists(filename) || string.IsNullOrWhiteSpace(filename))
				return false;
			try {
				Stream userFileStream = new FileStream(filename, FileMode.Open, FileAccess.Read);
				StreamReader userFileStreamReader = new StreamReader(userFileStream);
				string userFileString = userFileStreamReader.ReadToEnd();
				richTextBoxScript.Text = userFileString;
				ScriptFileName = openFileDialog.FileName;
				this.Text = FormTitle + " - " + ScriptFileName;
				userFileStreamReader.Close();
				userFileStream.Close();
				appMRU.AddRecentFile(filename);
				richTextBoxScript.Focus();
			} catch (Exception Ex) {
				MessageBox.Show("There was a problem while trying to open the selected file : " + Ex.Message);
			}
			return true;
		}

		/// <summary>
		/// Open and read a script file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void ReadScriptFile(object sender, CancelEventArgs e)
		{
			OpenFile(openFileDialog.FileName);
		}
		
		/// <summary>
		/// Save script to a file
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void saveFileDialog_FileOk(object sender, CancelEventArgs e)
		{
			ScriptFileName = saveFileDialog.FileName;
			this.Text = FormTitle + " - " + ScriptFileName;
			writeFile();
		}

		/// <summary>
		/// Show the about box
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void AboutCsSpringproxScriptorToolStripMenuItemClick(object sender, EventArgs e)
		{
			shade = new ShadowForm();
			shade.ShowShadow(this);
			AboutForm about = new AboutForm();
			about.ShowDialog();
			shade.Dispose();
			shade = null;
		}
		
		/// <summary>
		/// Instanciate the reader object
		/// </summary>
		private void LoadReader()
		{
			if(string.IsNullOrWhiteSpace(this.ComPort))
			{
				MessageBox.Show("Error, we can't open a communication with the reader as long as its port is not set");
				return;
			}
			
			eReaderName.Text = "";
			eReaderPort.Text = "";
			
			reader = new Reader(this.ComPort);
			if(!reader.Connect())
			{
				MessageBox.Show("Error while connecting to the reader : "+reader.LastError);
				return;
			}
			eReaderPort.Text = this.ComPort;
			
			// Backup user choice to registry
			WriteSettingString("COM", this.ComPort);

			string FirmwareVersion = "";
			if(reader.getFirmwareVersion(out FirmwareVersion))
				eReaderName.Text = FirmwareVersion;
			else
				MessageBox.Show("Error while searching for the reader's firmware : "+reader.LastError);
		}

		/// <summary>
		/// Load COM Port's list into main menu
		/// </summary>
		public void LoadComPortsList()
		{
			this.OnComPortClick = ComPortClicked_handler;
			ToolStripItem SubMenuComPort;
			string defaultComPort = ReadSettingString("COM", "");
			int ComPortsCount = 0;
			bool ComPortAssigned = false;
			
			foreach (var comPort in Reader.enumerateComPorts())
			{
				ComPortsCount++;
				SubMenuComPort = mnuComPorts.DropDownItems.Add(comPort.decsription + " (" + comPort.name + ")");
				SubMenuComPort.Tag = comPort.name;
				SubMenuComPort.Click += new EventHandler(this.OnComPortClick);
				if(!string.IsNullOrEmpty(defaultComPort) && comPort.name.Equals(defaultComPort))
				{
					ComPortAssigned = true;
					((ToolStripMenuItem)SubMenuComPort).Checked = true;
					this.ComPort = defaultComPort;
				}
			}
			
			// Add the "USB" COM Port *****************************************
			ComPortsCount++;
			SubMenuComPort = mnuComPorts.DropDownItems.Add("USB");
			SubMenuComPort.Tag = "USB";
			SubMenuComPort.Click += new EventHandler(this.OnComPortClick);
			// ****************************************************************
			// TCP Port
			mnuComPorts.DropDownItems.Add(new ToolStripSeparator());
			this.OnComPortClick = TcpPortClicked_handler;
			SubMenuComPort = mnuComPorts.DropDownItems.Add("TCP...");
			SubMenuComPort.Tag = "TCP";
			SubMenuComPort.Click += new EventHandler(this.OnComPortClick);
			
			// If user previously entered a TCP Port
			if(!ComPortAssigned && !string.IsNullOrEmpty(defaultComPort) && !defaultComPort.ToUpper().Equals("USB"))
			{
				// Add the entry, create the event and mark it as default value
				AddTcpPort(defaultComPort.Replace("TCP:", ""));
				ComPortAssigned = true;
			}
			// ****************************************************************
			
			if(ComPortsCount == 0)
			{
				MessageBox.Show("Warning, we were unable to get the list of COM Ports on this system, please verify your computer");
			}
		}
		
		/// <summary>
		/// Uncheck all items in the Ports list
		/// </summary>
		private void UncheckComItems()
		{
			int menusCount = mnuComPorts.DropDownItems.Count;
			for (int i=0; i < menusCount; i++)
			{
				if(mnuComPorts.DropDownItems[i].GetType() == typeof(System.Windows.Forms.ToolStripMenuItem))
					((ToolStripMenuItem)mnuComPorts.DropDownItems[i]).Checked = false;
			}

		}
		
		/// <summary>
		/// Add a TCP Port to the menu and check it
		/// </summary>
		/// <param name="ComPort"></param>
		private void AddTcpPort(string ComPort)
		{
			UncheckComItems();
			ToolStripItem SubMenuComPort;
			this.OnComPortClick = ComPortClicked_handler;
			SubMenuComPort = mnuComPorts.DropDownItems.Add(ComPort);
			
			if(!ComPort.StartsWith("TCP:"))
			   ComPort = "TCP:" + ComPort;
			   
			SubMenuComPort.Tag = ComPort;
			
			SubMenuComPort.Click += new EventHandler(this.OnComPortClick);
			((ToolStripMenuItem)SubMenuComPort).Checked = true;
			this.ComPort = ComPort;
		}
		
		
		void MainFormLoad(object sender, EventArgs e)
		{
			this.ComPort = ReadSettingString("COM", "");	// Load last used port
			if(!string.IsNullOrWhiteSpace(this.ComPort.Trim()))
				LoadReader();
			LoadComPortsList();
			FileVersionInfo i = FileVersionInfo.GetVersionInfo(System.AppDomain.CurrentDomain.FriendlyName);
			string version = i.ProductVersion;
			this.Text += " - Version " + version;			
		}
		
		/// <summary>
		/// Run current script
		/// </summary>
		private void Run()
		{
			if(reader == null)
			{
				MessageBox.Show("Error, reader is not open");
				return;
			}
			
			if(string.IsNullOrWhiteSpace(this.ComPort.Trim()))
			{
				MessageBox.Show("Please select a COM port to use in the settings menu");
				return;
			}
			
			if(richTextBoxScript.Text.Trim() == "")
			{
				MessageBox.Show("Type a script before to run it");
				return;
			}
			btnStop.Enabled = true;
			buttonRun.Enabled = false;
			
			// Setting reader's parameters
			reader.runThread = true;
			reader.Script = richTextBoxScript.Text.Trim();
			reader.StopOnError = cbStopOnError.Checked;
			reader.AsciiFormat = ASCIIFormat;
			reader.Answer78164Format = chkAnswer78164.Checked;
			reader.HexFormat = HEXFormat;
			ReaderThread = new Thread((ThreadStart)delegate { reader.ReaderDialog(this); });
			ReaderThread.IsBackground = true;
			ReaderThread.Start();
			return;
		}
		
		/// <summary>
		/// Run current script
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void ButtonRunClick(object sender, EventArgs e)
		{
			Run();
		}
		
		/// <summary>
		/// Run current script
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void RunScriptToolStripMenuItemClick(object sender, EventArgs e)
		{
			Run();
		}
		
		/// <summary>
		/// Method called from the reader's thread to update the GUI
		/// </summary>
		/// <param name="uid"></param>
		public void setInformationFromThread(UiAction uiAction)
		{
			ctx.Send(delegate(object state)
			         {
			         	switch(uiAction.Action)
			         	{
			         		case UiAction.ActionCode.SetCardUid:
			         			eCardAtr.Text = "";
			         			eCardFamily.Text = "";

			         			if(uiAction.ActionString.Length > 0)	// Display card's UID formated xx xx xx xx ... and its family if available
			         			{
			         				eCardAtr.Text = uiAction.ActionString;
			         				eCardFamily.Text = uiAction.SecondActionString;
			         			}
			         			break;
			         			
			         		case UiAction.ActionCode.SetSelectionColor:	// Set Selection's color
			         			richTextBoxResult.SelectionColor = uiAction.ActionColor;
			         			break;
			         			
			         		case UiAction.ActionCode.AddTextToOutput:	// Append text to the output
			         			richTextBoxResult.AppendText(uiAction.ActionString);
			         			break;
			         			
			         		case UiAction.ActionCode.JobFinished:	// Job finished
			         			btnStop.Enabled = false;
			         			buttonRun.Enabled = true;
			         			break;
			         			
			         		default:
			         			MessageBox.Show("Thread called UI with an unknown action");
			         			break;
			         	}
			         	
			         }, null);
		}
		
		/// <summary>
		/// Stop reader's thread
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void BtnStopClick(object sender, EventArgs e)
		{
			if(reader == null || this.ComPort.Trim().Equals(""))
				return;
			
			reader.runThread = false;
			Thread.Sleep(100);	// To give thread the necessary time to close itself
			
			if(ReaderThread != null && ReaderThread.IsAlive)
				ReaderThread.Abort();
			
			btnStop.Enabled = false;
			buttonRun.Enabled = true;
		}

		/// <summary>
		/// Open a recent script (from the MRU)
		/// </summary>
		/// <param name="obj"></param>
		/// <param name="evt"></param>
		void myOwnRecentFileGotClicked_handler(object obj, EventArgs evt)
		{
			Cursor.Current = Cursors.WaitCursor;
			string FileName = (obj as ToolStripItem).Text;
			if (!OpenFile(FileName))
			{
				appMRU.RemoveRecentFile(FileName);
				Cursor.Current = Cursors.Default;
				return;
			}
			this.Text = FormTitle + " - " + FileName;
			Cursor.Current = Cursors.Default;
		}
		
		/// <summary>
		/// Open reader's information window
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void ReaderToolStripMenuItemClick(object sender, EventArgs e)
		{
			if(reader == null || reader.ComPort.Trim().Equals(""))
			{
				MessageBox.Show("Please connect to your reader by choosing a port before to access this part");
				return;
			}
			shade = new ShadowForm();
			shade.ShowShadow(this);
			ReaderInformation RI = new ReaderInformation(ComPort);
			RI.ShowDialog();
			shade.Dispose();
			shade = null;
		}
		
		/// <summary>
		/// Called when user clicks on a COM Port from the Settings menu
		/// </summary>
		/// <param name="obj"></param>
		/// <param name="evt"></param>
		void ComPortClicked_handler(object obj, EventArgs evt)
		{
			ToolStripItem SubMenuComPort = (ToolStripItem) obj;
			string comPort = (string) SubMenuComPort.Tag;
			
			if(string.IsNullOrEmpty(comPort))
			{
				MessageBox.Show("Not a valid port");
				return;
			}

			UncheckComItems();							// Uncheck everything
			((ToolStripMenuItem)obj).Checked = true;	// Check current item
			
			if(this.reader != null)
				reader.UnConnect();
			this.ComPort = comPort;
			LoadReader();
		}
		
		/// <summary>
		/// When user click on the sub menu called "TCP Port"
		/// </summary>
		/// <param name="obj"></param>
		/// <param name="evt"></param>
		void TcpPortClicked_handler(object obj, EventArgs evt)
		{
			InputBoxResult result = InputBox.Show("Enter TCP Address (for example 192.168.16.248)", "TCP Adress");
			if (result.ReturnCode == DialogResult.OK && !string.IsNullOrWhiteSpace(result.Text.Trim()))
			{
				AddTcpPort(result.Text);
				LoadReader();
			}
		}
		
		
		#region Utilities
		/// <summary>
		/// Read a string from registry
		/// </summary>
		/// <param name="name"></param>
		/// <param name="default_value"></param>
		/// <returns></returns>
		public static string ReadSettingString(string name, string default_value)
		{
			string r = default_value;
			try
			{
				RegistryKey k = Registry.CurrentUser.OpenSubKey("SOFTWARE\\SpringCard\\" + Application.ProductName, false);
				r = (string) k.GetValue(name, default_value);
			}
			catch (Exception) { }
			return r;
		}

		/// <summary>
		/// Write a string to registry
		/// </summary>
		/// <param name="name"></param>
		/// <param name="value"></param>
		public static void WriteSettingString(string name, string value)
		{
			try
			{
				RegistryKey k = Registry.CurrentUser.CreateSubKey("SOFTWARE\\SpringCard\\" + Application.ProductName);
				k.SetValue(name, value);
			}
			catch (Exception) { }
		}
		#endregion
		
		
	}
}
