﻿/*
 * ******************************************************************
 * Springcard - CsbDiag2
 * Scripting application using SpringProx API
 * Copyright (c) 2004-2014 Pro Active SAS, France
 * Go to www.springcard.com for information and updates
 * ******************************************************************
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.Threading;
using System.Reflection;
using System.IO;
using SpringProxAPI;
using Microsoft.Win32;
using System.Diagnostics;
using NLog;
using NLog.Targets;

namespace CsbDiag2
{
	/// <summary>
	/// Description of MainForm.
	/// </summary>
	public partial class MainForm : Form
	{
		private Dictionary<string, string> ComPortsList = new Dictionary<string, string> ();	// Used to have an equivalent between the friendly COM port and its "real" name
		private string ReaderPort = "";		// Port used by reader
		private Reader reader = null;		// Object handling current reader
		Thread ReaderThread = null;			// Thread to communicate with the card
		SynchronizationContext ctx;			// To update User Interface from the Thread
		private bool DebugMode = false;		// Activate / Deactivate debug mode
		private bool PollingRunning = false;		// Start / stop Card polling
		bool ComPortAssigned = false;		// Is default Port set ?
		string defaultComPort = "";			// Last port used
		private static Logger logger = LogManager.GetCurrentClassLogger();	// Log
		
		public MainForm()
		{
			InitializeComponent();
			ctx = SynchronizationContext.Current;	// To update GUI from the readrer's thread
		}
		
		/// <summary>
		/// Set connection string according to user settings
		/// </summary>
		/// <returns></returns>
		private bool SetReaderPort()
		{
			logger.Trace("Set Reader Port");
			if(rdAutoDetect.Checked)				// Auto detect
			{
				this.ReaderPort = "AUTO";
				return true;
			}
			else if (rdUsbDirect.Checked)			// USB connection
			{
				this.ReaderPort = "USB";
				return true;
			}
			else if(rdSerial.Checked)				// Serial port
			{
				if(cbSerialPorts.SelectedIndex != -1)
				{
					string FridenlyName = cbSerialPorts.GetItemText(cbSerialPorts.SelectedItem);
					if(ComPortsList.ContainsKey(FridenlyName))
					{
						this.ReaderPort = ComPortsList[FridenlyName];
						return true;
					}
				}
			}
			else if(chkOtherSerialPort.Checked)		// Serial port - Other
			{
				if(!txtSerialOther.Text.Trim().Equals(""))
				{
					this.ReaderPort = txtSerialOther.Text;
					return true;
				}
			}
			else if(rdTcp.Checked)					// TCP connection
			{
				if(!txtDomain.Text.Trim().Equals(""))
				{
					this.ReaderPort = "TCP:" + txtDomain.Text.Trim();
					return true;
				}
			}
			logger.Debug("Can't find a valid connection parameter");
			return false;
		}
		
		
		void WwwSpringcardLinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			System.Diagnostics.Process.Start("http://www.springcard.com");
		}
		
		/// <summary>
		/// Explain Reader's Features
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void Button2Click(object sender, EventArgs e)
		{
			logger.Trace("Explain features");
			FeaturesList FL = new FeaturesList();
			FL.SetFeatures(txtProductFeatures.Text.Trim());
			FL.Show();
		}
		
		/// <summary>
		/// Load all available COM ports on the current computer
		/// </summary>
		private void ListSystemComPorts()
		{
			logger.Trace("List System Com Ports");
			int ComPortsCount = 0;
			int defaultIndex = 0;
			string ComPortName =  "";
			foreach(var comPort in Win32DeviceMgmt.GetAllCOMPorts())
			{
				ComPortName = comPort.name + " : " + comPort.decsription;
				ComPortsList.Add(ComPortName, comPort.name);
				cbSerialPorts.Items.Add(ComPortName);
				if(!string.IsNullOrEmpty(defaultComPort) && comPort.name.Equals(defaultComPort))
				{
					ComPortAssigned = true;
					this.ReaderPort = defaultComPort;
					rdSerial.Checked = true;
					defaultIndex = ComPortsCount;
				}
				ComPortsCount++;
			}
			
			if(ComPortAssigned)
				cbSerialPorts.SelectedIndex = defaultIndex;
			
			if(ComPortsCount == 0)
			{
				string errorMessage = "Error, we were unable to list COM ports on your computer";
				logger.Debug(errorMessage);
				MessageBox.Show(errorMessage);
			}
		}
		
		/// <summary>
		/// Search for the location of springprox.dll
		/// </summary>
		private void LoadDllLocation()
		{
			logger.Trace("Load DLL location");
			string assemblyFolder = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
			txtDllLocation.Text = Path.Combine(assemblyFolder,"springprox.dll");
		}

		/// <summary>
		/// Display DLL's version
		/// </summary>
		private void LoadDllVersion()
		{
			logger.Trace("Load DLL Version");
			try {
				txtDllVersion.Text = SPROX.GetLibrary();
			} catch (Exception Ex) {
				txtDllVersion.Text = "Error : "+Ex.Message;
			}
		}
		
		/// <summary>
		/// Set the full path to the default debug file
		/// </summary>
		private void SetDebugFile()
		{
			logger.Trace("Set Debug File");
			string filename = System.IO.Path.GetTempPath() + "csbdebug.log";
			txtDebugPath.Text = filename;
			
			// Set logger's output filename
			FileTarget target = LogManager.Configuration.FindTargetByName("logfile") as FileTarget;
			target.FileName = filename;
		}
		
		/// <summary>
		/// Try to reload the previous port used (if it exists)
		/// </summary>
		private void SetDefaultPort()
		{
			logger.Trace("Set Default Port");
			if(ComPortAssigned)
				SetReaderPort();
			else
			{
				if(!string.IsNullOrEmpty(defaultComPort))
				{
					if(defaultComPort.Trim().ToUpper().Equals("USB"))
					{
						rdUsbDirect.Checked = true;
						ComPortAssigned = true;
					}
					else if(defaultComPort.Trim().ToUpper().Equals("AUTO"))
					{
						rdAutoDetect.Checked = true;
						ComPortAssigned = true;
					}
					else if(defaultComPort.Trim().ToUpper().StartsWith("TCP:"))
					{
						rdTcp.Checked = true;
						ComPortAssigned = true;
						txtDomain.Text = defaultComPort.Replace("TCP:", "");
					}
					
					if(ComPortAssigned)
					{
						this.ReaderPort = defaultComPort;
						SetReaderPort();
					}
				}
			}
		}
		
		private string GetAssemblyAttribute<T>(Func<T, string> value)
			where T : Attribute
		{
			T attribute = (T)Attribute.GetCustomAttribute(Assembly.GetExecutingAssembly(), typeof (T));
			return value.Invoke(attribute);
		}
		
		/// <summary>
		/// Product's version in title bar and in the label
		/// </summary>
		private void SetVersionInformation()
		{
			logger.Trace("Set Version Information");
			FileVersionInfo i = FileVersionInfo.GetVersionInfo(System.AppDomain.CurrentDomain.FriendlyName);
			string version = i.ProductVersion;
			this.Text += " - Version " + version;
			StatusLabel.Text = "Not connected";
			label9.Text = "CsbDiag2 v" + version;
			label10.Text = GetAssemblyAttribute<AssemblyCopyrightAttribute>(a => a.Copyright);
		}
		
		
		/// <summary>
		/// On form loading
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void MainFormLoad(object sender, EventArgs e)
		{
			logger.Trace("Main Form Load");
			SetVersionInformation();
			defaultComPort = ReadSettingString("COM", "");		// Load last used port
			ListSystemComPorts();
			SetDefaultPort();
			LoadDllLocation();
			LoadDllVersion();
			SetDebugFile();
			if(cbDebugLevel.Items.Count > 0)
				cbDebugLevel.SelectedIndex = 0;
		}
		
		void RdAutoDetectCheckedChanged(object sender, EventArgs e)
		{
			SetReaderPort();
		}
		
		void RdUsbDirectCheckedChanged(object sender, EventArgs e)
		{
			SetReaderPort();
		}
		
		void RdSerialCheckedChanged(object sender, EventArgs e)
		{
			SetReaderPort();
		}
		
		void CbSerialPortsSelectedIndexChanged(object sender, EventArgs e)
		{
			SetReaderPort();
		}
		
		/// <summary>
		/// Try to set reader's connection according to user choices
		/// </summary>
		private void SetReaderConfiguration()
		{
			logger.Trace("Set Reader Configuration");
			uint settings = 0;
			short result = 0;
			string currentProtocol = "Unknown protocol";
			string currentChannel = "Unknow channel";
			string currentBaudrate = "Unkown baudrate";
			
			result = SPROX.ReaderGetDeviceSettings(ref settings);
			if(result == SPROX.MI_OK)
			{
				logger.Debug("We succeed to get reader's configuration");
				switch (settings & 0x0003)	// We are searching for current protocol
				{
					case 0x00:
						currentProtocol = "OSI";
						break;
						
					case 0x01:
						currentProtocol = "ASCII";
						break;

					case 0x02:
						currentProtocol = "Fast Binary";
						break;
						
					case 0x03:
						currentProtocol = "Binary bus";
						break;
				}
				
				switch (settings & 0x60)				// Channel
				{
					case 0x00:
						currentChannel = "RS232/RS422";
						break;
						
					case 0x20:
						currentChannel = "RS485";
						break;
						
					case 0x40:
						currentChannel = "USB";
						break;
						
					case 0x60:
						currentChannel = "TCP";
						break;
				}
				
				switch (settings & 0x08)	// Baud rates
				{
					case 0x00:
						currentBaudrate = "38400bps";
						break;
						
					case 0x08:
						currentBaudrate = "115200bps";
						break;
				}
				StatusLabel.Text += " - " +  currentProtocol + " - " + currentChannel + " - " + currentBaudrate;
			}
		}
		
		/// <summary>
		/// Fill fields for reader's informations
		/// </summary>
		private void SetDeviceData()
		{
			logger.Trace("Set Device Data");
			uint features = 0;
			short result = 0;
			txtProductName.Text = txtProductFeatures.Text = txtChipsetInfo.Text = txtChipsetId.Text = "";
			txtProductName.Text = SPROX.ReaderGetFirmware();
			result = SPROX.ReaderGetFeatures(ref features);
			if(result == SPROX.MI_OK)
				txtProductFeatures.Text = features.ToString("X");
			
			byte[] rc500_type = new byte[5];
			byte[] rc500_snr = new byte[4];
			result = SPROX.ReaderGetRc500Id(rc500_type, rc500_snr);
			if(result == SPROX.MI_OK)
			{
				txtChipsetInfo.Text = reader.ArrayToHex(rc500_type, 5, false, true);
				txtChipsetId.Text = reader.ArrayToHex(rc500_snr, 4, false, true);
			}
		}
		
		/// <summary>
		/// Test communication with reader
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void BtnTestCommunicationClick(object sender, EventArgs e)
		{
			logger.Trace("Test Communication");
			if(reader == null)
				reader = new Reader("");
			
			if(!reader.IsConnected)
			{
				string ErrorMessage = "";
				Cursor.Current = Cursors.WaitCursor;
				Thread.Sleep(100);	// To see the mouse change
				if(!SetReaderPort())
				{
					ErrorMessage = "Error, please verify your connection settings";
					MessageBox.Show(ErrorMessage);
					logger.Debug(ErrorMessage);
					Cursor.Current = Cursors.Default;
					return;
				}
				if(this.ReaderPort.Trim().Equals(""))
				{
					ErrorMessage = "Error, please select a port to use for the reader";
					logger.Debug(ErrorMessage);
					MessageBox.Show(ErrorMessage);
					Cursor.Current = Cursors.Default;
					return;
				}

				StatusLabel.Text = "Connecting to " + this.ReaderPort;
				reader = new Reader(this.ReaderPort);
				if(!reader.Connect())
				{
					ErrorMessage = "Error while connecting to the reader : "+reader.LastError;
					logger.Debug(ErrorMessage);
					MessageBox.Show(ErrorMessage);
					StatusLabel.Text = "Error while connecting with " + this.ReaderPort;
					Cursor.Current = Cursors.Default;
					return;
				}
				StatusLabel.Text = "";
				Cursor.Current = Cursors.Default;
				StatusLabel.Text = "Connected with : " + this.ReaderPort;
				SetReaderConfiguration();
				SetDeviceData();
				WriteSettingString("COM", this.ReaderPort);	// Save port used for the next time we will use the program
				logger.Trace("Successfully connected with port "+ this.ReaderPort);
				btnTestCommunication.Text = "Disconnect";
				btnContactLessPolling.Enabled = true;
				btnExplain.Enabled = true;
			}
			else	// Unconnect
			{
				if(PollingRunning)
				{
					StopPolling();
					PollingRunning = false;
				}
				
				logger.Trace("Unconnect");
				txtProductName.Text = txtProductFeatures.Text = txtChipsetInfo.Text = txtChipsetId.Text = "";
				StatusLabel.Text = "Not connected";
				Cursor.Current = Cursors.WaitCursor;
				btnTestCommunication.Text = "Test communication";
				Cursor.Current = Cursors.Default;
				reader.UnConnect();
				btnContactLessPolling.Enabled = false;
				btnExplain.Enabled = false;
			}
		}
		
		/// <summary>
		/// When user is asking for quitting
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void MainFormFormClosing(object sender, FormClosingEventArgs e)
		{
			if(reader != null && reader.IsConnected)	// Close reader
			{
				reader.UnConnect();
				reader.runThread = false;
				Thread.Sleep(100);	// To give thread the necessary time to close itself
				if(ReaderThread != null && ReaderThread.IsAlive)
					ReaderThread.Abort();
			}
		}

		/// <summary>
		/// Method called from the reader's thread to update the GUI
		/// </summary>
		/// <param name="uid"></param>
		public void setInformationFromThread(UiAction uiAction)
		{
			logger.Trace("Set information from reader's thread");
			ctx.Send(delegate(object state)
			         {
			         	switch(uiAction.Action)
			         	{
			         		case UiAction.ActionCode.SetCardUid:
			         			lblOutput.Text = "";

			         			if(uiAction.ActionString.Length > 0)	// Display card's UID formated xx xx xx xx ... and its family if available
			         			{
			         				lblOutput.Text = uiAction.ActionString;
			         				lblOutput.Text += Environment.NewLine + uiAction.SecondActionString;
			         			}
			         			break;
			         			
			         		default:
			         			MessageBox.Show("Thread called UI with an unknown action");
			         			break;
			         	}
			         }, null);
		}
		
		/// <summary>
		/// Activate or de activate debug's mode
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void BtnActiveDebugClick(object sender, EventArgs e)
		{
			logger.Trace("Activate of de activate debug's mode");
			DebugMode = !DebugMode;
			if(DebugMode)
				btnActiveDebug.Text = "Desactivate debug file";
			else
				btnActiveDebug.Text = "Activate debug file";
			
			btnSetDebugLevel.Visible = DebugMode;
			label11.Visible = DebugMode;
			cbDebugLevel.Visible = DebugMode;
		}
		
		/// <summary>
		/// On click on "Change debug level" button
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void BtnSetDebugLevelClick(object sender, EventArgs e)
		{
			logger.Trace("Change debug level");
			if(cbDebugLevel.SelectedIndex >= 0)
			{
				LogLevel level = LogLevel.FromString(cbDebugLevel.SelectedItem.ToString());
				logger.Trace("Debug level set to : "+level.ToString());
				var LogginRule = LogManager.Configuration.LoggingRules[0];
				LogginRule.EnableLoggingForLevel(level);
			}
		}
		
		/// <summary>
		/// Automatically activate the TCP connection (if not already done)
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void TxtDomainClick(object sender, EventArgs e)
		{
			rdTcp.Checked = true;
		}
		
		/// <summary>
		/// Automatically activate the Serial connection
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void CbSerialPortsMouseClick(object sender, MouseEventArgs e)
		{
			if(!rdSerial.Checked)
				rdSerial.Checked = true;
		}
		
		/// <summary>
		/// Contactless Polling
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void BtnContactLessPollingClick(object sender, EventArgs e)
		{
			logger.Trace("Contactless Polling");
			lblOutput.Text = "";
			PollingRunning = ! PollingRunning;
			if(PollingRunning)
				StartPolling();
			else
				StopPolling();
		}

		/// <summary>
		/// Start card(s) polling
		/// </summary>
		private void StartPolling()
		{
			string ErrorMessage = "";
			logger.Trace("Start polling");
			if(string.IsNullOrWhiteSpace(this.ReaderPort.Trim()))
			{
				ErrorMessage = "Please select a COM port to use in the settings menu";
				logger.Debug(ErrorMessage);
				MessageBox.Show(ErrorMessage);
				return;
			}
			
			if(reader == null || !reader.IsConnected)
			{
				ErrorMessage = "Error, reader is not open";
				logger.Debug(ErrorMessage);
				MessageBox.Show(ErrorMessage);
				return;
			}
			btnContactLessPolling.Text = "Terminate pooling";
			reader.runThread = true;
			ReaderThread = new Thread((ThreadStart)delegate { reader.ReaderDialog(this); });
			ReaderThread.IsBackground = true;
			ReaderThread.Start();
			logger.Trace("Reader's thread is started");
		}
		
		/// <summary>
		/// Stop card(s) polling
		/// </summary>
		private void StopPolling()
		{
			logger.Trace("Stop polling");
			btnContactLessPolling.Text = "Contactless pooling";
			if(reader == null || this.ReaderPort.Trim().Equals(""))
				return;
			
			reader.runThread = false;
			Thread.Sleep(100);
			if(ReaderThread != null && ReaderThread.IsAlive)
				ReaderThread.Abort();
			lblOutput.Text = "";
		}
		
		/// <summary>
		/// When use type something in the "other" COM port
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void TxtSerialOtherTextChanged(object sender, EventArgs e)
		{
			if(!chkOtherSerialPort.Checked)
				chkOtherSerialPort.Checked = true;
			if(!rdSerial.Checked)
				rdSerial.Checked = true;
		}

		/// <summary>
		/// Automatically activate the Serial connection
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void CbSerialPortsClick(object sender, EventArgs e)
		{
			if(!rdSerial.Checked)
				rdSerial.Checked = true;
		}
		
		#region Utilities
		/// <summary>
		/// Read a string from registry
		/// </summary>
		/// <param name="name"></param>
		/// <param name="default_value"></param>
		/// <returns></returns>
		public static string ReadSettingString(string name, string default_value)
		{
			string r = default_value;
			try
			{
				RegistryKey k = Registry.CurrentUser.OpenSubKey("SOFTWARE\\SpringCard\\" + Application.ProductName, false);
				r = (string) k.GetValue(name, default_value);
			}
			catch (Exception) { }
			return r;
		}

		/// <summary>
		/// Write a string to registry
		/// </summary>
		/// <param name="name"></param>
		/// <param name="value"></param>
		public static void WriteSettingString(string name, string value)
		{
			try
			{
				RegistryKey k = Registry.CurrentUser.CreateSubKey("SOFTWARE\\SpringCard\\" + Application.ProductName);
				k.SetValue(name, value);
			}
			catch (Exception) { }
		}
		#endregion
		
	}
}
