﻿/*
 * ******************************************************************
 * Springcard - CsbDiag2
 * Scripting application using SpringProx API
 * Copyright (c) 2004-2014 Pro Active SAS, France
 * Go to www.springcard.com for information and updates
 * ******************************************************************
 */
using System;
using SpringProxAPI;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Drawing;
using NLog;

namespace CsbDiag2
{
	
	public class FormatException : ApplicationException
	{
		public FormatException(string message) : base(message) { }
	}
	
	/// <summary>
	/// A class used to dialog with the UI when using the reader's thread
	/// </summary>
	public class UiAction
	{
		public enum ActionCode
		{
			SetCardUid,				// Display card'UID (or PUPI) on main's form status bar
			AddTextToOutput,		// Add a text to the richtextbox used for ouput
			SetSelectionColor,		// Set color of text of the richtextbox used for ouput
			JobFinished,			// Inform main form that job is finished
		};
		
		/// <summary>
		/// Action to run
		/// </summary>
		public ActionCode Action { get; set; }

		/// <summary>
		/// Action parameters
		/// </summary>
		public byte[] ActionBytes { get; set; }
		public string ActionString { get; set; }
		public string SecondActionString { get; set; }
		public Color ActionColor { get; set; }
	}
	
	/// <summary>
	/// A class used to talk with reader and cards
	/// </summary>
	public class Reader
	{
		#region Constants
		public const short ERROR_CONTINUE = -1;
		public const short ERROR_STOP = -2;
		public const short GO_ON = 0;
		#endregion
		
		#region Class private's vars
		private bool isConnected = false;		// Is reader connected ?
		private string readerComPort = "";		// Reader's COM Port
		public bool runThread = false;			// Used to start and stop the thread
		private static Logger logger = LogManager.GetCurrentClassLogger();
		#endregion

		#region Properties
		public string LastError { get; private set; }	// Last error code from this class
		public string Script { get; set; }				// Content of user script
		public bool StopOnError  { get; set; }			// Stop execution upon error
		public bool AsciiFormat   { get; set; }			// Output format in Ascii
		public bool Answer78164Format   { get; set; }	// Answer is in ISO 7816-4 format
		public bool HexFormat   { get; set; }			// Output format in Hexa.
		public bool  IsConnected { 
			get { 
				return this.isConnected; 
				} 
			private set {
				this.isConnected = value;
			}
		}
		#endregion
		
		
		public Reader(string ComPort)
		{
			logger.Trace("Reader Constructor");
			this.isConnected = false;
			this.readerComPort = ComPort;
		}
		
		/// <summary>
		/// Set reader's COM Port
		/// </summary>
		public string ComPort
		{
			get {
				logger.Trace("Set reader's port");
				return this.readerComPort;
			}
			set {
				this.ComPort = value;
			}
		}

		/// <summary>
		/// Connect to the reader
		/// </summary>
		/// <returns></returns>
		public bool Connect()
		{
			logger.Trace("Connect reader");
			this.isConnected = false;
			if(this.readerComPort.Trim().Equals(""))
			{
				LastError = "Error, COM port not set";
				logger.Debug(LastError);
				return false;
			}
			
			short res = SPROX.ReaderOpen(readerComPort);
			if(res != SPROX.MI_OK)
			{
				LastError = "Error while opening the COM port " + readerComPort + ", error code : " + res.ToString();
				logger.Debug(LastError);
				return false;
			}
			this.isConnected = true;
			return true;
		}
		
		/// <summary>
		/// Close connection with reader
		/// </summary>
		~Reader()
		{
			logger.Trace("Reader ask for unconnect");
			UnConnect();
		}
		
		
		/// <summary>
		/// Unconnect from current reader
		/// </summary>
		/// <returns></returns>
		public void UnConnect()
		{
			logger.Trace("Reader unconnect");
			this.runThread = false;
			if(this.isConnected)
			{
				logger.Debug("Unconnect from reader");
				this.isConnected = false;
				SPROX.ReaderClose();
			}
		}
		
		/// <summary>
		/// Enumerate available COM ports on current system
		/// </summary>
		/// <returns></returns>
		public static List<Win32DeviceMgmt.DeviceInfo> enumerateComPorts()
		{
			logger.Trace("List all COM ports installed on computer");
			return Win32DeviceMgmt.GetAllCOMPorts();
		}
		
		/// <summary>
		/// Get library's version
		/// </summary>
		/// <returns></returns>
		public string getDllVersion()
		{
			logger.Trace("Get DLL Version");
			return SPROX.GetLibrary();
		}

		
		/// <summary>
		/// Returns firmware's version
		/// </summary>
		/// <param name="FirmwareVersion"></param>
		/// <returns></returns>
		public bool getFirmwareVersion(out string FirmwareVersion)
		{
			logger.Trace("Get Firmware version");
			FirmwareVersion = "";
			if(!this.isConnected)
			{
				LastError = "Error, reader connection not set";
				logger.Debug(LastError);
				return false;
			}
			FirmwareVersion = SPROX.ReaderGetFirmware();
			return true;
		}
		
		/// <summary>
		/// Search for card's family
		/// </summary>
		/// <param name="cardFamily"></param>
		/// <returns></returns>
		private string getFamilyName(ushort cardFamily)
		{
			logger.Trace("Get Family Name");
			Dictionary<ushort, string> families = new Dictionary<ushort, string>();
			families.Add(0x0000, "No family");
			families.Add(0x0001, "ISO 14443-A (including Mifare) / ISO 18092 @ 106kbit/s / NFC Forum Type 2 and Type 4-A Tags");
			families.Add(0x0002, "ISO 14443-B / NFC Forum Type 4-B Tags");
			families.Add(0x0004, "ISO 15693");
			families.Add(0x0008, "NXP ICODE1");
			families.Add(0x0010, "Inside Contactless PicoPass (also HID iClass)");
			families.Add(0x0020, "ST MicroElectronics SRxxx");
			families.Add(0x0040, "ASK CTS256B or CTS512B");
			families.Add(0x0080, "Innovatron / (legacy Calypso cards – sometimes called 14443-B’)");
			families.Add(0x0100 , "RFU");
			families.Add(0x0200 , "RFU");
			families.Add(0x0400 , "NFC Forum Type 1 Tags (Innovision/Broadcomm chips)");
			families.Add(0x0800 , "Kovio RF barcode");
			families.Add(0x1000 , "JIS:X6319-4 (Felica) / ISO 18092 @ 212 kbit/s and 424 kbit/s / NFC Forum Type 3 Tags");
			families.Add(0x2000 , "RFU");
			families.Add(0x4000 , "RFU");
			families.Add(0x8000 , "RFU");
			foreach(var pair in families)
			{
				if( (cardFamily & pair.Key) == cardFamily)
					return pair.Value;
			}
			logger.Debug("Family card not found");
			return "";
		}
		
	
		/// <summary>
		/// Thread to manage the reader
		/// </summary>
		/// <param name="Form"></param>
		public void ReaderDialog(MainForm Form)
		{
			logger.Trace("Reader's thread");
			while(runThread)
			{
				if(this.isConnected)
				{
					UiAction uiAction = new UiAction();
					DateTime now = DateTime.Now;
					logger.Trace("Loop in the thread - "+now.ToString());
					
					short rc;										// Result of calls to DLL
					ushort protocols = 0xFFFF;						// Protocols we want to find
					ushort cardFamily = 0;							// Find's result, on exit, found contactless family (one bit only is set, or none)
					byte[] uid = new byte[16];						// Card's UID (when it is found)
					byte[] protocolInfo = new byte[32];				// Protocol related information			
					byte uidLength = 16;							// Length of card's UID
					byte ProtocolInfoLength = 32;					// Length of protocol information
					rc = SPROX.Find(protocols, ref cardFamily, uid, ref uidLength, protocolInfo, ref ProtocolInfoLength);

					uiAction.ActionString = "";
					uiAction.SecondActionString = "";
					uiAction.ActionBytes = null;
					
					if(rc == SPROX.MI_OK)
					{
						logger.Debug("Connection with card ok");
						uiAction.Action = UiAction.ActionCode.SetCardUid;
						uiAction.ActionString = ArrayToHex(uid, uidLength);
						uiAction.SecondActionString = HexFormatWithSpace(getFamilyName(cardFamily));
						Form.setInformationFromThread(uiAction);
						logger.Debug("Card ATR : "+ uiAction.ActionString + ", Card's Family : " + uiAction.SecondActionString);
					}
					else	// Problem while communication
					{
						uiAction.Action = UiAction.ActionCode.SetCardUid;
						uiAction.ActionString = SPROX.GetErrorMessage(rc);
						logger.Debug("Problem while communication : " + uiAction.ActionString);
						Form.setInformationFromThread(uiAction);
					}
				}
			}
		}
		
		
		#region Utilities not directly in relation with the reader, the card or the API
		private bool isASCIIChar(byte b)
		{
			if ((b < 0x20) || (b>= 0x7F))
				return false;
			
			return true;
		}
		
		private string getASCIIChar(byte b)
		{
			byte[] ab = new byte[1];
			ab[0] = b;
			return System.Text.Encoding.ASCII.GetString(ab);
		}
		
		/// <summary>
		/// "Convert" an array of bytes to its ASCII (string) representation
		/// </summary>
		/// <param name="array"></param>
		/// <returns></returns>
		private string ConvertToString(byte[] array)
		{
			string s = "";
			for (int i=0; i<array.Length; i++)
			{
				if (isASCIIChar(array[i]))
				{
					s += getASCIIChar(array[i]);
				} else
				{
					s += " ";
				}
			}
			return s;
		}
		
		byte h2b(char h1, char h2)
		{
			return (byte)((h2q(h1) << 4) + h2q(h2));
		}
		
		byte h2q(char h)
		{
			if ((h >= '0') && (h <= '9')) return (byte)(h - '0');
			if ((h >= 'A') && (h <= 'F')) return (byte)(h - 'A' + 10);
			if ((h >= 'a') && (h <= 'f')) return (byte)(h - 'a' + 10);
			return 0x00;
		}

		
		/// <summary>
		/// Converts a string composed of hexadecimal commands in a bytes array
		/// </summary>
		/// <param name="array"></param>
		/// <param name="str"></param>
		/// <param name="size"></param>
		/// <param name="reverse"></param>
		/// <returns></returns>
		private bool HexToArray(ref byte[] array, string str, int size, bool reverse = false)
		{
			int i, l;
			if ((str == null) || (array == null))
				return false;

			l = str.Length/2;
			for (i=0; i<size; i++)
			{
				if (reverse)
				{
					if (i<l)
						array[size-i-1] = h2b(str[2*i], str[(2*i) + 1]);
					else
						array[size-i-1] = 0x00;
				}
				else
				{
					if (i<l)
						array[i] = h2b(str[2 * i], str[(2 * i) + 1]);
					else
						array[i] = 0x00;
				}
			}
			return true;
		}

		/// <summary>
		/// Convert an array of bytes to its hexadecimal representation
		/// </summary>
		/// <param name="array"></param>
		/// <param name="len"></param>
		/// <param name="reverse"></param>
		/// <returns></returns>
		public string ArrayToHex(byte[] array, int len, bool reverse = false, bool withSpaces = false)
		{
			string str = "";
			int i;

			if (reverse)
			{
				for (i = len - 1; i >= 0; i--)
				{
					str += String.Format("{0:X02}", array[i]);
					if(withSpaces)
						str += " ";
				}
			}
			else
			{
				for (i = 0; i < len; i++)
				{
					str += String.Format("{0:X02}", array[i]);
					if(withSpaces)
						str += " ";					
				}
			}

			return str;
		}
		
		
		/// <summary>
		/// Format a string from 010203AAFF to 01 02 03 AA FF
		/// </summary>
		/// <param name="text"></param>
		/// <returns></returns>
		private string HexFormatWithSpace(string text)
		{
			return Regex.Replace(String.Format("{0:X8}", text), "([0-9A-F]{2})(?!$)","$1 ");
		}
		#endregion
	}
	
}
